classdef DG4202Interface < FuncGenInterface
    %DG4202INTERFACE Interface class for Rigol DG4202
    %   Detailed explanation goes here
    
    properties (Constant)
        NUM_CHANNELS = 2
        CHANNEL_VOLTAGE_RANGE = [0,2.5;0,2.5]
    end
    
    methods
        function obj = DG4202Interface(portName, logHandle)
            arguments
                portName (1,1) string
                logHandle (1,1) Logger
            end
            %DG4202INTERFACE Construct an instance of this class
            %   Call FuncGenInterface constructor.
            obj = obj@FuncGenInterface(portName, 0, logHandle);
            obj.Port.configureTerminator("LF")

            % Instrument specific setup
            obj.log("Rigol DG4202 setup done")
        end
        
        function setChannelState(obj, channel, state)
            arguments
                obj (1,1) DG4202Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                state (1,1) logical
            end
            %SETCHANNELSTATE Set the output state of a channel
            if state
                obj.Port.writeline(sprintf(":OUTP%d ON", channel))
                obj.log("Channel %d turned on", channel)
            else
                obj.Port.writeline(sprintf(":OUTP%d OFF", channel))
                obj.log("Channel %d turned off", channel)
            end
        end

        function setChannelWaveform(obj, channel, waveform)
            arguments
                obj (1,1) DG4202Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                waveform (1,1) string {mustBeMember(waveform, ["SIN", ...
                    "SQU", "RAMP", "PULS", "NOIS"])}
            end
            %SETCHANNELWAVEFORM Set the output waveform of a channel
            % Only a selected number of shapes supported for now, only SIN
            % really needed anyway.
            obj.Port.writeline(sprintf(":SOUR%d:FUNC %s", channel, ...
                waveform))
            obj.log("Channel %d waveform set to %s", channel, waveform)
        end

        function setChannelFrequency(obj, channel, frequency)
            arguments
                obj (1,1) DG4202Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                frequency (1,1) double
            end
            %SETCHANNELFREQUENCY Set the output frequency of a channel
            obj.Port.writeline(sprintf(":SOUR%d:FREQ %.6f", channel, ...
                frequency))
            obj.log("Channel %d frequency set to %.6f", channel, frequency)
        end

        function setChannelVoltageRange(obj, channel, voltageRange)
            arguments
                obj (1,1) DG4202Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                voltageRange (1,2) double
            end
            %SETCHANNELVOLTAGERANGE Set the output voltage range of the 
            % channel
            lowVoltage = min(voltageRange);
            highVoltage = max(voltageRange);
            if lowVoltage < obj.CHANNEL_VOLTAGE_RANGE(channel, 1)
                error("Low voltage limit too low")
            end
            if highVoltage > obj.CHANNEL_VOLTAGE_RANGE(channel, 2)
                error("High voltage limit too high")
            end
            obj.Port.writeline(sprintf(":SOUR%d:VOLT:LOW %.6f", ...
                channel, lowVoltage))
            obj.Port.writeline(sprintf(":SOUR%d:VOLT:HIGH %.6f", ...
                channel, highVoltage))
            obj.log("Channel %d output voltage range set to %.6f to " + ...
                "%.6f V", lowVoltage, highVoltage)
        end
    end
end

