classdef DP832Interface < SupplyInterface
    %DP832INTERFACE Interface class for Rigol DP832
    %   Detailed explanation goes here
    
    properties (Constant)
        NUM_CHANNELS = 3
        %CHANNEL_VOLTAGE_RANGE = [0,32;0,32;0,5]
        CHANNEL_VOLTAGE_RANGE = [0,6;0,6;0,5] % No need to go above 6V
        CHANNEL_CURRENT_RANGE = [0,3;0,3;0,3]
    end
    
    methods
        function obj = DP832Interface(portName, logHandle)
            arguments
                portName (1,1) string
                logHandle (1,1) Logger
            end
            %DP832INTERFACE Construct an instance of this class
            %   Call SupplyInterface constructor.
            obj = obj@SupplyInterface(portName, 0, logHandle);
            obj.Port.configureTerminator("LF")

            % Instrument specific setup
            obj.log("Rigol DP832 setup done")
        end

        function setChannelState(obj, channel, state)
            arguments
                obj (1,1) DP832Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
                state (1,1) logical
            end
            %SETCHANNELSTATE Set output state of selected channel
            if state
                obj.Port.writeline(sprintf(":OUTP CH%d,ON", channel))
                obj.log("Channel %d turned on", channel)
            else
                obj.Port.writeline(sprintf(":OUTP CH%d,OFF", channel))
                obj.log("Channel %d turned off", channel)
            end
        end
        
        function setChannelVoltage(obj, channel, voltage)
            arguments
                obj (1,1) DP832Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
                voltage (1,1) double
            end
            %SETCHANNELVOLTAGE Set output voltage of selected channel
            obj.log("Try to set channel %d to %f V.", channel, voltage)
            if voltage < obj.CHANNEL_VOLTAGE_RANGE(channel, 1) || ...
                    voltage > obj.CHANNEL_VOLTAGE_RANGE(channel, 2)
                obj.log("Voltage %f out of range for channel %d", ...
                    voltage, channel)
                error("Invalid voltage")
            end
            obj.Port.writeline(sprintf(":SOUR%d:VOLT %2.4f", ...
                channel, voltage))
            obj.log("Channel %d voltage set to %2.3f V", channel, voltage)
        end
        
        function setChannelCurrent(obj, channel, current)
            arguments
                obj (1,1) DP832Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
                current (1,1) double
            end
            %SETCHANNELCURRENT Set output current of selected channel
            obj.log("Try to set channel %d to %f A.", channel, current)
            if current < obj.CHANNEL_CURRENT_RANGE(channel, 1) || ...
                    current > obj.CHANNEL_CURRENT_RANGE(channel, 2)
                obj.log("Current %f out of range for channel %d", ...
                    current, channel)
                error("Invalid current")
            end
            obj.Port.writeline(sprintf(":SOUR%d:CURR %2.4f", ...
                channel, current))
            obj.log("Channel %d current set to %2.3f A", channel, current)
        end

        function voltage = getChannelVoltage(obj, channel)
            arguments (Input)
                obj (1,1) DP832Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
            end
            arguments (Output)
                voltage (1,1) double
            end
            %GETCHANNELVOLTAGE Get output voltage of selected channel
            voltage = obj.Port.writeread(sprintf(":SOUR%d:VOLT?", ...
                channel));
            obj.log("Channel %d voltage = %2.3f V", channel, voltage)
        end

        function current = getChannelCurrent(obj, channel)
            arguments (Input)
                obj (1,1) DP832Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
            end
            arguments (Output)
                current (1,1) double
            end
            %GETCHANNELCURRENT Get output current of selected channel
            current = obj.Port.writeread(sprintf(":SOUR%d:CURR?", ...
                channel));
            obj.log("Channel %d current = %2.3f A", channel, current)
        end

        function voltage = measureChannelVoltage(obj, channel)
            arguments (Input)
                obj (1,1) DP832Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
            end
            arguments (Output)
                voltage (1,1) double
            end
            %GETCHANNELVOLTAGE Get output voltage of selected channel
            voltage = obj.Port.writeread(sprintf(":MEAS:VOLT? CH%d", ...
                channel));
            obj.log("Channel %d measured voltage = %2.3f V", channel, ...
                voltage)
        end

        function current = measureChannelCurrent(obj, channel)
            arguments (Input)
                obj (1,1) DP832Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
            end
            arguments (Output)
                current (1,1) double
            end
            %GETCHANNELCURRENT Get output current of selected channel
            current = obj.Port.writeread(sprintf(":MEAS:CURR? CH%d", ...
                channel));
            obj.log("Channel %d measured current = %2.3f A", channel, ...
                current)
        end
    end
end

