classdef FpgaInterface < InstrumentInterface
    %FPGAINTERFACE Summary of this class goes here
    %   Detailed explanation goes here

    properties (Constant)
        INSTRUMENT_TYPE = "FPGA"
        DEFAULT_SHIFT_SPEED = 500 % 200 kHz
    end

    properties
        WritePause (1,3) double {mustBePositive} = 1.25 * [1920, 45, 2] ...
            / 9600
        ShiftPause (1,2) double {mustBePositive} = 65535 * [1920, 45] / 1e8
        reset logical
    end
    
    methods
        function obj = FpgaInterface(serialPort, baudRate, logHandle)
            arguments
                serialPort (1,1) string
                baudRate (1,1) double
                logHandle (1,1) Logger
            end
            %FPGAINTERFACE Construct an instance of this class
            %   Call InstrumentInterface constructor with FPGA type.
            obj = obj@InstrumentInterface(...
                FpgaInterface.INSTRUMENT_TYPE, ...
                serialPort, ...
                baudRate, ...
                logHandle);

            % Instrument specific setup
            obj.WritePause = 1.25 * [1920, 45, 2] / baudRate;
            obj.ShiftPause = FpgaInterface.DEFAULT_SHIFT_SPEED * ...
                [1920, 45] / 1e8;
            obj.setShiftSpeed(FpgaInterface.DEFAULT_SHIFT_SPEED)
            obj.setReset(1)

            obj.setToZero(0)
            obj.shift(0)
            obj.setToZero(0)

            obj.setToZero(1)
            obj.shift(1)
            obj.setToZero(1)

            obj.log("Instrument setup done")
        end

        function result = isDone(obj, register)
            arguments
                obj (1,1) FpgaInterface
                register (1,1) double {mustBeMember(register, [0, 1])}
            end
            %ISDONE Check if results can be shifted out
            %   Retrieve the status of the selected shift register, true
            %   when results can be shifted out.
            if ~obj.isMock()
                if register == 0
                    obj.Port.write('D', "char")
                    result = (read(obj.Port, 1, "uint8") ~= 0);
                else
                    obj.Port.write('d', "char")
                    result = (read(obj.Port, 1, "uint8") == 0);
                end
            else
                result = true;
            end
            obj.log("Register %d status = %d", register, result)
        end

        function shift(obj, register)
            arguments
                obj (1,1) FpgaInterface
                register (1,1) double {mustBeMember(register, [0, 1])}
            end
            %SHIFT Swap selected register between FPGA and IC
            %   Shift shift register contents from FPGA to IC and from IC
            %   to FPGA.
            if ~obj.isMock()
                if register == 0
                    obj.Port.write('S', "char")
                else
                    obj.Port.write('s', "char")
                end
                pause(obj.ShiftPause(register + 1))
            end
            obj.log("Register %d shifted", register)
        end

        function setToZero(obj, register)
            arguments
                obj (1,1) FpgaInterface
                register (1,1) double {mustBeMember(register, [0, 1])}
            end
            %SETTOZERO Set selected register to all 0
            if ~obj.isMock()
                if register == 0
                    obj.Port.write('W', "char")
                    obj.Port.write(zeros(1, 240), "uint8")
                else
                    obj.Port.write('w', "char")
                    obj.Port.write(zeros(1, 6), "uint8")
                end
                pause(obj.WritePause(register + 1))
            end
            obj.log("Register %d set to all-0", register)
        end

        function writeRegister(obj, register, binString)
            arguments
                obj (1,1) FpgaInterface
                register (1,1) double {mustBeMember(register, [0, 1])}
                binString (1,1) string
            end
            %WRITEREGISTER Write binary string contents to register
            if ~obj.isMock()
                if register == 0
                    if strlength(binString) ~= 1920
                        error("BinString length = %d instead of 1920", ...
                            strlength(binString))
                    end
                    enc = bin2dec(reshape(char(binString),[8,240])');
                    obj.Port.write('W', "char")
                    obj.Port.write(enc, "uint8")
                else
                    if strlength(binString) ~= 45
                        error("BinString length = %d instead of 45", ...
                            strlength(binString))
                    end
                    enc = bin2dec(reshape(char(binString + "000"),[8,6])');
                    obj.Port.write('w', "char")
                    obj.Port.write(enc, "uint8")
                end
                pause(obj.WritePause(register + 1))
            end
            obj.log("Binary string uploaded to register %d", register)
        end

        function result = readRegister(obj, register)
            arguments (Input)
                obj (1,1) FpgaInterface
                register (1,1) double {mustBeMember(register, [0, 1])}
            end
            arguments (Output)
                result (1,1) string
            end
            if ~obj.isMock
                if register == 0
                    obj.Port.write('R', "char")
                    enc = obj.Port.read(240, "uint8");
                    bin = dec2bin(enc, 8)';
                    result = string(bin(:)');
                else
                    obj.Port.write('r', "char")
                    enc = obj.Port.read(6, "uint8");
                    bin = dec2bin(enc, 8)';
                    result = bin(:)';
                    result = string(result(1:45));
                end
            else
                if register == 0
                    result = string(char('0' * ones(1,1920)));
                else
                    result = string(char('0' * ones(1,45)));
                end
            end
            obj.log("Binary string downloaded from register %d", register)
        end

        function setReset(obj, reset)
            arguments
                obj (1,1) FpgaInterface
                reset (1,1) logical
            end
            if ~obj.isMock
                if reset
                    obj.Port.write('C', "char")
                else
                    obj.Port.write('c', "char")
                end
                pause(obj.WritePause(3))
            end
            obj.reset = reset;
            obj.log("Reset register set to %d", reset)
        end

        function setShiftSpeed(obj, period)
            arguments
                obj (1,1) FpgaInterface
                period (1,1) uint16
            end
            if ~obj.isMock
                data = bin2dec(reshape(dec2bin(period, 16)', [8,2])');
                obj.Port.write('F', "char")
                obj.Port.write(data, "uint8")
                pause(obj.WritePause(3))
            end
            obj.ShiftPause = double(period) * [1920, 45] / 1e8;
            obj.log("Set shift register period to %d cycles", period)
        end

        
    end
end

