classdef HMC8043Interface < SupplyInterface
    %HMC8043INTERFACE Interface class for R&S HMC8043
    %   Detailed explanation goes here
    
    properties (Constant)
        NUM_CHANNELS = 3
        %CHANNEL_VOLTAGE_RANGE = [0,32;0,32;0,32]
        CHANNEL_VOLTAGE_RANGE = [0,1.2;0,1.2;0,1.2]
        CHANNEL_CURRENT_RANGE = [0,3;0,3;0,3]
    end
    
    methods
        function obj = HMC8043Interface(portName, logHandle)
            arguments
                portName (1,1) string
                logHandle (1,1) Logger
            end
            %HMC8043INTERFACE Construct an instance of this class
            %   Call SupplyInterface constructor.
            obj = obj@SupplyInterface(portName, 0, logHandle);
            obj.Port.configureTerminator("off", "LF")

            % Instrument specific setup
            obj.log("Rohde & Schwarz HMC8043 setup done")
        end
        
        function setChannelState(obj, channel, state)
            arguments
                obj (1,1) HMC8043Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
                state (1,1) logical
            end
            %SETCHANNELSTATE Set output state of selected channel
            obj.Port.writeline(sprintf("INST:NSEL %d", channel))
            if state
                obj.Port.writeline("OUTP ON")
                obj.log("Channel %d turned on", channel)
            else
                obj.Port.writeline("OUTP OFF")
                obj.log("Channel %d turned off", channel)
            end
        end
        
        function setChannelVoltage(obj, channel, voltage)
            arguments
                obj (1,1) HMC8043Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
                voltage (1,1) double
            end
            %SETCHANNELVOLTAGE Set output voltage of selected channel
            obj.log("Try to set channel %d to %f V", channel, voltage)
            if voltage < obj.CHANNEL_VOLTAGE_RANGE(channel, 1) || ...
                    voltage > obj.CHANNEL_VOLTAGE_RANGE(channel, 2)
                obj.log("Voltage %f out of range for channel %d", ...
                    voltage, channel)
                error("Invalid voltage")
            end
            obj.Port.writeline(sprintf("INST:NSEL %d", channel))
            obj.Port.writeline(sprintf("SOUR:VOLT %2.4f", voltage))
            obj.log("Channel %d set to %2.3f V", channel, voltage)
        end
        
        function setChannelCurrent(obj, channel, current)
            arguments
                obj (1,1) HMC8043Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
                current (1,1) double
            end
            %SETCHANNELCURRENT Set output current of selected channel
            obj.log("Try to set channel %d to %f A", channel, current)
            if current < obj.CHANNEL_CURRENT_RANGE(channel, 1) || ...
                    current > obj.CHANNEL_CURRENT_RANGE(channel, 2)
                obj.log("Current %f out of range for channel %d", ...
                    current, channel)
                error("Invalid current")
            end
            obj.Port.writeline(sprintf("INST:NSEL %d", channel))
            obj.Port.writeline(sprintf("SOUR:CURR %2.4f", current))
            obj.log("Channel %d set to %2.3f A", channel, current)
        end

        function voltage = getChannelVoltage(obj, channel)
            arguments (Input)
                obj (1,1) HMC8043Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
            end
            arguments (Output)
                voltage (1,1) double
            end
            %GETCHANNELVOLTAGE Get output voltage of selected channel
            obj.Port.writeline(sprintf("INST:NSEL %d", channel))
            voltage = obj.Port.writeread("SOUR:VOLT?");
            obj.log("Channel %d voltage = %2.3f V", channel, voltage)
        end

        function current = getChannelCurrent(obj, channel)
            arguments (Input)
                obj (1,1) HMC8043Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
            end
            arguments (Output)
                current (1,1) double
            end
            %GETCHANNELCURRENT Get output current of selected channel
            obj.Port.writeline(sprintf("INST:NSEL %d", channel))
            current = obj.Port.writeread("SOUR:CURR?");
            obj.log("Channel %d current = %2.3f A", channel, current)
        end

        function voltage = measureChannelVoltage(obj, channel)
            arguments (Input)
                obj (1,1) HMC8043Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
            end
            arguments (Output)
                voltage (1,1) double
            end
            %MEASURECHANNELVOLTAGE Get measured output voltage of selected 
            % channel
            obj.Port.writeline(sprintf("INST:NSEL %d", channel))
            obj.Port.writeread("*OPC?");
            voltage = double(obj.Port.writeread("MEAS?"));
            obj.Port.writeread("*OPC?");
            obj.log("Channel %d measured voltage = %2.3f V", channel, ...
                voltage)
        end

        function current = measureChannelCurrent(obj, channel)
            arguments (Input)
                obj (1,1) HMC8043Interface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
            end
            arguments (Output)
                current (1,1) double
            end
            %MEASURECHANNELCURRENT Get measured output current of selected 
            % channel
            obj.Port.writeline(sprintf("INST:NSEL %d", channel))
            current = obj.Port.writeread("MEAS:CURR?");
            obj.log("Channel %d measured current = %2.3f A", channel, ...
                current)
        end
    end
end

