classdef Instruction < handle
    %INSTRUCTION Class describing a microcontroller instruction
    %   This class describes a single microcontroller instruction and is
    %   able to construct it from an assembly line or from/to a binary 
    %   machine code line.
    
    properties (Constant)
        INSTRUCTIONS_ASM = ["beq", "bneq", "blt", "bltq", "add", "xor", ...
            "shl", "shr", "write", "refresh", "read", "read_acc", ...
            "read_stop", "addi", "shli", "shri"]
    end

    properties
        Opcode (1,1) uint8 {mustBeInRange(Opcode, 0, 15)}
        Rs1 (1,1) uint8 {mustBeInRange(Rs1, 0, 31)}
        Rs2Imm (1,1) uint8 {mustBeInRange(Rs2Imm, 0, 31)}
        RdImmb (1,1) uint8 {mustBeInRange(RdImmb, 0, 31)}
    end

    methods (Static)
        function result = reg2uint(str)
            arguments (Input)
                str (1,1) string
            end
            arguments (Output)
                result (1,1) uint8
            end
            result = str2double(erase(str, "$r"));
            if result < 0 || result > 31
                error("Register $r%d does not exist", result)
            end
            result = uint8(result);
        end

        function result = imm2uint(str)
            arguments (Input)
                str (1,1) string
            end
            arguments (Output)
                result (1,1) uint8
            end
            result = mod(str2double(str) + 32, 32);
            if result < 0 || result > 31
                error("Immediate value %d is out of range", result)
            end
            result = uint8(result);
        end
    end
    
    methods
        function obj = Instruction()
            %INSTRUCTION Construct an instance of this class
            %   Detailed explanation goes here
            obj.Opcode = 0;
            obj.Rs1 = 0;
            obj.Rs2Imm = 0;
            obj.RdImmb = 0;
        end
        
        function fromAsm(obj, asmString)
            arguments
                obj (1,1) Instruction
                asmString (1,1) string
            end
            %FROMASM Set properties based on an assembly string
            %   Decode an assembly string into the right field values.
            asmStringSplit = split(asmString);
            opcode = find(asmStringSplit(1) == obj.INSTRUCTIONS_ASM) - 1;
            if isempty(opcode)
                error("Unknown opcode '%s'", asmStringSplit(1))
            end
            if opcode <= 3
                % Branch instruction
                obj.Rs1 = obj.reg2uint(asmStringSplit(2));
                obj.Rs2Imm = obj.reg2uint(asmStringSplit(3));
                obj.RdImmb = obj.imm2uint(asmStringSplit(4));
            elseif opcode >= 13
                % Immediate compute instruction
                obj.Rs1 = obj.reg2uint(asmStringSplit(2));
                obj.Rs2Imm = obj.imm2uint(asmStringSplit(3));
                obj.RdImmb = obj.reg2uint(asmStringSplit(4));
            elseif opcode <= 7 || opcode >= 11
                % Register compute or comparison read instruction
                obj.Rs1 = obj.reg2uint(asmStringSplit(2));
                obj.Rs2Imm = obj.reg2uint(asmStringSplit(3));
                obj.RdImmb = obj.reg2uint(asmStringSplit(4));
            elseif opcode == 8
                % Write instruction
                obj.Rs1 = obj.reg2uint(asmStringSplit(2));
                obj.Rs2Imm = obj.reg2uint(asmStringSplit(3));
            elseif opcode == 9
                % Refresh instruction
                obj.Rs2Imm = obj.reg2uint(asmStringSplit(2));
            elseif opcode == 10
                % Read instruction
                obj.Rs2Imm = obj.reg2uint(asmStringSplit(2));
                obj.RdImmb = obj.reg2uint(asmStringSplit(3));
            end
            obj.Opcode = opcode;
        end

        function asmString = toAsm(obj)
            arguments (Input)
                obj (1,1) Instruction
            end
            arguments (Output)
                asmString (1,1) string
            end
            %TOASM Return ASM string of instruction
            %   Translate the instruction into an ASM string.
            if obj.Opcode <= 3
                % Branch instruction
                asmString = sprintf("%s $r%d $r%d %d", ...
                    obj.INSTRUCTIONS_ASM(obj.Opcode + 1), obj.Rs1, ...
                    obj.Rs2Imm, mod(double(obj.RdImmb) + 16, 32) - 16);
            elseif obj.Opcode >= 13
                % Immediate compute instruction
                asmString = sprintf("%s $r%d %d $r%d", ...
                    obj.INSTRUCTIONS_ASM(obj.Opcode + 1), obj.Rs1, ...
                    mod(double(obj.Rs2Imm) + 16, 32) - 16, obj.RdImmb);
            elseif obj.Opcode <= 7 || obj.Opcode >= 11
                % Register compute or comparison read instruction
                asmString = sprintf("%s $r%d $r%d $r%d", ...
                    obj.INSTRUCTIONS_ASM(obj.Opcode + 1), obj.Rs1, ...
                    obj.Rs2Imm, obj.RdImmb);
            elseif obj.Opcode == 8
                % Write instruction
                asmString = sprintf("%s $r%d $r%d", ...
                    obj.INSTRUCTIONS_ASM(obj.Opcode + 1), obj.Rs1, ...
                    obj.Rs2Imm);
            elseif obj.Opcode == 9
                % Refresh instruction
                asmString = sprintf("%s $r%d", ...
                    obj.INSTRUCTIONS_ASM(obj.Opcode + 1), obj.Rs2Imm);
            elseif obj.Opcode == 10
                % Read instruction
                asmString = sprintf("%s $r%d $r%d", ...
                    obj.INSTRUCTIONS_ASM(obj.Opcode + 1), obj.Rs2Imm, ...
                    obj.RdImmb);
            end
        end

        function fromBinary(obj, binString)
            arguments
                obj (1,1) Instruction
                binString (1,1) string
            end
            %FROMBINARY Set properties based on a binary string
            %   Decode a binary string into the right field values.
            obj.Opcode = bin2dec(extractBetween(binString, 1, 4));
            obj.RdImmb = bin2dec(extractBetween(binString, 5, 9));
            obj.Rs2Imm = bin2dec(extractBetween(binString, 10, 14));
            obj.Rs1 = bin2dec(extractBetween(binString, 15, 19));
        end

        function binString = toBinary(obj)
            arguments (Input)
                obj (1,1) Instruction
            end
            arguments (Output)
                binString (1,1) string
            end
            %TOBINARY Return binary string of instruction
            %   Translate the instruction into a binary string.
            binString = string(dec2bin(obj.Opcode, 4)) + ...
                string(dec2bin(obj.RdImmb, 5)) + ...
                string(dec2bin(obj.Rs2Imm, 5)) + ...
                string(dec2bin(obj.Rs1, 5));
        end

        function clear(obj)
            arguments
                obj (1,1) Instruction
            end
            %CLEAR Clear the instruction by setting everything to 0
            obj.fromBinary("0000000000000000000")
        end
    end
end

