classdef InstrumentInterface < handle
    %INSTRUMENTINTERFACE Superclass for instrument handles
    %   Superclass keeping track of the port names and serial port handles
    %   for all instruments.
    
    properties (Constant)
        MOCK_LITERAL = "MOCK"
    end

    properties
        PortName (1,1) string = ""
        Port = []
        LogHandle Logger = Logger.empty
        Type (1,1) string = ""
    end
    
    methods
        function obj = InstrumentInterface(type, portName, baudRate, ...
                logHandle)
            arguments
                type (1,1) string
                portName (1,1) string
                baudRate (1,1) double {mustBeMember(baudRate, ...
                    [0, 110, 300, 600, 1200, 2400, 4800, 9600, 14400, ...
                    19200, 38400, 57600, 115200, 128000, 256000])}
                logHandle (1,1) Logger
            end
            %INSTRUMENTINTERFACE Construct an instance of this class
            %   Try to connect to instrument on specified port name. If no
            %   port name is provided, a mock instrument is created for
            %   testing.
            obj.Type = type;
            obj.LogHandle = logHandle;
            obj.PortName = portName;
            if obj.PortName == ""
                obj.PortName = InstrumentInterface.MOCK_LITERAL;
            else
                obj.Port = visadev(obj.PortName);
                if baudRate ~= 0
                    obj.Port.BaudRate = baudRate;
                end
            end
            obj.log("Serial connection opened")
        end
        
        function delete(obj)
            %DELETE Destruct class instance
            %   Use destructor to close the connection.
            obj.log("Serial connection closed")
            clear obj.Port
        end

        function result = isMock(obj)
            arguments (Input)
                obj (1,1) InstrumentInterface
            end
            arguments (Output)
                result (1,1) logical
            end
            %ISMOCK Returns whether object is a mock object
            %   Compare the port name with mock literal.
            result = (obj.PortName == InstrumentInterface.MOCK_LITERAL);
        end

        function log(obj, str, varargin)
            arguments
                obj (1,1) InstrumentInterface
                str (1,1) string
            end
            arguments (Repeating)
                varargin
            end
            %LOG Write formatted to log file with timestamp
            %   Write specified string to the log file including a
            %   timestamp and type of instrument prefixed.
            text = sprintf("%s (%s): %s", obj.Type, obj.PortName, str);
            obj.LogHandle.twritef(text, varargin{:})
        end
    end
end

