classdef K2636BInterface < SmuInterface
    %K2636BINTERFACE Summary of this class goes here
    %   Detailed explanation goes here
    
    properties (Constant)
        NUM_CHANNELS = 2
        CHANNEL_STRING = ["smua", "smub"]
        % No need to go above 8V
        CHANNEL_VOLTAGE_RANGE = [0,8;0,8] 
        % No need to go above 10mA
        CHANNEL_CURRENT_RANGE = [0,10e-3;0,10e-3] 
    end
    
    methods
        function obj = K2636BInterface(portName, logHandle)
            arguments
                portName (1,1) string
                logHandle (1,1) Logger
            end
            %K2636BINTERFACE Construct an instance of this class
            %   Call SupplyInterface constructor.
            obj = obj@SmuInterface(portName, 0, logHandle);

            obj.Port.writeline("beeper.enable = 0")
            obj.Port.writeline("smua.reset()")
            obj.Port.writeline("smub.reset()")

            obj.Port.writeline("format.asciiprecision = 10")

            % Set the default voltage/current range
            obj.setChannelRangeVoltage(1, obj.CHANNEL_VOLTAGE_RANGE(1,2))
            obj.setChannelRangeCurrent(1, obj.CHANNEL_CURRENT_RANGE(1,2))
            obj.setChannelRangeVoltage(2, obj.CHANNEL_VOLTAGE_RANGE(2,2))
            obj.setChannelRangeCurrent(2, obj.CHANNEL_CURRENT_RANGE(2,2))

            % Instrument specific setup
            obj.log("Keithley 2636B setup done")
        end

        function setChannelSourceSettling(obj, channel, settlingTime)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                settlingTime (1,1) double {mustBeNonnegative}
            end
            obj.Port.writeline(sprintf("%s.source.delay = %g", ...
                obj.CHANNEL_STRING(channel), settlingTime))
            obj.log("Channel %d delay set to %g", channel, settlingTime)
        end

        function setChannelNplc(obj, channel, nplc)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                nplc (1,1) double {mustBeInRange(nplc, 0.001, 25)}
            end
            obj.Port.writeline(sprintf("%s.measure.nplc = %g", ...
                obj.CHANNEL_STRING(channel), nplc))
            obj.log("Channel %d nplc set to %g", channel, nplc)
        end

        function setChannelSourceState(obj, channel, state)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                state (1,1) logical
            end
            %SETCHANNELSOURCESTATE Set output state of selected channel
            obj.Port.writeline(sprintf("%s.source.output = %d", ...
                obj.CHANNEL_STRING(channel), state))
            if state
                obj.log("Channel %d turned on", channel)
            else
                obj.log("Channel %d turned off", channel)
            end
        end

        function setChannelSourceVoltage(obj, channel, voltage)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                voltage (1,1) double
            end
            %SETCHANNELSOURCEVOLTAGE Set output voltage of selected channel
            obj.log("Try to set channel %d to %g V", channel, voltage)
            if voltage < obj.CHANNEL_VOLTAGE_RANGE(channel, 1) || ...
                    voltage > obj.CHANNEL_VOLTAGE_RANGE(channel, 2)
                obj.log("Voltage %g out of range for channel %d", ...
                    voltage, channel)
                error("Invalid voltage")
            end
            obj.Port.writeline(sprintf("%s.source.levelv = %g", ...
                obj.CHANNEL_STRING(channel), voltage))
            obj.log("Channel %d voltage set to %g V", channel, voltage)
        end

        function setChannelSourceCurrent(obj, channel, current)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                current (1,1) double
            end
            %SETCHANNELCURRENT Set output current of selected channel
            obj.log("Try to set channel %d to %g A", channel, current)
            if current < obj.CHANNEL_CURRENT_RANGE(channel, 1) || ...
                    current > obj.CHANNEL_CURRENT_RANGE(channel, 2)
                obj.log("Current %g out of range for channel %d", ...
                    current, channel)
                error("Invalid current")
            end
            obj.Port.writeline(sprintf("%s.source.leveli = %g", ...
                obj.CHANNEL_STRING(channel), current))
            obj.log("Channel %d current set to %g A", channel, current)
        end

        function setChannelRangeVoltage(obj, channel, voltage)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                voltage (1,1) double
            end
            %SETCHANNELRANGEVOLTAGE Set output voltage range of selected 
            % channel (use with VSOURCE mode)
            obj.log("Try to set channel %d voltage range to %g V", ...
                channel, voltage)
            if voltage < obj.CHANNEL_VOLTAGE_RANGE(channel, 1) || ...
                    voltage > obj.CHANNEL_VOLTAGE_RANGE(channel, 2)
                obj.log("Voltage %g out of range for channel %d", ...
                    voltage, channel)
                error("Invalid voltage")
            end
            obj.Port.writeline(sprintf("%s.source.rangev = %g", ...
                obj.CHANNEL_STRING(channel), voltage))
            obj.log("Channel %d voltage range set to %g V", channel, ...
                voltage)
        end

        function setChannelRangeCurrent(obj, channel, current)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                current (1,1) double
            end
            %SETCHANNELRANGECURRENT Set output current range of selected 
            % channel (use with ISOURCE mode)
            obj.log("Try to set channel %d current range to %g A", ...
                channel, current)
            if current < obj.CHANNEL_CURRENT_RANGE(channel, 1) || ...
                    current > obj.CHANNEL_CURRENT_RANGE(channel, 2)
                obj.log("Current %g out of range for channel %d", ...
                    current, channel)
                error("Invalid current")
            end
            obj.Port.writeline(sprintf("%s.source.rangei = %g", ...
                obj.CHANNEL_STRING(channel), current))
            obj.log("Channel %d current range set to %g A", channel, ...
                current)
        end

        function setChannelLimitVoltage(obj, channel, voltage)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                voltage (1,1) double
            end
            %SETCHANNELLIMITVOLTAGE Set output voltage limit of selected 
            % channel (use with ISOURCE mode)
            obj.log("Try to set channel %d voltage limit to %g V", ...
                channel, voltage)
            if voltage < obj.CHANNEL_VOLTAGE_RANGE(channel, 1) || ...
                    voltage > obj.CHANNEL_VOLTAGE_RANGE(channel, 2)
                obj.log("Voltage %g out of range for channel %d", ...
                    voltage, channel)
                error("Invalid voltage")
            end
            obj.Port.writeline(sprintf("%s.source.limitv = %g", ...
                obj.CHANNEL_STRING(channel), voltage))
            obj.log("Channel %d voltage limit set to %g V", channel, ...
                voltage)
        end

        function setChannelLimitCurrent(obj, channel, current)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                current (1,1) double
            end
            %SETCHANNELRANGECURRENT Set output current limit of selected 
            % channel (use with VSOURCE mode)
            obj.log("Try to set channel %d current limit to %g A", ...
                channel, current)
            if current < obj.CHANNEL_CURRENT_RANGE(channel, 1) || ...
                    current > obj.CHANNEL_CURRENT_RANGE(channel, 2)
                obj.log("Current %g out of range for channel %d", ...
                    current, channel)
                error("Invalid current")
            end
            obj.Port.writeline(sprintf("%s.source.limiti = %g", ...
                obj.CHANNEL_STRING(channel), current))
            obj.log("Channel %d current limit set to %g A", channel, ...
                current)
        end

        function setChannelSense(obj, channel, mode)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                mode (1,1) string {mustBeMember(mode, ["LOCAL", ...
                    "REMOTE"])}
            end
            obj.Port.writeline(sprintf("%s.sense = %s.SENSE_%s", ...
                obj.CHANNEL_STRING(channel), ...
                obj.CHANNEL_STRING(channel), mode))
            obj.log("Channel %d sense mode set to %s", channel, mode)
        end

        function setChannelMode(obj, channel, mode)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                mode (1,1) string {mustBeMember(mode, ["ISOURCE", ...
                    "VSOURCE"])}
            end
            %SETCHANNELMODE Set channel to current source or voltage source
            if mode == "ISOURCE"
                obj.Port.writeline(sprintf("%s.source.func = 0", ...
                    obj.CHANNEL_STRING(channel)))
            else
                obj.Port.writeline(sprintf("%s.source.func = 1", ...
                    obj.CHANNEL_STRING(channel)))
            end
            obj.log("Channel %d mode set to %s", channel, mode)
        end

        function setChannelHighCap(obj, channel, mode)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                mode (1,1) string {mustBeMember(mode, ["ENABLE", ...
                    "DISABLE"])}
            end
            obj.Port.writeline(sprintf("%s.source.highc = " + ...
                "%s.%s", obj.CHANNEL_STRING(channel), ...
                obj.CHANNEL_STRING(channel), mode))
            obj.log("Channel %d high cap mode %s", channel, mode)
        end

        function setChannelDisplay(obj, channel, mode)
            arguments
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
                mode (1,1) string {mustBeMember(mode, ["DCAMPS", ...
                    "DCVOLTS", "OHMS", "WATTS"])}
            end
            obj.Port.writeline(sprintf("display.%s.measure.func = " + ...
                "display.MEASURE_%s", obj.CHANNEL_STRING(channel), mode))
            obj.log("Channel %d display mode set to %s", channel, mode)
        end

        function voltage = getChannelVoltage(obj, channel)
            arguments (Input)
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
            end
            arguments (Output)
                voltage (1,1) double
            end
            %GETCHANNELVOLTAGE Get output voltage of selected channel
            voltage = obj.Port.writeread(sprintf(...
                "print(%s.source.levelv)", obj.CHANNEL_STRING(channel)));
            obj.log("Channel %d voltage = %g V", channel, ...
                voltage)
        end

        function current = getChannelCurrent(obj, channel)
            arguments (Input)
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,2)}
            end
            arguments (Output)
                current (1,1) double
            end
            %GETCHANNELCURRENT Get output current of selected channel
            current = obj.Port.writeread(sprintf(...
                "print(%s.source.leveli)", obj.CHANNEL_STRING(channel)));
            obj.log("Channel %d current = %g A", channel, ...
                current)
        end

        function voltage = measureChannelVoltage(obj, channel)
            arguments (Input)
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
            end
            arguments (Output)
                voltage (1,1) double
            end
            %MEASURECHANNELVOLTAGE Get measured output voltage of selected 
            % channel
            voltage = obj.Port.writeread(sprintf(...
                "print(%s.measure.v())", obj.CHANNEL_STRING(channel)));
            obj.log("Channel %d measured voltage = %g V", channel, ...
                voltage)
        end

        function power = measureChannelPower(obj, channel)
            arguments (Input)
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
            end
            arguments (Output)
                power (1,1) double
            end
            %MEASURECHANNELVOLTAGE Get measured output voltage of selected 
            % channel
            power = obj.Port.writeread(sprintf(...
                "print(%s.measure.p())", obj.CHANNEL_STRING(channel)));
            obj.log("Channel %d measured power = %g W", channel, ...
                power)
        end

        function current = measureChannelCurrent(obj, channel)
            arguments (Input)
                obj (1,1) K2636BInterface
                channel (1,1) uint8 {mustBeInRange(channel,1,3)}
            end
            arguments (Output)
                current (1,1) double
            end
            %MEASURECHANNELCURRENT Get measured output current of selected 
            % channel
            current = obj.Port.writeread(sprintf(...
                "print(%s.measure.i())", obj.CHANNEL_STRING(channel)));
            obj.log("Channel %d measured current = %g A", channel, ...
                current)
        end
    end
end

