classdef MeasurementSetupInterface < handle
    %MEASUREMENTSETUPINTERFACE Controls the entire measurement setup
    %   This class describes the entire measurement setup and allows for an
    %   abstraction of all the instruments.
    
    properties (Constant)
        LOG_FILE_NAME = "MSI"
        FPGA_PORT = "COM5"
        FPGA_BAUD_RATE = 115200
        SMU_PORT = "USB0::0x05E6::0x2636::4093504::0::INSTR"
        DP832_PORT = "USB0::0x1AB1::0x0E11::DP8C233203749::0::INSTR"
        HMC8043_PORT = "USB0::0x0AAD::0x0135::026373825::0::INSTR"
        RELAY_PORT = "COM3"
        RELAY_BAUD_RATE = 9600
        DG4202_PORT = "USB0::0x1AB1::0x0641::DG4E175003418::0::INSTR"

        SMU_TEMP_CHANNEL = 2
        TEMP_SENS_BIAS_CURRENT = 10e-6
        TEMP_STD_CURVE_FILE = "C:\Users\radamsteegt\Documents\MATLAB" + ...
            "\DT-600 Standard Curve Interpolation Table.txt"

        SETTINGS_REGISTER_RESET = [
            bin2dec("000"+"000"+"111111"+"1000"+"10000000"+"11111111");
            bin2dec("000"+"000"+"111111"+"1000"+"10000000"+"11111111");
            bin2dec("000"+"000"+"111111"+"1000"+"10000000"+"11111111");
            bin2dec("000"+"000"+"111111"+"1000"+"10000000"+"11111111");
            bin2dec("000"+"000"+"111111"+"1000"+"10000000"+"11111111");
            bin2dec("000"+"000"+"111111"+"1000"+"10000000"+"11111111");
            bin2dec("0"+"1000"+"0"+"111111"+"1000"+"00100000"+"11111111");
            bin2dec("0"+"1000"+"0"+"111111"+"1000"+"00100000"+"11111111");
            bin2dec("00000000000000000000000"+"0"+"11111111");
            bin2dec("00000000000000000000"+"1"+"0"+"0"+"0000"+"00000")
        ]
        SUPPLY_CHANNEL_REF = 5%3
        SUPPLY_CHANNEL_TREF = 4%2
        SUPPLY_CHANNEL_SNM = 6
    end
    
    properties
        LogHandle (1,:) Logger = Logger.empty
        FpgaHandle (1,:) FpgaInterface = FpgaInterface.empty
        SmuHandle (1,:) SmuInterface = SmuInterface.empty
        SupplyHandle (1,:) SupplyInterface = SupplyInterface.empty
        RelayHandle (1,:) RelayInterface = RelayInterface.empty
        FuncGenHandle (1,:) FuncGenInterface = FuncGenInterface.empty
        TemperatureStandardCurve (2,:) double
    end
    
    methods
        function obj = MeasurementSetupInterface()
            %MEASUREMENTSETUPINTERFACE Construct an instance of this class
            % Create all the instrument handles
            obj.LogHandle = Logger(...
                MeasurementSetupInterface.LOG_FILE_NAME);
            obj.FpgaHandle = FpgaInterface(...
                MeasurementSetupInterface.FPGA_PORT, ...
                MeasurementSetupInterface.FPGA_BAUD_RATE, ...
                obj.LogHandle);
            obj.SmuHandle(1) = K2636BInterface(...
                MeasurementSetupInterface.SMU_PORT, ...
                obj.LogHandle);
            obj.SupplyHandle(1) = DP832Interface(...
                MeasurementSetupInterface.DP832_PORT, ...
                obj.LogHandle);
            obj.SupplyHandle(2) = HMC8043Interface(...
                MeasurementSetupInterface.HMC8043_PORT, ...
                obj.LogHandle);
            obj.RelayHandle = RelayInterface(...
                MeasurementSetupInterface.RELAY_PORT, ...
                MeasurementSetupInterface.RELAY_BAUD_RATE, ...
                obj.LogHandle);
            obj.FuncGenHandle = DG4202Interface(...
                MeasurementSetupInterface.DG4202_PORT, ...
                obj.LogHandle);

            % Set up the system clock on channel 1
            obj.setupFuncGenClock(1)

            % Initialise the temperature sensor SMU channel
            obj.initTemperatureSensor()
            % Load the temperature sensor standard curve
            temp = readtable(obj.TEMP_STD_CURVE_FILE);
            obj.TemperatureStandardCurve = table2array(temp(:,1:2))';

            % Reset all relays
            for iRelay = 1:4
                obj.resetRelay(iRelay)
            end

            % Set initial supply voltages
            obj.setSupplyChannelVoltage(obj.SUPPLY_CHANNEL_TREF, 0.2)
            obj.setSupplyChannelState(obj.SUPPLY_CHANNEL_TREF, 1)
            obj.setSupplyChannelVoltage(obj.SUPPLY_CHANNEL_REF, 0.2)
            obj.setSupplyChannelState(obj.SUPPLY_CHANNEL_REF, 1)
            obj.setSupplyChannelVoltage(obj.SUPPLY_CHANNEL_SNM, 1.1)
            obj.setSupplyChannelState(obj.SUPPLY_CHANNEL_SNM, 1)

            % Reset the memory setting registers
            for iMemory = 1:numel(obj.SETTINGS_REGISTER_RESET)
                address = (iMemory - 1) * 64 + 32;
                data = obj.SETTINGS_REGISTER_RESET(iMemory);
                obj.memoryWrite(data, address)
            end
        end
        
        function destroy(obj)
            %DESTROY Destroy the class instance
            %   Close all interfaces and the log file.
            obj.FpgaHandle.delete()
            obj.SmuHandle.delete()
            obj.SupplyHandle.delete()
            obj.RelayHandle.delete()
            obj.LogHandle.delete()
        end

        % ------------------------ FPGA METHODS ------------------------

        function result = runProgram(obj, program, varargin)
            arguments (Input)
                obj (1,1) MeasurementSetupInterface
                program (1,1) Program
            end
            arguments (Input, Repeating)
                varargin
            end
            arguments (Output)
                result (1,1) Program
            end
            %RUNPROGRAM Run specified program and collect results
            %   Upload the specified program, deassert the reset, wait for
            %   the program to finish, and download the result.
            if numel(varargin) >= 1
                timeoutDuration = varargin{1};
            else
                timeoutDuration = Inf;
            end

            result = Program(program.Name + "_OUT");
            obj.FpgaHandle.setReset(1)
            obj.FpgaHandle.writeRegister(0, program.toBinary)
            obj.FpgaHandle.shift(0)
            obj.FpgaHandle.setReset(0)
            obj.LogHandle.twritef("Program '%s' uploaded", program.Name)
            startTime = tic;
            % Check every 100 milliseconds if the program is done yet
            pause(0.1)
            while ~obj.FpgaHandle.isDone(0) && toc(startTime) < ...
                    timeoutDuration
                pause(0.1)
            end
            obj.FpgaHandle.setReset(1)
            obj.FpgaHandle.shift(0)
            result.fromBinary(obj.FpgaHandle.readRegister(0))
            obj.LogHandle.twritef("Program '%s' downloaded", result.Name)
        end

        function uploadProgram(obj, program)
            arguments
                obj (1,1) MeasurementSetupInterface
                program (1,1) Program
            end
            %UPLOADPROGRAM Upload specified program
            %   Upload the specified program and deassert the reset.
            obj.FpgaHandle.setReset(1)
            obj.FpgaHandle.writeRegister(0,program.toBinary)
            obj.FpgaHandle.shift(0)
            obj.FpgaHandle.setReset(0)
            obj.LogHandle.twritef("Program '%s' uploaded", program.Name)
        end

        function memoryWrite(obj, data, address)
            arguments
                obj (1,1) MeasurementSetupInterface
                data (1,1) uint32
                address (1,1) uint16 {mustBeInRange(address, 0, 1023)}
            end
            binString = "00" + dec2bin(address, 10) + ...
                dec2bin(data, 32) + "1";
            resetState = obj.FpgaHandle.reset;
            obj.FpgaHandle.writeRegister(1, reverse(binString));
            obj.FpgaHandle.shift(1);
            if resetState
                obj.FpgaHandle.setReset(0)
            end
            % Check every 100 milliseconds if the write is done yet
            % If operation is not done in 10s, something went wrong
            tStart = tic;
            while ~obj.FpgaHandle.isDone(1)
                pause(0.1)
                if toc(tStart) > 10
                    error("Manual bus instruction timeout reached")
                end
            end
            if resetState
                obj.FpgaHandle.setReset(1)
            end
            obj.LogHandle.twritef("Written data %d to address %d", ...
                data, address)
        end

        function data = memoryRead(obj, address)
            arguments (Input)
                obj (1,1) MeasurementSetupInterface
                address (1,1) uint16 {mustBeInRange(address, 0, 1023)}
            end
            arguments (Output)
                data (1,1) uint32
            end
            binString = "01" + dec2bin(address, 10) + dec2bin(0, 32) + "1";
            resetState = obj.FpgaHandle.reset;
            obj.FpgaHandle.writeRegister(1, reverse(binString));
            obj.FpgaHandle.shift(1);
            if resetState
                obj.FpgaHandle.setReset(0)
            end
            % Check every 100 milliseconds if the read is done yet
            % If operation is not done in 10s, something went wrong
            tStart = tic;
            while ~obj.FpgaHandle.isDone(1)
                pause(0.1)
                if toc(tStart) > 10
                    error("Manual bus instruction timeout reached")
                end
            end
            obj.FpgaHandle.shift(1);
            if resetState
                obj.FpgaHandle.setReset(1)
            end
            readString = reverse(obj.FpgaHandle.readRegister(1));
            data = uint32(bin2dec(extractBetween(readString, 13, 44)));
            obj.LogHandle.twritef("Read data %d from address %d", ...
                data, address)
        end

        function memoryRefresh(obj, address)
            arguments
                obj (1,1) MeasurementSetupInterface
                address (1,1) uint16 {mustBeInRange(address, 0, 1023)}
            end
            binString = "10" + dec2bin(address, 10) + dec2bin(0, 32) + "1";
            resetState = obj.FpgaHandle.reset;
            obj.FpgaHandle.writeRegister(1, reverse(binString));
            obj.FpgaHandle.shift(1);
            if resetState
                obj.FpgaHandle.setReset(0)
            end
            % Check every 100 milliseconds if the refresh is done yet
            % If operation is not done in 20s, something went wrong
            tStart = tic;
            while ~obj.FpgaHandle.isDone(1)
                pause(0.1)
                if toc(tStart) > 10
                    error("Manual bus instruction timeout reached")
                end
            end
            if resetState
                obj.FpgaHandle.setReset(1)
            end
            obj.LogHandle.twritef("Refreshed address %d", address)
        end

        function memoryReset(obj, resetState)
            arguments
                obj (1,1) MeasurementSetupInterface
                resetState (1,1) logical
            end
            obj.FpgaHandle.setReset(resetState)
        end
        
        % ------------------------ SMU METHODS ------------------------

        function setSmuVoltage(obj, channel, voltage)
            arguments
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBePositive}
                voltage (1,1) double
            end
            %SETSMUVOLTAGE Set output voltage of SMU channel
            [smuIdx, channelIdx] = obj.decodeSmuChannel(channel);
            obj.SmuHandle(smuIdx).setChannelSourceVoltage(channelIdx, ...
                voltage);
            obj.LogHandle.twritef("Set SMU channel %d to %g V", ...
                channel, voltage)
        end

        function setSmuVoltageSafe(obj, channel, voltage, currentLimit)
            arguments
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBePositive}
                voltage (1,1) double
                currentLimit (1,1) double
            end
            %SETSMUVOLTAGESAFE Set output voltage of SMU channel, enable
            %the channel output and wait until the measured voltage is
            %close enough to the set voltage.
            [smuIdx, channelIdx] = obj.decodeSmuChannel(channel);
            obj.SmuHandle(smuIdx).setChannelMode(channelIdx, "VSOURCE")
            obj.SmuHandle(smuIdx).setChannelRangeVoltage(channelIdx, ...
                1.5 * voltage)
            obj.SmuHandle(smuIdx).setChannelSourceVoltage(channelIdx, ...
                voltage)
            obj.SmuHandle(smuIdx).setChannelLimitCurrent(channelIdx, ...
                currentLimit)
            obj.SmuHandle(smuIdx).setChannelDisplay(channelIdx, "DCAMPS")
            obj.SmuHandle(smuIdx).setChannelHighCap(channelIdx, "ENABLE")
            obj.SmuHandle(smuIdx).setChannelNplc(channelIdx, 25)
            obj.SmuHandle(smuIdx).setChannelSourceState(channelIdx, 1)
            pause(0.1)
            while abs(obj.SmuHandle(smuIdx)...
                    .measureChannelVoltage(channelIdx) - voltage) > 1e-2
                % Wait for SMU output to settle within 10mV of desired
                % voltage
                pause(0.1)
            end
            obj.LogHandle.twritef("Set SMU channel %d to %g V", ...
                channel, voltage)
        end

        function setSmuCurrentSafe(obj, channel, current, voltageLimit)
            arguments
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBePositive}
                current (1,1) double
                voltageLimit (1,1) double
            end
            %SETSMUVOLTAGESAFE Set output voltage of SMU channel, enable
            %the channel output and wait until the measured voltage is
            %close enough to the set voltage.
            [smuIdx, channelIdx] = obj.decodeSmuChannel(channel);
            obj.SmuHandle(smuIdx).setChannelMode(channelIdx, "ISOURCE")
            obj.SmuHandle(smuIdx).setChannelRangeCurrent(channelIdx, ...
                1.5 * current)
            obj.SmuHandle(smuIdx).setChannelSourceCurrent(channelIdx, ...
                current)
            obj.SmuHandle(smuIdx).setChannelLimitVoltage(channelIdx, ...
                voltageLimit)
            obj.SmuHandle(smuIdx).setChannelDisplay(channelIdx, "DCVOLTS")
            obj.SmuHandle(smuIdx).setChannelHighCap(channelIdx, "DISABLE")
            obj.SmuHandle(smuIdx).setChannelSourceSettling(channelIdx, 1)
            obj.SmuHandle(smuIdx).setChannelNplc(channelIdx, 25)
            obj.SmuHandle(smuIdx).setChannelSourceState(channelIdx, 1)
            pause(0.1)
            while abs(obj.SmuHandle(smuIdx)...
                    .measureChannelCurrent(channelIdx) - current) > 1e-8
                % Wait for SMU output to settle within 100nA of desired
                % current
                pause(0.1)
            end
            obj.LogHandle.twritef("Set SMU channel %d to %g A", ...
                channel, current)
        end

        function setSmuChannelState(obj, channel, state)
            arguments
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBePositive}
                state (1,1) logical
            end
            %SETSMUCHANNELSTATE Enable or disable output of SMU channel
            [smuIdx, channelIdx] = obj.decodeSmuChannel(channel);
            obj.SmuHandle(smuIdx).setChannelSourceState(channelIdx, state)
            obj.LogHandle.twritef("Set SMU channel %d state to %d", ...
                channel, state)
        end

        function current = getSmuCurrent(obj, channel)
            arguments (Input)
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBePositive}
            end
            arguments (Output)
                current (1,1) double
            end
            %GETSMUCURRENT Get output current of SMU channel
            [smuIdx, channelIdx] = obj.decodeSmuChannel(channel);
            current = obj.SmuHandle(smuIdx)...
                .measureChannelCurrent(channelIdx);
            obj.LogHandle.twritef("SMU channel %d current = %g A", ...
                channel, current)
        end
        
        function setSmuCurrent(obj, channel, current)
            arguments
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBePositive}
                current (1,1) double
            end
            %SETSMUCURRENT Set output current of SMU channel
            [smuIdx, channelIdx] = obj.decodeSmuChannel(channel);
            obj.SmuHandle(smuIdx).setChannelSourceCurrent(channelIdx, ...
                current);
            obj.LogHandle.twritef("Set SMU channel %d to %g A", ...
                channel, current)
        end

        function voltage = getSmuVoltage(obj, channel)
            arguments (Input)
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBePositive}
            end
            arguments (Output)
                voltage (1,1) double
            end
            %GETSMUVOLTAGE Get output voltage of SMU channel
            [smuIdx, channelIdx] = obj.decodeSmuChannel(channel);
            voltage = obj.SmuHandle(smuIdx)...
                .measureChannelVoltage(channelIdx);
            obj.LogHandle.twritef("SMU channel %d voltage = %g V", ...
                channel, voltage)
        end

        function power = getSmuPower(obj, channel)
            arguments (Input)
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBePositive}
            end
            arguments (Output)
                power (1,1) double
            end
            %GETSMUVOLTAGE Get output voltage of SMU channel
            [smuIdx, channelIdx] = obj.decodeSmuChannel(channel);
            power = obj.SmuHandle(smuIdx)...
                .measureChannelPower(channelIdx);
            obj.LogHandle.twritef("SMU channel %d power = %g W", ...
                channel, power)
        end
        
        function [smuIdx, channelIdx] = decodeSmuChannel(obj, ...
                channel)
            arguments (Input)
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger, mustBePositive}
            end
            arguments (Output)
                smuIdx (1,1) double {mustBeInteger, mustBePositive}
                channelIdx (1,1) double {mustBeInteger, mustBePositive}
            end
            %DECODESMUCHANNEL Compute SMU and channel number from global 
            % channel number.
            nChannels = zeros(size(obj.SmuHandle));
            for iSmu = 1:numel(obj.SmuHandle)
                nChannels(iSmu) = ...
                    obj.SmuHandle(iSmu).getNumChannels;
            end
            cumChannels = cumsum(nChannels);
            smuIdx = find(channel <= cumChannels, 1);
            if isempty(smuIdx)
                error("Channel out of range")
            else
                if smuIdx == 1
                    channelIdx = channel;
                else
                    channelIdx = channel - cumChannels(smuIdx - 1);
                end
            end
        end

        % ----- TEMP SENSOR -----

        function initTemperatureSensor(obj)
            arguments
                obj (1,1) MeasurementSetupInterface
            end
            %INITTEMPERATURESENSOR Initialise SMU channel for temperature
            %sensor diode (DT-670B(-CU))
            [smuIdx, channelIdx] = obj.decodeSmuChannel(...
                obj.SMU_TEMP_CHANNEL);
            % Set channel to current source with maximum voltage of 1.8 V
            obj.SmuHandle(smuIdx).setChannelMode(channelIdx, "ISOURCE")
            obj.SmuHandle(smuIdx).setChannelRangeCurrent(channelIdx, ...
                1.5 * obj.TEMP_SENS_BIAS_CURRENT)
            obj.SmuHandle(smuIdx).setChannelSourceCurrent(channelIdx, ...
                obj.TEMP_SENS_BIAS_CURRENT)
            obj.SmuHandle(smuIdx).setChannelLimitVoltage(channelIdx, 1.8)
            obj.SmuHandle(smuIdx).setChannelSense(channelIdx, "REMOTE")
            % Enable the channel output
            obj.SmuHandle(smuIdx).setChannelSourceState(channelIdx, 1)
            obj.LogHandle.twritef("Temperature sensor initialised")
        end

        function temperature = getTemperature(obj)
            arguments
                obj (1,1) MeasurementSetupInterface
            end
            %GETTEMPERATURE Return the temperature range (in K) from the
            %standard curve
            [smuIdx, channelIdx] = obj.decodeSmuChannel(...
                obj.SMU_TEMP_CHANNEL);
            sensorVoltage = obj.SmuHandle(smuIdx)...
                .measureChannelVoltage(channelIdx);
            temperature = interp1(obj.TemperatureStandardCurve(2,:), ...
                obj.TemperatureStandardCurve(1,:), sensorVoltage);
            obj.LogHandle.twritef("Temperature sensor voltage = %f V" + ...
                " (%f K)", sensorVoltage, temperature)
        end
        
        % ------------------------ SUPPLY METHODS ------------------------

        function setSupplyChannelVoltage(obj, channel, voltage)
            arguments
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBePositive}
                voltage (1,1) double
            end
            %SETSUPPLYCHANNELVOLTAGE Set voltage of SUPPLY channel
            [supplyIdx, channelIdx] = obj.decodeSupplyChannel(channel);
            obj.SupplyHandle(supplyIdx).setChannelVoltage(channelIdx, ...
                voltage)
            obj.LogHandle.twritef("Set SUPPLY channel %d to %g V", ...
                channel, voltage)
        end

        function setSupplyChannelVoltageSafe(obj, channel, voltage)
            arguments
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBePositive}
                voltage (1,1) double
            end
            %SETSUPPLYCHANNELVOLTAGE Set voltage of SUPPLY channel
            [supplyIdx, channelIdx] = obj.decodeSupplyChannel(channel);
            obj.SupplyHandle(supplyIdx).setChannelVoltage(channelIdx, ...
                voltage)
            obj.SupplyHandle(supplyIdx).setChannelState(channelIdx, 1)
            pause(0.1)
            voltageMeasured = obj.SupplyHandle(supplyIdx)...
                    .measureChannelVoltage(channelIdx);
            while abs(voltageMeasured - voltage) > 1e-2
                pause(0.1)
                voltageMeasured = obj.SupplyHandle(supplyIdx)...
                    .measureChannelVoltage(channelIdx);
            end
            pause(0.1)
            newVoltageMeasured = obj.SupplyHandle(supplyIdx)...
                    .measureChannelVoltage(channelIdx);
            while newVoltageMeasured ~= voltageMeasured
                pause(0.1)
                voltageMeasured = newVoltageMeasured;
                newVoltageMeasured = obj.SupplyHandle(supplyIdx)...
                    .measureChannelVoltage(channelIdx);
            end
            obj.LogHandle.twritef("Set SUPPLY channel %d to %g V", ...
                channel, voltage)
        end

        function voltage = measureSupplyChannelVoltage(obj, channel)
            arguments (Input)
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBePositive}
            end
            arguments (Output)
                voltage (1,1) double
            end
            %MEASURESUPPLYCHANNELVOLTAGE Measure actual SUPPLY channel
            %output voltage
            [supplyIdx, channelIdx] = obj.decodeSupplyChannel(channel);
            voltage = obj.SupplyHandle(supplyIdx)...
                .measureChannelVoltage(channelIdx);
        end

        function setSupplyChannelState(obj, channel, state)
            arguments
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBePositive}
                state (1,1) logical
            end
            %SETSUPPLYCHANNELSTATE Set state of SUPPLY channel
            [supplyIdx, channelIdx] = obj.decodeSupplyChannel(channel);
            obj.SupplyHandle(supplyIdx).setChannelState(channelIdx, ...
                state)
            obj.LogHandle.twritef("Set SUPPLY channel %d state to %d", ...
                channel, state)
        end
        
        function [supplyIdx, channelIdx] = decodeSupplyChannel(obj, ...
                channel)
            arguments (Input)
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger, mustBePositive}
            end
            arguments (Output)
                supplyIdx (1,1) double {mustBeInteger, mustBePositive}
                channelIdx (1,1) double {mustBeInteger, mustBePositive}
            end
            %DECODESUPPLYCHANNEL Compute supply and channel number from
            %global channel number.
            nChannels = zeros(size(obj.SupplyHandle));
            for iSupply = 1:numel(obj.SupplyHandle)
                nChannels(iSupply) = ...
                    obj.SupplyHandle(iSupply).getNumChannels;
            end
            cumChannels = cumsum(nChannels);
            supplyIdx = find(channel <= cumChannels, 1);
            if isempty(supplyIdx)
                error("Channel out of range")
            else
                if supplyIdx == 1
                    channelIdx = channel;
                else
                    channelIdx = channel - cumChannels(supplyIdx - 1);
                end
            end
        end
        
        % ------------------------ RELAY METHODS ------------------------
        
        function setRelay(obj, channel)
            arguments
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBeNonnegative}
            end
            %SETRELAY Set specified RELAY channel
            obj.RelayHandle.setRelay(channel)
            obj.RelayHandle.waitForDriver()
            obj.LogHandle.twritef("Set RELAY channel %d", channel)
        end

        function resetRelay(obj, channel)
            arguments
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBeNonnegative}
            end
            %RESETRELAY Reset specified RELAY channel
            obj.RelayHandle.resetRelay(channel)
            obj.RelayHandle.waitForDriver()
            obj.LogHandle.twritef("Reset RELAY channel %d", channel)
        end
        
        % ----------------------- FUNC GEN METHODS -----------------------

        function setupFuncGenClock(obj, channel)
            arguments
                obj (1,1) MeasurementSetupInterface
                channel (1,1) double {mustBeInteger,mustBeNonnegative}
            end
            %SETUPFUNCGENCLOCK Configure the clock on the specified channel
            obj.FuncGenHandle.setChannelWaveform(channel, "SIN")
            obj.FuncGenHandle.setChannelFrequency(channel, 1e8)
            obj.FuncGenHandle.setChannelVoltageRange(channel, [0, 2.5])
            obj.FuncGenHandle.setChannelState(channel, 1)
            obj.LogHandle.twritef("Clock configured on function " + ...
                "generator channel %d", channel)
        end

    end
end

