classdef Program < handle
    %PROGRAM Class describing a microprocessor program
    %   Class keeping track of a microprocessor program: the instructions,
    %   register values, and error counters. Can compile a program into a
    %   string that can be uploaded to the IC and decompile the results
    %   into readable data.
    
    properties
        ErrorCounter (32,1) uint16 {mustBeInRange(ErrorCounter,0,1023)}
        Register (31,1) uint32
        InstructionList (32,1) Instruction
        Name (1,1) string
    end
    
    methods
        function obj = Program(name)
            arguments
                name (1,1) string
            end
            %PROGRAM Construct an instance of this class
            %   File the instruction list with 32 separate instructions
            obj.Name = name;
            for iInstruction = 1:32
                obj.InstructionList(iInstruction) = Instruction;
            end
        end

        function fromAsmFile(obj, asmFileName)
            arguments (Input)
                obj (1,1) Program
                asmFileName (1,1) string
            end
            %FROMASMFILE Construct a program from an ASM file
            %   Fill the registers and instruction list based on the
            %   instructions in an ASM file.
            % Clear the program to prevent short ASM files loaded after 
            % long ones being followed by remnants from the long one.
            obj.clear
            % Read ASM file
            asmLines = readlines(asmFileName);
            iInstruction = 1;
            for iLine = asmLines'
                line = strtrim(iLine);
                if strlength(line) == 0 || extract(line,1) == ';'
                    % Skip to the next line when the line is empty or a
                    % comment (starting with ';')
                    continue
                end
                splitLine = split(line);
                if splitLine(1) == "init"
                    regNo = str2double(erase(splitLine(2), "$r"));
                    if contains(splitLine(3), "0b")
                        value = bin2dec(extractAfter(splitLine(3), 2));
                    else
                        value = str2double(splitLine(3));
                    end
                    obj.Register(regNo) = uint32(value);
                else
                    obj.InstructionList(iInstruction).fromAsm(line);
                    iInstruction = iInstruction + 1;
                end
            end
        end

        function asmString = toAsmString(obj)
            arguments (Input)
                obj (1,1) Program
            end
            arguments (Output)
                asmString (64,1) string
            end
            asmString = strings(64,1);
            asmString(1) = sprintf("; %s", obj.Name);
            for iRegister = 1:31
                asmString(iRegister + 1) = sprintf("init $r%d %d", ...
                    iRegister, obj.Register(iRegister));
            end
            for iInstruction = 1:32
                asmString(iInstruction + 32) = ...
                    obj.InstructionList(iInstruction).toAsm;
            end
        end

        function fromBinary(obj, binString)
            arguments
                obj (1,1) Program
                binString (1,1) string
            end
            %FROMBINARY Decode binary string to a program
            %   Set the program properties (error counters, registers,
            %   instruction list) based on the read binary string.
            if strlength(binString) ~= 1920
                error("Incorrect binary string length")
            end
            % Read error counters
            for iErrorCounter = 1:32
                binSubStr = extractBefore(binString, 11);
                obj.ErrorCounter(iErrorCounter) = ...
                    bin2dec(reverse(binSubStr));
                binString = extractAfter(binString, 10);
            end
            % Read registers
            for iRegister = 1:31
                binSubStr = extractBefore(binString, 33);
                obj.Register(iRegister) = ...
                    bin2dec(reverse(binSubStr));
                binString = extractAfter(binString, 32);
            end
            % Read instructions
            for iInstruction = 1:32
                binSubStr = extractBefore(binString, 20);
                obj.InstructionList(iInstruction).fromBinary(...
                    reverse(binSubStr));
                binString = extractAfter(binString, 19);
            end
        end
        
        function binString = toBinary(obj)
            arguments (Input)
                obj (1,1) Program
            end
            arguments (Output)
                binString (1,1) string
            end
            %TOBINARY Returns a binary string representation of the program
            %   Translate the values of the program properties into a
            %   single binary string.
            binString = "";
            % Add error counters
            for iErrorCounter = 1:32
                binString = binString + ...
                    reverse(dec2bin(obj.ErrorCounter(iErrorCounter), 10));
            end
            % Add registers
            for iRegister = 1:31
                binString = binString + ...
                    reverse(dec2bin(obj.Register(iRegister), 32));
            end
            % Add instructions
            for iInstruction = 1:32
                binString = binString + ...
                    reverse(obj.InstructionList(iInstruction).toBinary);
            end
        end

        function clear(obj)
            arguments
                obj (1,1) Program
            end
            %CLEAR Clear the program by resetting all registers and 
            % instructions.
            for iErrorCounter = 1:32
                obj.ErrorCounter(iErrorCounter) = 0;
            end
            for iRegister = 1:31
                obj.Register(iRegister) = 0;
            end
            for iInstruction = 1:32
                obj.InstructionList(iInstruction).clear()
            end
        end
    end
end

