classdef DynamicMemoryCharacterisation < handle
    %DYNAMICMEMORYCHARACTERISATION Summary of this class goes here
    %   Detailed explanation goes here
    
    properties (Constant)
        PROGRAM_FOLDER = "C:\Users\radamsteegt\Documents\MATLAB\" + ...
            "memoryCharacterisation\programs\"
    end

    properties
        MemoryNumber (1,1) uint32
        AddressRows (32,1) uint32
        AddressSettings (1,1) uint32
        Config
        fullMemoryReadTimeCharacterisationReferences (:,2,32) double
        fullMemoryReadTimeCharacterisationErrorCounters (32,:,2,32) double
        fullMemoryReadTimeCharacterisationRBL (2,32,32) double
        fullMemoryReadTimeCharacterisationRBLstd (2,32,32) double
        fullMemoryReadTimeCharacterisationWorstCells (2,3) double
        worstCellsReadTimeCharacterisationReferences (11,:,2) double
        worstCellsReadTimeCharacterisationErrorCounters (32,11,:,2) double
        worstCellsReadTimeCharacterisationRBL (2,:) double
        worstCellsReadTimeCharacterisationRBLstd (2,:) double
        minimumReadTime (1,1) double
        readTimeReferenceEstimate (1,:) double
        firstFailingCellErrorCounters (32, 32, 2, :) double
        firstFailingCellDelay (32, 2, :) double
        fullMemoryRetentionWorstCells (2, 2, :) double
        readSettingReference (1,:) double
        worstCellsRetentionCharacterisationReferences (11,2,:,:) double
        worstCellsRetentionCharacterisationErrorCounters (32,11,2,:,:) double
        worstCellsRetentionCharacterisationHoldTimes (:,:) double
        worstCellsRetentionCharacterisationRBL (2,:,:) double
        worstCellsRetentionCharacterisationRBLstd (2,:,:) double
    end
    
    methods
        function obj = DynamicMemoryCharacterisation(config)
            %MEMORYCHARACTERISATION Construct an instance of this class
            %   Detailed explanation goes here
            obj.MemoryNumber = config.memoryNumber;
            obj.AddressRows = config.memoryNumber * 64 + 0:31;
            obj.AddressSettings = config.memoryNumber * 64 + 32;
            obj.Config = config;
        end
        
        function characteriseReadTimeFullMemory(obj, MSI)
            %CHARACTERISEREADTIMEFULLMEMORY Characterise all cells for a
            % single fixed read time

            % Load the memory settings
            readT1 = obj.Config.fullMemoryReadTimeRT1;
            readT2 = obj.Config.fullMemoryReadTimeRT2;
            writeT1 = obj.Config.fullMemoryReadTimeWT1;
            settingsRegister = obj.createSetting(255, readT1, readT2, ...
                writeT1, 0, 1, 1);

            fprintf("Characterise all cells RT1 = %d\n", readT1)

            % Measurement settings
            maxNumberOfReferences = 200;
            rowRange = 1:32;

            % Allocate result matrices
            errorCounters = NaN(32, maxNumberOfReferences, 2, ...
                numel(rowRange));
            references = NaN(maxNumberOfReferences, 2, numel(rowRange));

            % Load program
            inputProgram = Program("writeReadAccumulate");
            inputProgram.fromAsmFile(obj.PROGRAM_FOLDER + ...
                "writeReadAccumulate.asm")
            
            % Set row/data-independent settings
            inputProgram.Register(4) = obj.AddressSettings;
            inputProgram.Register(5) = settingsRegister;

            for rowIdx = 1:numel(rowRange)
                fprintf("Row %d/%d\n", rowIdx, numel(rowRange))
                % Set row-dependent settings
                inputProgram.Register(6) = obj.AddressRows(rowRange(...
                    rowIdx));
                for data = 0:1
                    dataIdx = data + 1;
                    % Set data-dependent settings
                    inputProgram.Register(5) = bitset(...
                        inputProgram.Register(5), 29, 1-data);
                    inputProgram.Register(7) = 0xFFFFFFFF * data;

                    % Measure characterisation limits first, then perform
                    % binary search on each cell of the row.
                    references(1, dataIdx, rowIdx) = ...
                        obj.Config.characterisationLimits(1);
                    references(2, dataIdx, rowIdx) = ...
                        obj.Config.characterisationLimits(2);

                    for refIdx = 1:2
                        MSI.setSupplyChannelVoltageSafe( ...
                            MSI.SUPPLY_CHANNEL_REF, references( ...
                            refIdx, dataIdx, rowIdx))
                        outputProgram = MSI.runProgram(inputProgram);
                        errorCounters(:, refIdx, dataIdx, rowIdx) = ...
                            outputProgram.ErrorCounter(:);
                    end
                    
                    % Check if references are far enough apart for all
                    % cells
                    if max(errorCounters(:, 1, dataIdx, rowIdx)) > 511 ...
                            || min(errorCounters(:, 2, dataIdx, ...
                            rowIdx)) < 512
                        error("Not all cells can be characterised " + ...
                            "using the current characterisation limit!")
                    end
                    
                    % Loop over the cells
                    target = 1023 / 2;
                    refIdx = 3;
                    for cellIdx = 1:32
                        % Find initial lower/upper bound
                        minRef = max(references(errorCounters(cellIdx, ...
                            :, dataIdx, rowIdx) < target, dataIdx, ...
                            rowIdx));
                        maxRef = min(references(errorCounters(cellIdx, ...
                            :, dataIdx, rowIdx) > target, dataIdx, ...
                            rowIdx));

                        while abs(maxRef - minRef) > 1e-3
                            newRef = (minRef + maxRef) / 2;
                            references(refIdx, dataIdx, rowIdx) = newRef;
                            MSI.setSupplyChannelVoltageSafe( ...
                                MSI.SUPPLY_CHANNEL_REF, references( ...
                                refIdx, dataIdx, rowIdx))
                            outputProgram = MSI.runProgram( ...
                                inputProgram);
                            errorCounters(:, refIdx, dataIdx, rowIdx) = ...
                                outputProgram.ErrorCounter(:);
                            if errorCounters(cellIdx, refIdx, dataIdx, ...
                                    rowIdx) < target
                                minRef = newRef;
                            else
                                maxRef = newRef;
                            end
                            refIdx = refIdx + 1;
                            if refIdx > maxNumberOfReferences
                                break
                            end
                        end
                        fprintf("%c", 128)
                        if refIdx > maxNumberOfReferences
                            break
                        end
                    end
                    fprintf("\n")
                end
            end

            obj.fullMemoryReadTimeCharacterisationReferences = references;
            obj.fullMemoryReadTimeCharacterisationErrorCounters = ...
                errorCounters;

            fprintf("All cells characterisation measurements done\n")
        end

        function processReadTimeFullMemoryData(obj)
            %PROCESSREADTIMEFULLMEMORYDATA Compute the RBL voltages from
            %the full memory read time characterisation data
            references = obj.fullMemoryReadTimeCharacterisationReferences;
            errorCounters = ...
                obj.fullMemoryReadTimeCharacterisationErrorCounters;
            obj.fullMemoryReadTimeCharacterisationWorstCells = NaN(2,3);
            % Calculate the RBL voltage of each cell/data combination
            for rowIdx = 1:size(references, 3)
                for cellIdx = 1:32
                    for dataIdx = 1:2
                        refVoltages = references(:, dataIdx, rowIdx);
                        avgOutputData = errorCounters(cellIdx, :, ...
                            dataIdx, rowIdx) / 1023;
                        nanMask = isnan(refVoltages);
                        refVoltages = refVoltages(~nanMask);
                        avgOutputData = avgOutputData(~nanMask);

                        fcn = @(b,x) normcdf(x, b(1), b(2));
                        NRCF = @(b) norm(avgOutputData - fcn(b, ...
                            refVoltages));
                        [~,voltageEstimateIdx] = min(abs( ...
                            avgOutputData - 0.5));
                        voltageEstimate = refVoltages(voltageEstimateIdx);
                        B = fminsearch(NRCF, [voltageEstimate; 2e-3]);

                        obj.fullMemoryReadTimeCharacterisationRBL(...
                            dataIdx, rowIdx, cellIdx) = B(1);
                        obj.fullMemoryReadTimeCharacterisationRBLstd(...
                            dataIdx, rowIdx, cellIdx) = B(2);

                        if dataIdx == 1
                            if ~(B(1) > obj.fullMemoryReadTimeCharacterisationWorstCells(dataIdx, 1))
                                obj.fullMemoryReadTimeCharacterisationWorstCells(dataIdx, 1) = B(1);
                                obj.fullMemoryReadTimeCharacterisationWorstCells(dataIdx, 2) = rowIdx;
                                obj.fullMemoryReadTimeCharacterisationWorstCells(dataIdx, 3) = cellIdx;
                            end
                        else
                            if ~(B(1) < obj.fullMemoryReadTimeCharacterisationWorstCells(dataIdx, 1))
                                obj.fullMemoryReadTimeCharacterisationWorstCells(dataIdx, 1) = B(1);
                                obj.fullMemoryReadTimeCharacterisationWorstCells(dataIdx, 2) = rowIdx;
                                obj.fullMemoryReadTimeCharacterisationWorstCells(dataIdx, 3) = cellIdx;
                            end
                        end
                    end
                end
            end
        end

        function characteriseReadTimeWorstCells(obj, MSI)
            %CHARACTERISEREADTIMEWORSTCELLS Characterise worst cells for a
            % range of read times

            % Load the memory settings
            readT1Range = obj.Config.worstCellReadTimeRT1Range;
            readT2 = obj.Config.fullMemoryReadTimeRT2;
            writeT1 = obj.Config.fullMemoryReadTimeWT1;

            fprintf("Characterise worst cells\n")

            % Measurement settings
            maxNumberOfReferences = 11;

            % Allocate result matrices
            errorCounters = NaN(32, maxNumberOfReferences, numel(readT1Range), 2);
            references = NaN(maxNumberOfReferences, numel(readT1Range), 2);

            % Load program
            inputProgram = Program("writeReadAccumulate");
            inputProgram.fromAsmFile(obj.PROGRAM_FOLDER + ...
                "writeReadAccumulate.asm")
            
            % Set row/data-independent settings
            inputProgram.Register(4) = obj.AddressSettings;

            for data = 0:1
                dataIdx = data + 1;
                rowIdx = obj.fullMemoryReadTimeCharacterisationWorstCells(dataIdx, 2);
                colIdx = obj.fullMemoryReadTimeCharacterisationWorstCells(dataIdx, 3);
                % Set row/data-dependent settings
                inputProgram.Register(6) = obj.AddressRows(rowIdx);
                inputProgram.Register(7) = 0xFFFFFFFF * data;
                for settingIdx = 1:numel(readT1Range)
                    readT1 = readT1Range(settingIdx);
                    settingsRegister = obj.createSetting(255, readT1, ...
                        readT2, writeT1, 0, 1, 1 - data);
                    inputProgram.Register(5) = settingsRegister;
                    
                    % Measure characterisation limits first, then perform
                    % binary search on each cell of the row.
                    references(1, settingIdx, dataIdx) = ...
                        obj.Config.characterisationLimits(1);
                    references(2, settingIdx, dataIdx) = ...
                        obj.Config.characterisationLimits(2);

                    for refIdx = 1:2
                        MSI.setSupplyChannelVoltageSafe( ...
                            MSI.SUPPLY_CHANNEL_REF, references( ...
                            refIdx, settingIdx, dataIdx))
                        outputProgram = MSI.runProgram(inputProgram);
                        errorCounters(:, refIdx, settingIdx, dataIdx) = ...
                            outputProgram.ErrorCounter(:);
                    end
                    
                    % Check if references are far enough apart for all
                    % cells
                    if errorCounters(colIdx, 1, settingIdx, dataIdx) > ...
                            511 || errorCounters(colIdx, 2, settingIdx, ...
                            dataIdx) < 512
                        warning("Worst cell for data = %d can not be" + ...
                            " characterised with delay %d with the " + ...
                            "current characterisation limits!", data, ...
                            readT1)
                        continue
                    end

                    % Binary search
                    minRef = references(1, settingIdx, dataIdx);
                    maxRef = references(2, settingIdx, dataIdx);

                    for refIdx = 3:maxNumberOfReferences
                        newRef = (minRef + maxRef) / 2;
                        references(refIdx, settingIdx, dataIdx) = newRef;
                        MSI.setSupplyChannelVoltageSafe( ...
                            MSI.SUPPLY_CHANNEL_REF, newRef);
                        outputProgram = MSI.runProgram(inputProgram);
                        errorCounters(:, refIdx, settingIdx, dataIdx) = ...
                            outputProgram.ErrorCounter(:);
                        cellErrorCount = errorCounters(colIdx, refIdx, ...
                            settingIdx, dataIdx);
                        if cellErrorCount < 512
                            minRef = newRef;
                        else
                            maxRef = newRef;
                        end
                    end
                    fprintf("%c", 128)
                end
                fprintf("\n")
            end
            obj.worstCellsReadTimeCharacterisationReferences = references;
            obj.worstCellsReadTimeCharacterisationErrorCounters = ...
                errorCounters;

            fprintf("Worst cell characterisation measurements done\n")
        end

        function processReadTimeWorstCellsData(obj)
            %PROCESSREADTIMEWORSTCELLSDATA Compute the RBL voltages from
            %the worst cell read time sweep characterisation data
            references = obj.worstCellsReadTimeCharacterisationReferences;
            errorCounters = ...
                obj.worstCellsReadTimeCharacterisationErrorCounters;
            % Calculate the RBL voltage of each cell/data combination
            for dataIdx = 1:2
                colIdx = obj.fullMemoryReadTimeCharacterisationWorstCells(dataIdx, 3);
                for readTimeIdx = 1:size(references, 2)
                    refVoltages = references(:, readTimeIdx, dataIdx);
                    avgOutputData = errorCounters(colIdx, :, ...
                        readTimeIdx, dataIdx) / 1023;

                    fcn = @(b,x) normcdf(x, b(1), b(2));
                    NRCF = @(b) norm(avgOutputData - fcn(b, ...
                        refVoltages));
                    voltageEstimate = refVoltages(end);
                    B = fminsearch(NRCF, [voltageEstimate; 2e-3]);

                    obj.worstCellsReadTimeCharacterisationRBL(...
                        dataIdx, readTimeIdx) = B(1);
                    obj.worstCellsReadTimeCharacterisationRBLstd(...
                        dataIdx, readTimeIdx) = B(2);
                end
            end

            obj.minimumReadTime = ceil(interp1(diff( ...
                obj.worstCellsReadTimeCharacterisationRBL), ...
                obj.Config.worstCellReadTimeRT1Range, 0, ...
                "linear", "extrap"));
            obj.readTimeReferenceEstimate = mean( ...
                obj.worstCellsReadTimeCharacterisationRBL);

            reference = interp1( ...
                obj.Config.worstCellReadTimeRT1Range, ...
                obj.readTimeReferenceEstimate, readT1, ...
                "linear", "extrap");
            obj.readSettingReference(readSettingIdx) = reference;
        end

        function findWorstRetentionCells(obj, MSI)
            %FINDWORSTRETENTIONCELLS Find worst cells in terms of retention
            %time for specified RT1 settings

            fprintf("Find worst retention cells\n")
            
            % Load program
            inputProgram = Program("rowRetentionFast");
            inputProgram.fromAsmFile(obj.PROGRAM_FOLDER + ...
                "rowRetentionFast.asm")
            
            % Set row/data-independent settings
            inputProgram.Register(4) = obj.AddressSettings;

            readT2 = obj.Config.fullMemoryReadTimeRT2;
            writeT1 = obj.Config.fullMemoryReadTimeWT1;

            obj.firstFailingCellDelay = NaN(32, 2, numel( ...
                obj.Config.retentionTimeRT1Range));
            obj.firstFailingCellErrorCounters = NaN(32, 32, 2, numel( ...
                obj.Config.retentionTimeRT1Range));

            for readSettingIdx = 1:numel(obj.Config.retentionTimeRT1Range)
                readT1 = obj.Config.retentionTimeRT1Range(readSettingIdx);
                fprintf("Read RT1 = %d\n", readT1)
                if readT1 < obj.minimumReadTime
                    warning("Read T1 %d is smaller than minimum %d", ...
                        readT1, obj.minimumReadTime)
                    continue
                end
                reference = obj.readSettingReference(readSettingIdx);
                if isnan(reference)
                    warning("Reference voltage undefined for RT1 = %d", ...
                        readT1)
                    continue
                end
                MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, ...
                    reference);
                for data = 0:1
                    dataIdx = data + 1;
                    for rowIdx = 1:32
                        settingsRegister = obj.createSetting(255, ...
                            readT1, readT2, writeT1, 0, 1, 1 - data);
                        inputProgram.Register(5) = settingsRegister;
                        inputProgram.Register(6) = obj.AddressRows(rowIdx);
                        inputProgram.Register(7) = 0xFFFFFFFF * data;

                        outputProgram = MSI.runProgram(inputProgram);
                        
                        obj.firstFailingCellErrorCounters(:, rowIdx, ...
                            dataIdx, readSettingIdx) = ...
                            outputProgram.ErrorCounter(:);
                        obj.firstFailingCellDelay(rowIdx, dataIdx, ...
                            readSettingIdx) = outputProgram.Register(1);
                        fprintf("%c", 128)
                    end
                    fprintf("\n")
                end
            end
            fprintf("Fast full memory worst retention measurements done\n")
        end

        function processRetentionFullMemoryData(obj)
            %PROCESSRETENTIONFULLMEMORYDATA Identify the worst cells for
            %all read settings
            [~, worstRetentionRowsIdx] = min(obj.firstFailingCellDelay, ...
                [], 1);
            for readSettingIdx = 1:numel(obj.Config.retentionTimeRT1Range)
                if any(isnan( ...
                        obj.firstFailingCellDelay(:,:,readSettingIdx)))
                    continue
                end
                for dataIdx = 1:2
                    worstRowIdx = worstRetentionRowsIdx(1, dataIdx, ...
                        readSettingIdx);
                    obj.fullMemoryRetentionWorstCells(1, dataIdx, ...
                        readSettingIdx) = worstRowIdx;
                    if sum(obj.firstFailingCellErrorCounters(:, ...
                            worstRowIdx, dataIdx, readSettingIdx)) > 1
                        warning("Multiple first failing cells for " + ...
                            "read setting %d/data %d", readSettingIdx, ...
                            dataIdx)
                    end
                    obj.fullMemoryRetentionWorstCells(2, dataIdx, ...
                        readSettingIdx) = find(obj...
                        .firstFailingCellErrorCounters(:, worstRowIdx, ...
                        dataIdx, readSettingIdx) > 0, 1);
                end
            end
        end

        function characteriseHoldTimeWorstCells(obj, MSI)
            %CHARACTERISEHOLDTIMEWORSTCELLS Characterise the identified
            %worst cells for a specified range of hold times

            readT2 = obj.Config.fullMemoryReadTimeRT2;
            writeT1 = obj.Config.fullMemoryReadTimeWT1;

            fprintf("Characterise worst retention cells\n")

            % Measurement settings
            maxNumberOfReferences = 11;

            % Allocate result matrices
            errorCounters = NaN(32, maxNumberOfReferences, 2, ...
                obj.Config.holdTimeSteps, numel(obj.Config.retentionTimeRT1Range));
            references = NaN(maxNumberOfReferences, 2, ...
                obj.Config.holdTimeSteps, numel(obj.Config.retentionTimeRT1Range));

            % Load program
            inputProgram = Program("holdTimeCharacterisation");
            inputProgram.fromAsmFile(obj.PROGRAM_FOLDER + ...
                "holdTimeCharacterisation.asm")
            
            % Set row/data-independent settings
            inputProgram.Register(2) = obj.Config.holdTimeComparisonCount;
            inputProgram.Register(4) = obj.AddressSettings;

            for readSettingIdx = 1:numel(obj.Config.retentionTimeRT1Range)
                readT1 = obj.Config.retentionTimeRT1Range(readSettingIdx);
                fprintf("Read RT1 = %d\n", readT1)
                if any(obj.fullMemoryRetentionWorstCells(:, ...
                        :, readSettingIdx) == 0)
                    continue
                end
                % Determine hold time range
                holdTimeLimits = min(7e-8 + 1e-7 * obj ...
                    .firstFailingCellDelay(obj ...
                    .fullMemoryRetentionWorstCells(1, :, ...
                    readSettingIdx), :, readSettingIdx));
                if diff(holdTimeLimits) == 0
                    continue
                end
                holdTimeRange = linspace(min(holdTimeLimits), ...
                    max(holdTimeLimits), obj.Config.holdTimeSteps);
                holdTimes = max((round(holdTimeRange / 1e-8 - 8) / 2), 0);
                obj.worstCellsRetentionCharacterisationHoldTimes(:, ...
                    readSettingIdx) = holdTimes;
                for data = 0:1
                    dataIdx = data + 1;
                    rowIdx = obj.fullMemoryRetentionWorstCells(1, ...
                        dataIdx, readSettingIdx);
                    colIdx = obj.fullMemoryRetentionWorstCells(2, ...
                        dataIdx, readSettingIdx);
                    settingsRegister = obj.createSetting(255, readT1, ...
                        readT2, writeT1, 0, 1, 1 - data);
                    inputProgram.Register(5) = settingsRegister;
                    inputProgram.Register(6) = obj.AddressRows(rowIdx);
                    inputProgram.Register(7) = 0xFFFFFFFF * data;

                    for holdTimeIdx = 1:numel(holdTimes)
                        inputProgram.Register(8) = holdTimes(holdTimeIdx);
                        
                        % Measure characterisation limits first, then perform
                        % binary search on each cell of the row.
                        references(1, dataIdx, holdTimeIdx, readSettingIdx) = ...
                            obj.Config.characterisationLimits(1);
                        references(2, dataIdx, holdTimeIdx, readSettingIdx) = ...
                            obj.Config.characterisationLimits(2);
    
                        for refIdx = 1:2
                            MSI.setSupplyChannelVoltageSafe( ...
                                MSI.SUPPLY_CHANNEL_REF, references( ...
                                refIdx, dataIdx, holdTimeIdx, readSettingIdx))
                            outputProgram = MSI.runProgram(inputProgram);
                            errorCounters(:, refIdx, dataIdx, holdTimeIdx, readSettingIdx) = ...
                                outputProgram.ErrorCounter(:);
                        end
                        
                        % Check if references are far enough apart for all
                        % cells
                        target = obj.Config.holdTimeComparisonCount / 2;
                        if errorCounters(colIdx, 1, dataIdx, holdTimeIdx, readSettingIdx) > ...
                                target || errorCounters(colIdx, 2, dataIdx, ...
                                holdTimeIdx, readSettingIdx) < target
                            warning("Worst cell for data = %d can " + ...
                                "not be characterised with hold time" + ...
                                " %d with the current " + ...
                                "characterisation limits!", data, ...
                                holdTimes(holdTimeIdx))
                            continue
                        end
    
                        % Binary search
                        minRef = references(1, dataIdx, holdTimeIdx, readSettingIdx);
                        maxRef = references(2, dataIdx, holdTimeIdx, readSettingIdx);
    
                        for refIdx = 3:maxNumberOfReferences
                            newRef = (minRef + maxRef) / 2;
                            references(refIdx, dataIdx, holdTimeIdx, readSettingIdx) = newRef;
                            MSI.setSupplyChannelVoltageSafe( ...
                                MSI.SUPPLY_CHANNEL_REF, newRef);
                            outputProgram = MSI.runProgram(inputProgram);
                            errorCounters(:, refIdx, dataIdx, holdTimeIdx, readSettingIdx) = ...
                                outputProgram.ErrorCounter(:);
                            cellErrorCount = errorCounters(colIdx, refIdx, ...
                                dataIdx, holdTimeIdx, readSettingIdx);
                            if cellErrorCount < target
                                minRef = newRef;
                            else
                                maxRef = newRef;
                            end
                        end

                        fprintf("%c", 128)
                    end
                    fprintf("\n")
                end
            end
            obj.worstCellsRetentionCharacterisationReferences = references;
            obj.worstCellsRetentionCharacterisationErrorCounters = ...
                errorCounters;

            fprintf("Worst retention cell characterisation " + ...
                "measurements done\n")
        end

        function processRetentionWorstCellsData(obj)
            %PROCESSRETENTIONWORSTCELLSDATA Compute the RBL voltages from
            %the worst cell hold time sweep characterisation data
            references = obj.worstCellsRetentionCharacterisationReferences;
            errorCounters = ...
                obj.worstCellsRetentionCharacterisationErrorCounters;
            % Calculate the RBL voltage of each cell/data combination
            for readSettingIdx = 1:numel(obj.Config.retentionTimeRT1Range)
                if any(isnan(references(readSettingIdx, :, :)))
                    continue
                end
                for dataIdx = 1:2
                    colIdx = obj.fullMemoryRetentionWorstCells(2, ...
                            dataIdx, readSettingIdx);
                    for holdTimeIdx = 1:size(references, 3)
                        if any(isnan(references(:,:,holdTimeIdx, readSettingIdx)))
                            continue
                        end
                        refVoltages = references(:, dataIdx, holdTimeIdx, readSettingIdx);
                        avgOutputData = errorCounters(colIdx, :, ...
                            dataIdx, holdTimeIdx, readSettingIdx) / 1023;
    
                        fcn = @(b,x) normcdf(x, b(1), b(2));
                        NRCF = @(b) norm(avgOutputData - fcn(b, ...
                            refVoltages));
                        voltageEstimate = refVoltages(end);
                        B = fminsearch(NRCF, [voltageEstimate; 2e-3]);
    
                        obj.worstCellsRetentionCharacterisationRBL(...
                            dataIdx, holdTimeIdx, readSettingIdx) = B(1);
                        obj.worstCellsRetentionCharacterisationRBLstd(...
                            dataIdx, holdTimeIdx, readSettingIdx) = B(2);
                    end
                end

                % Improve the reference voltage to maximise the retention
                % time
                worstRBLdifference = ...
                    obj.worstCellsRetentionCharacterisationRBL(2, :, ...
                    readSettingIdx) - obj...
                    .worstCellsRetentionCharacterisationRBL(2, 1, ...
                    readSettingIdx) + obj...
                    .worstCellsRetentionCharacterisationRBL(1, end, ...
                    readSettingIdx) - obj...
                    .worstCellsRetentionCharacterisationRBL(1, :, ...
                    readSettingIdx) + rand(1, size(obj...
                    .worstCellsRetentionCharacterisationRBL, 2)) * 1e-9;
                nanmask = ~isnan(worstRBLdifference);
                improvedRef = interp1(worstRBLdifference(nanmask), obj...
                    .worstCellsRetentionCharacterisationRBL(1, nanmask, ...
                    readSettingIdx), 0);
                obj.readSettingReference(readSettingIdx) = improvedRef;
            end
        end

    end

    methods (Static)
        function setting = createSetting(captureDelay, readT1, readT2, ...
                writeT1, saTest, backgroundEnable, backgroundData)
            arguments (Input)
                captureDelay (1,1) uint32 {mustBeInRange(captureDelay, ...
                    0, 255)}
                readT1 (1,1) uint32 {mustBeInRange(readT1, 0, 191)}
                readT2 (1,1) uint32 {mustBeInRange(readT2, 0, 15)}
                writeT1 (1,1) uint32 {mustBeInRange(writeT1, 0, 63)}
                saTest (1,1) logical
                backgroundEnable (1,1) logical
                backgroundData (1,1) logical
            end
            arguments (Output)
                setting (1,1) uint32
            end
            %CREATESETTING Construct a setting number from specified 
            % settings 
            setting = captureDelay + readT1 * 2^8 + readT2 * 2^16 + ...
                writeT1 * 2^20 + saTest * 2^26 + ...
                backgroundEnable * 2^27 + backgroundData * 2^28;
        end
    end
end

