%% Characterise memories
% Using the result of findBestReferenceVoltage.m, we can find the energy
% consumption and latency of each memory/setting combination

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\" + ...
    "memoryCharacterisation\";
workspaceName = sprintf("%scharMems_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Settings
% "outputData" can be the result of a findBestReferenceVoltage.m run or a
% concatenation of data from various runs (different sweep ranges for
% different memories)
% Cols: [memory number, read T1, read T2, write T1, reference, retention]
inputDataTemp = [bestRefOutputData; bestStaticSettingsData];
% Keep only the input configurations with non-zero retention time
inputData = inputDataTemp;%(inputDataTemp(:,6) > 0, :);

names = cell(size(inputData, 1), 1);
for iName = 1:numel(names)
    names{iName} = sprintf("Mem %d (%d, %d, %d)", inputData(iName, 1:4));
end

%% Allocate results matrix
% Cols: [retention, E read, E write, E refresh, P leakage, latency]
memoryCharacterisationOutputData = NaN(size(inputData, 1), 6);
memoryCharacterisationOutputData(:,1) = inputData(:,6);

writeEnergyMeasurementAveragingRuns = 2;
writeReadEnergyMeasurementAveragingRuns = 2;
refreshEnergyMeasurementAveragingRuns = 2;
leakageEnergyMeasurementAveragingRuns = 2;

%% Measure latency
fprintf("Measure latency\n")
for configIdx = 1:size(inputData, 1)
    reference = inputData(configIdx, 5);
    MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, reference)

    inputProgram = Program("findReadLatency");
    inputProgram.fromAsmFile("findReadLatency.asm")

    inputProgram.Register(4) = 64 * inputData(configIdx, 1) + 32;
    inputProgram.Register(5) = DynamicMemoryCharacterisation...
        .createSetting(0, inputData(configIdx, 2), ...
        inputData(configIdx, 3), inputData(configIdx, 4), 0, 0, 0);
    inputProgram.Register(6) = 64 * inputData(configIdx, 1);
    
    outputProgram = MSI.runProgram(inputProgram);
    % SET DELAY CHAIN RESOLUTION BASED ON RO CHARACTERISATION!!!
    memoryCharacterisationOutputData(configIdx, 6) = (double( ...
        outputProgram.Register(7)) + 1) * 17e-12 + 180e-12;

    fprintf("%c", 128)
end
fprintf("\n")

%% Measure write energy
% Measure write energy through random pattern accesses
% (Quite some variation between runs)
% 
% Address switching energy is of all memories corrected to the same supply
% (6x too large for dynamic, 2x too large for static).
% Solution: estimate the address switching energy by running the programs
% on a memory not connected to the measured supply. The memories connected
% to the supply then only consume the address switching energy. This can be
% divided by the number of memories to estimate the address switching
% energy per memory (roughly the same design) and multiplied by number of
% memories - 1 to be subtracted from the measured energy consumption,
% resulting in only a single address switching energy.

fprintf("Measure write energy\n")

% Set up SMU
smuChannel = 1;
smuVoltage = 1.1;
MSI.setSmuVoltageSafe(smuChannel, smuVoltage, 1e-4);

supplyRelayChannel = [3, 3, 3, 3, 3, 3, 1, 2];
controlRelayChannel = 4;

for configIdx = 1:size(inputData, 1)
    reference = inputData(configIdx, 5);
    MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, reference)

    % Measure write power for selected memory/settings
    inputProgram = Program("loopRandomWrite");
    inputProgram.fromAsmFile("loopRandomWrite.asm")

    totalCoreWriteEnergy = 0;
    totalCoreAddressEnergy = 0;
    for averagingRunIdx = 1:writeEnergyMeasurementAveragingRuns
        inputProgram.Register(4) = 64 * inputData(configIdx, 1) + 32;
        inputProgram.Register(5) = DynamicMemoryCharacterisation...
            .createSetting(255, inputData(configIdx, 2), ...
            inputData(configIdx, 3), inputData(configIdx, 4), 0, 0, 0);
        inputProgram.Register(6) = 64 * inputData(configIdx, 1);
        
        MSI.uploadProgram(inputProgram)
    
        [~, coreWritePower, ~] = ...
            test_9_measure_power(MSI, smuChannel, supplyRelayChannel(...
                inputData(configIdx, 1) + 1), controlRelayChannel);
        totalCoreWriteEnergy = totalCoreWriteEnergy + ...
            (coreWritePower / 1e8 * 21);
    
        % Measure address switching power only for selected memory/settings
        if inputData(configIdx, 1) < 6
            otherMemoryNo = 6;
            memoryRelayCount = 6;
        else
            otherMemoryNo = 0;
            memoryRelayCount = 2;
        end
        inputProgram.Register(4) = 64 * otherMemoryNo + 32;
        inputProgram.Register(5) = DynamicMemoryCharacterisation...
            .createSetting(255, inputData(configIdx, 2), ...
            inputData(configIdx, 3), inputData(configIdx, 4), 0, 0, 0);
        inputProgram.Register(6) = 64 * otherMemoryNo;
        
        MSI.uploadProgram(inputProgram)
    
        [~, coreAddressPower, ~] = ...
            test_9_measure_power(MSI, smuChannel, supplyRelayChannel(...
                inputData(configIdx, 1) + 1), controlRelayChannel);
        totalCoreAddressEnergy = totalCoreAddressEnergy + ...
            coreAddressPower / 1e8 * 21 / memoryRelayCount * ...
            (memoryRelayCount - 1);
    end

    memoryCharacterisationOutputData(configIdx, 3) = ...
        (totalCoreWriteEnergy - totalCoreAddressEnergy) ...
        / writeEnergyMeasurementAveragingRuns;

    fprintf("%c", 128)
end
fprintf("\n")

%% Measure (write +) read energy
% Measure (write +) read energy through random pattern accesses
% (Quite some variation between runs)
% 
% Address switching energy is of all memories corrected to the same supply
% (6x too large for dynamic, 2x too large for static).
% Solution: estimate the address switching energy by running the programs
% on a memory not connected to the measured supply. The memories connected
% to the supply then only consume the address switching energy. This can be
% divided by the number of memories to estimate the address switching
% energy per memory (roughly the same design) and multiplied by number of
% memories - 1 to be subtracted from the measured energy consumption,
% resulting in only a single address switching energy.
%
% Measure writes + reads (should result in reading varying data, though
% maybe limited variation at RT due to leakage) and subtract random write
% energy measured before

fprintf("Measure read energy\n")

% Set up SMU
smuChannel = 1;
smuVoltage = 1.1;
MSI.setSmuVoltageSafe(smuChannel, smuVoltage, 1e-4);

supplyRelayChannel = [3, 3, 3, 3, 3, 3, 1, 2];
controlRelayChannel = 4;

for configIdx = 1:size(inputData, 1)
    reference = inputData(configIdx, 5);
    MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, reference)

    % Measure write + read power for selected memory/settings
    inputProgram = Program("loopRandomWriteRead");
    inputProgram.fromAsmFile("loopRandomWriteRead.asm")

    totalCoreWriteReadEnergy = 0;
    totalCoreAddressEnergy = 0;
    for averagingRunIdx = 1:writeReadEnergyMeasurementAveragingRuns
        inputProgram.Register(4) = 64 * inputData(configIdx, 1) + 32;
        inputProgram.Register(5) = DynamicMemoryCharacterisation...
            .createSetting(255, inputData(configIdx, 2), ...
            inputData(configIdx, 3), inputData(configIdx, 4), 0, 0, 0);
        inputProgram.Register(6) = 64 * inputData(configIdx, 1);
        
        MSI.uploadProgram(inputProgram)
    
        [~, coreWriteReadPower, ~] = ...
            test_9_measure_power(MSI, smuChannel, supplyRelayChannel(...
                inputData(configIdx, 1) + 1), controlRelayChannel);
        totalCoreWriteReadEnergy = totalCoreWriteReadEnergy + ...
            (coreWriteReadPower / 1e8 * 31);
    
        % Measure address switching power only for selected memory/settings
        if inputData(configIdx, 1) < 6
            otherMemoryNo = 6;
            memoryRelayCount = 6;
        else
            otherMemoryNo = 0;
            memoryRelayCount = 2;
        end
        inputProgram.Register(4) = 64 * otherMemoryNo + 32;
        inputProgram.Register(5) = DynamicMemoryCharacterisation...
            .createSetting(255, inputData(configIdx, 2), ...
            inputData(configIdx, 3), inputData(configIdx, 4), 0, 0, 0);
        inputProgram.Register(6) = 64 * otherMemoryNo;
        
        MSI.uploadProgram(inputProgram)
    
        [~, coreAddressPower, ~] = ...
            test_9_measure_power(MSI, smuChannel, supplyRelayChannel(...
                inputData(configIdx, 1) + 1), controlRelayChannel);
        totalCoreAddressEnergy = totalCoreAddressEnergy + ...
            coreAddressPower / 1e8 * 31 / memoryRelayCount * ...
            (memoryRelayCount - 1);
    end

    memoryCharacterisationOutputData(configIdx, 2) = ...
        (totalCoreWriteReadEnergy - totalCoreAddressEnergy) ...
        / writeReadEnergyMeasurementAveragingRuns ...
        - memoryCharacterisationOutputData(configIdx, 3);

    fprintf("%c", 128)
end
fprintf("\n")

%% Measure refresh energy (sequential access pattern)
% Measure refresh energy with sequential pattern accesses
% 
% Address switching energy is of all memories corrected to the same supply
% (6x too large for dynamic, 2x too large for static).
% Solution: estimate the address switching energy by running the programs
% on a memory not connected to the measured supply. The memories connected
% to the supply then only consume the address switching energy. This can be
% divided by the number of memories to estimate the address switching
% energy per memory (roughly the same design) and multiplied by number of
% memories - 1 to be subtracted from the measured energy consumption,
% resulting in only a single address switching energy.

fprintf("Measure refresh energy\n")

% Set up SMU
smuChannel = 1;
smuVoltage = 1.1;
MSI.setSmuVoltageSafe(smuChannel, smuVoltage, 1e-4);

supplyRelayChannel = [3, 3, 3, 3, 3, 3, 1, 2];
controlRelayChannel = 4;

for configIdx = 1:size(inputData, 1)
    % Skip for static memories
    if inputData(configIdx, 1) >= 6
        memoryCharacterisationOutputData(configIdx, 4) = 0;
        fprintf("%c", 128)
        continue
    end

    reference = inputData(configIdx, 5);
    MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, reference)

    % Measure refresh power for selected memory/settings
    inputProgram = Program("loopRandomRefresh");
    inputProgram.fromAsmFile("loopRandomRefresh.asm")

    totalCoreRefreshEnergy = 0;
    totalCoreAddressEnergy = 0;
    for averagingRunIdx = 1:refreshEnergyMeasurementAveragingRuns
        inputProgram.Register(4) = 64 * inputData(configIdx, 1) + 32;
        inputProgram.Register(5) = DynamicMemoryCharacterisation...
            .createSetting(255, inputData(configIdx, 2), ...
            inputData(configIdx, 3), inputData(configIdx, 4), 0, 0, 0);
        inputProgram.Register(6) = 64 * inputData(configIdx, 1);
        inputProgram.Register(8) = 64 * inputData(configIdx, 1) + 31;
        
        MSI.uploadProgram(inputProgram)
    
        [~, coreRefreshPower, ~] = ...
            test_9_measure_power(MSI, smuChannel, supplyRelayChannel(...
                inputData(configIdx, 1) + 1), controlRelayChannel);
        totalCoreRefreshEnergy = totalCoreRefreshEnergy + ...
            (coreRefreshPower / 1e8 * 290);
    
        % Measure address switching power only for selected memory/settings
        if inputData(configIdx, 1) < 6
            otherMemoryNo = 6;
            memoryRelayCount = 6;
        else
            otherMemoryNo = 0;
            memoryRelayCount = 2;
        end
        inputProgram.Register(4) = 64 * otherMemoryNo + 32;
        inputProgram.Register(5) = DynamicMemoryCharacterisation...
            .createSetting(255, inputData(configIdx, 2), ...
            inputData(configIdx, 3), inputData(configIdx, 4), 0, 0, 0);
        inputProgram.Register(6) = 64 * otherMemoryNo;
        
        MSI.uploadProgram(inputProgram)
    
        [~, coreAddressPower, ~] = ...
            test_9_measure_power(MSI, smuChannel, supplyRelayChannel(...
                inputData(configIdx, 1) + 1), controlRelayChannel);
        totalCoreAddressEnergy = totalCoreAddressEnergy + ...
            coreAddressPower / 1e8 * 290 / memoryRelayCount * ...
            (memoryRelayCount - 1);
    end

    memoryCharacterisationOutputData(configIdx, 4) = ...
        (totalCoreRefreshEnergy - totalCoreAddressEnergy) ...
        / refreshEnergyMeasurementAveragingRuns;

    fprintf("%c", 128)
end
fprintf("\n")

%% Measure leakage energy
fprintf("Measure leakage energy\n")

% Set up SMU
smuChannel = 1;
smuVoltage = 1.1;
MSI.setSmuVoltageSafe(smuChannel, smuVoltage, 1e-4);

supplyRelayChannel = [3, 3, 3, 3, 3, 3, 1, 2];
controlRelayChannel = 4;

for configIdx = 1:size(inputData, 1)
    reference = inputData(configIdx, 5);
    MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, reference)

    % Measure leakage power for selected memory/settings
    inputProgram = Program("leakage");
    inputProgram.clear;
    MSI.uploadProgram(inputProgram)

    totalCoreLeakagePower = 0;
    for averagingRunIdx = 1:leakageEnergyMeasurementAveragingRuns
        [rawData, ~, ~] = ...
            test_9_measure_power(MSI, smuChannel, supplyRelayChannel(...
                inputData(configIdx, 1) + 1), controlRelayChannel);
        totalCoreLeakagePower = totalCoreLeakagePower + mean(rawData( ...
            1:end/2));
    end

    if inputData(configIdx, 1) < 6
        memoryRelayCount = 6;
    else
        memoryRelayCount = 1;
    end

    memoryCharacterisationOutputData(configIdx, 5) = ...
        totalCoreLeakagePower / refreshEnergyMeasurementAveragingRuns ...
        / memoryRelayCount;

    fprintf("%c", 128)
end
fprintf("\n")

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "inputData", ...
    "names", "memoryCharacterisationOutputData", ...
    "writeEnergyMeasurementAveragingRuns", ...
    "writeReadEnergyMeasurementAveragingRuns", ...
    "refreshEnergyMeasurementAveragingRuns", ...
    "leakageEnergyMeasurementAveragingRuns")