%% Compute energy over application space
% Compute the energy for each point in the application space for each
% memory type

%% Data format
% [Retention time, read energy, write energy, refresh energy, leakage
% power, operation latency]

% Example data
% data = [...
%     300e-9, 500e-15, 150e-15, 32 * 620e-15, 0, 1e-9;
%     1e-6, 450e-15, 200e-15, 32 * 600e-15, 0, 2e-9;
%     Inf, 600e-15, 500e-15, 0, 100e-9, 0.5e-9];
data = memoryCharacterisationOutputData;

colorLimits = [...
    1.0, 1.0, 0, 0.7, 0.1, 0.1;
    1.0, 1.0, 0, 0.7, 0.9, 0.9;
    0.1, 0.1, 1.0, 1.0, 0, 0.7;
    0.9, 0.9, 1.0, 1.0, 0, 0.7;
    0, 0.7, 0.1, 0.1, 1.0, 1.0;
    0, 0.7, 0.9, 0.9, 1.0, 1.0;
    0.4, 0.4, 0.4, 0.4, 0.4, 0.4;
    0.7, 0.7, 0.7, 0.7, 0.7, 0.7
    ];

dataColors = [];
memoryNumbers = inputData(:,1);
for memoryNumber = 0:7
    count = sum(memoryNumbers == memoryNumber);
    if count > 0
        tmp = [linspace(colorLimits(memoryNumber + 1, 1), ...
            colorLimits(memoryNumber + 1, 2), count)', ...
            linspace(colorLimits(memoryNumber + 1, 3), ...
            colorLimits(memoryNumber + 1, 4), count)', ...
            linspace(colorLimits(memoryNumber + 1, 5), ...
            colorLimits(memoryNumber + 1, 6), count)'];
        dataColors = [dataColors; tmp];
    end
end

%% Compute power over application space for each memory type
writePerRead = linspace(0, 1, 101);
readFrequency = logspace(log10(1e3), log10(1e9), 501);

[rF, wPr] = meshgrid(readFrequency, writePerRead);

power = NaN([size(rF), size(data, 1)]);

for iData = 1:size(data,1)
    refreshFreq = 1/data(iData,1);
    readFreq = rF;
    writeFreq = rF .* wPr;
%     powerTemp = readFreq * data(iData, 2) + writeFreq * data(iData, 3) ...
%         + refreshFreq * data(iData, 4) + max(data(iData, 5) - mean(data(1:3, 5)), 0);
    powerTemp = readFreq * data(iData, 2) + writeFreq * data(iData, 3) ...
        + refreshFreq * data(iData, 4) + max(data(iData, 5), 0);
    powerTemp(rF .* (1 + wPr) > 1/data(iData, 6)) = NaN;
    power(:,:,iData) = powerTemp;
end

%% Plot
figure(11)
subplot(1,2,1)
plt = surf(readFrequency, writePerRead, power(:,:,1), 'EdgeColor', 'none', 'FaceColor', dataColors(1, :));
hold on
for iData = 2:size(data,1)
    surf(readFrequency, writePerRead, power(:,:,iData), 'EdgeColor', ...
        'none', 'FaceColor', dataColors(iData, :));
end
hold off
plt.Parent.YDir = 'normal';
plt.Parent.XScale = 'log';
plt.Parent.ZScale = 'log';
plt.Parent.ColorScale = 'log';
xlabel("Read frequency [Hz]")
ylabel("Writes per read")
zlabel("Power [W]")
legend(names, 'Location', 'bestoutside')
title("Power of Memory (RT1, RT2, WT1)")

subplot(1,2,2)
applicationSpace = zeros(size(rF));
minPower = min(power, [], 3);
for iData = 1:size(data,1)
    applicationSpace(power(:,:,iData) == minPower) = iData;
end
img = imagesc(readFrequency, writePerRead, applicationSpace);
img.Parent.YDir = 'normal';
img.Parent.XScale = 'log';
xlabel("Read frequency [Hz]")
xlim([min(readFrequency), max(readFrequency)])
ylabel("Writes per read")
ylim([0, 1])
colormap(img.Parent, [1, 1, 1; dataColors])
clim([0, size(data,1)])
title("Lowest power consumption")