%% Find best reference voltage
% Find the reference voltage that gives the maximum retention time for a
% given memory and setting

% Perform binary search with the reference voltage to minimise the
% difference between the retention times of the two data polarities.

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\" + ...
    "memoryCharacterisation\";
workspaceName = sprintf("%sfindBestRef_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Settings
memoryNo = [0, 2, 4, 5];
memoryNo = [0:5];
rowNo = 0:31;
readT1List = [...
    40:2:60;...
    30:2:50;...
    55:2:75;...
    40:2:60];
readT1List = [30:10:70;30:10:70;30:10:70;30:10:70;30:10:70;30:10:70];
readT2 = [10, 4, 15, 15];
readT2 = [10, 10, 4, 4, 15, 15];
writeT1 = [10, 15, 30, 30];
writeT1 = [10, 10, 15, 15, 30, 30];
referenceLimits = [...
    0, 0.6;
    0, 0.6;
    0.6, 1.1;
    0.6, 1.1];
referenceLimits = [0, 0.6; 0, 0.6; 0, 0.6; 0, 0.6; 0.6, 1.1; 0.6, 1.1];
numberOfReferences = max(ceil(log2(diff(referenceLimits, 1, 2) * 1e3)) - 1);
numRuns = 1;

% Allocate intermediate/final result matrices
references = NaN(numberOfReferences, size(readT1List,2), numel(memoryNo));
worstRetention = NaN(numel(rowNo), 2, numberOfReferences, ...
    size(readT1List, 2), numel(memoryNo));
errorCounters = NaN(32, numel(rowNo), 2, numberOfReferences, ...
    size(readT1List, 2), numel(memoryNo));
bestReference = NaN(size(readT1List, 2), numel(memoryNo));
retentionTime = NaN(size(readT1List, 2), numel(memoryNo));
worstCell = NaN(2, 2, size(readT1List, 2), numel(memoryNo));

% Output data: used in the next steps
% Cols: memory number, read T1, read T2, write T1, reference, retention
bestRefOutputData = NaN(numel(retentionTime), 6);

%% Main
for memoryIdx = 1:numel(memoryNo)
    fprintf("Memory %d/%d (= %d)\n", memoryIdx, numel(memoryNo), ...
        memoryNo(memoryIdx))
    for readT1Idx = 1:size(readT1List, 2)
        fprintf("Temp = %f K\n", MSI.getTemperature());
        readT1 = readT1List(memoryIdx, readT1Idx);
        fprintf("Read T1 %d/%d (= %d)\n", readT1Idx, ...
            size(readT1List, 2), readT1)

        % Load program
        inputProgram = Program("rowRetentionFast");
        inputProgram.fromAsmFile("rowRetentionFast.asm")
        % Set row/data-independent settings
        inputProgram.Register(4) = 64 * memoryNo(memoryIdx) + 32;

        minRef = referenceLimits(memoryIdx, 1);
        maxRef = referenceLimits(memoryIdx, 2);
        
        for refIdx = 1:numberOfReferences
            newRef = round(mean([minRef, maxRef]), 3);
            references(refIdx, readT1Idx, memoryIdx) = newRef;
    
            MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, ...
                references(refIdx, readT1Idx, memoryIdx))
            for rowIdx = 1:numel(rowNo)
                inputProgram.Register(6) = 64 * memoryNo(memoryIdx) ...
                    + rowNo(rowIdx);
                for data = 0:1
                    dataIdx = data + 1;
                    settingsRegister = DynamicMemoryCharacterisation...
                        .createSetting(255, readT1, readT2(memoryIdx), ...
                        writeT1(memoryIdx), 0, 1, 1 - data);
                    inputProgram.Register(5) = settingsRegister;
                    inputProgram.Register(7) = 0xFFFFFFFF * data;

                    worstRetentionTemp = NaN(numRuns, 1);
                    errorCountersTemp = NaN(32, numRuns);
                    for i = 1:numRuns
                        outputProgram = MSI.runProgram(inputProgram, 1);
                        worstRetentionTemp(i) = outputProgram.Register(1);
                        errorCountersTemp(:,i) = outputProgram...
                            .ErrorCounter(:);
                    end

                    worstRetention(rowIdx, dataIdx, refIdx, readT1Idx, ...
                        memoryIdx) = min(worstRetentionTemp);
                    errorCounters(:, rowIdx, dataIdx, refIdx, readT1Idx,...
                        memoryIdx) = sum(errorCountersTemp(:, ...
                        worstRetentionTemp == min(worstRetentionTemp)), 2);
                end
                %fprintf("%c", 128)
            end
            
            smallestRetention = min(worstRetention(:, :, refIdx, ...
                readT1Idx, memoryIdx), [], 1);
            if smallestRetention(1) > smallestRetention(2)
                minRef = newRef;
            else
                maxRef = newRef;
            end

            fprintf("%c", 128)
            %fprintf("\n")
        end
        fprintf("\n")

        % Process results
        minRetentions = reshape(min(worstRetention(:, :, :, readT1Idx, ...
            memoryIdx), [], [1,2]), 1, []);
        bestMinRetention = max(minRetentions);
        bestReferenceIdx = find(minRetentions == bestMinRetention, 1, ...
            'last');
        bestReference(readT1Idx, memoryIdx) = references( ...
            bestReferenceIdx, readT1Idx, memoryIdx);
        retentionTime(readT1Idx, memoryIdx) = bestMinRetention;
        fprintf("Best reference = %f V, retention = %d (%e to %e s)\n", ...
            bestReference(readT1Idx, memoryIdx), bestMinRetention, ...
            7e-8+1e-7*(bestMinRetention-1), 7e-8+1e-7*bestMinRetention)

        for dataIdx = 1:2
            [~, worstCellRowIdx] = min(worstRetention(:, dataIdx, ...
                bestReferenceIdx, readT1Idx, memoryIdx));
            worstCell(1, dataIdx, readT1Idx, memoryIdx) = worstCellRowIdx;
            [~, worstCell(2, dataIdx, readT1Idx, memoryIdx)] = max( ...
                errorCounters(:, worstCellRowIdx, dataIdx, ...
                bestReferenceIdx, readT1Idx, memoryIdx));
            fprintf("Worst data = %d cell: (%d, %d) [(row, col) " + ...
                "in 1-32]\n", dataIdx - 1, worstCell(:, dataIdx, ...
                readT1Idx, memoryIdx));
        end

        outputDataIdx = (memoryIdx-1) * size(readT1List, 2) + readT1Idx;
        bestRefOutputData(outputDataIdx, 1) = memoryNo(memoryIdx);
        bestRefOutputData(outputDataIdx, 2) = readT1;
        bestRefOutputData(outputDataIdx, 3) = readT2(memoryIdx);
        bestRefOutputData(outputDataIdx, 4) = writeT1(memoryIdx);
        bestRefOutputData(outputDataIdx, 5) = bestReference(readT1Idx, ...
            memoryIdx);
        bestRefOutputData(outputDataIdx, 6) = max(7e-8 + 1e-7 * ...
            (retentionTime(readT1Idx, memoryIdx) - 1), 0) ;
    end
end

%% Plot ideal reference vs read setting
for memoryIdx = 1:numel(memoryNo)
    figure(memoryIdx)
    plot(readT1List(memoryIdx, :), bestReference(:, memoryIdx), 'o-')
    xlabel("Read T1 setting")
    xlim([min(readT1List(memoryIdx, :)) - 5, ...
        max(readT1List(memoryIdx, :)) + 5])
    ylabel("Reference voltage [V]")
    yyaxis right
    ylabel("Retention time [s]")
    errorbar(readT1List(memoryIdx, :), 7e-8 + 1e-7 * ...
        (retentionTime(:, memoryIdx) - 0.5), min(5e-8, 7e-8 + 1e-7 * ...
        (retentionTime(:, memoryIdx) - 0.5)), 5e-8, 'o-')
    grid on
end

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "memoryNo", ...
    "rowNo", "readT1List", "readT2", "writeT1", "referenceLimits", ...
    "numberOfReferences", "numRuns", "references", "worstRetention", ...
    "errorCounters", "bestReference", "retentionTime", "worstCell", ...
    "bestRefOutputData")