%% Find best static settings
% Find the minimum write and read settings for the SRAMs
% - With large enough read time: decrease write time until a single cell 
% can not be flipped anymore
% - With minimum write time: decrease read time until a single cell can not
% be read anymore
%   - Take SA hysteresis into account: write data 0, read twice, write data
%   1, read twice (1->0, 0->0, 0->1, 1->1) for all rows before reducing the
%   read time setting

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\" + ...
    "memoryCharacterisation\";
workspaceName = sprintf("%sfindBestStaticSet_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Settings
memoryNo = [6, 7];
readT2Range = 0:15;

minimumWriteSetting = NaN(size(memoryNo));
minimumReadSetting = NaN(numel(readT2Range), numel(memoryNo));
bestStaticSettingsData = NaN(numel(memoryNo), 6);

%% Main
for memoryIdx = 1:numel(memoryNo)
    fprintf("Memory %d/%d (= %d)\n", memoryIdx, numel(memoryNo), ...
        memoryNo(memoryIdx))

    % Load program
    inputProgram = Program("findStaticWriteTime");
    inputProgram.fromAsmFile("findStaticWriteTime.asm")
    
    inputProgram.Register(4) = 64 * memoryNo(memoryIdx) + 32;
    inputProgram.Register(5) = bin2dec("0"+"0000"+"0"+"000000"+"1000"+ ...
        "01000000"+"11111111");
    inputProgram.Register(6) = 64 * memoryNo(memoryIdx);
    inputProgram.Register(7) = 64 * memoryNo(memoryIdx) + 31;

    outputProgram = MSI.runProgram(inputProgram);

    minimumWriteSetting(memoryIdx) = min(outputProgram.Register(1)+1, 63);

    % Load program
    inputProgram = Program("findStaticReadTime");
    inputProgram.fromAsmFile("findStaticReadTime.asm")
    
    for readT2Idx = 1:numel(readT2Range)
        inputProgram.Register(4) = 64 * memoryNo(memoryIdx) + 32;
        inputProgram.Register(5) = bin2dec("0"+"0000"+"0"+ ...
            dec2bin(minimumWriteSetting(memoryIdx), 6)+ ...
            dec2bin(readT2Range(readT2Idx), 4)+"00000000"+"11111111");
        inputProgram.Register(6) = 64 * memoryNo(memoryIdx);
        inputProgram.Register(7) = 64 * memoryNo(memoryIdx) + 31;
    
        outputProgram = MSI.runProgram(inputProgram);
    
        minimumReadSetting(readT2Idx, memoryIdx) = ...
            outputProgram.Register(1) + 1;
    end
end

%% Construct output data
% Cols: [memory number, read T1, read T2, write T1, reference, retention]
bestStaticSettingsData(:,1) = memoryNo;
[~,fastestReadIdx] = min(minimumReadSetting, [], 1);
for memoryIdx = 1:numel(memoryNo)
    bestStaticSettingsData(memoryIdx, 2) = minimumReadSetting( ...
        fastestReadIdx(memoryIdx), memoryIdx);
    bestStaticSettingsData(memoryIdx,3) = readT2Range( ...
        fastestReadIdx(memoryIdx));
end
bestStaticSettingsData(:,4) = minimumWriteSetting;
bestStaticSettingsData(:,5) = 1.1;
bestStaticSettingsData(:,6) = Inf(numel(memoryNo), 1);


%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "memoryNo", ...
    "readT2Range", "bestStaticSettingsData")