%% Full memory characterisation
% This script runs the full characterisation of all selected memories which
% includes the following steps for each memory:
% - Select constants depending on the memory, allocate container
%   (struct/class) for the measurement results of the memory.
% - Characterise all cells for a reasonable read time (RT1, manually 
%   specified per memory)
% - Select cells to characterise in detail
% - Perform detailed characterisation of the selected cells
% - Compute reference estimate for each read time setting (and minimum read
%   time setting needed to generate RBL margin)
% - For every read time setting:
%   - Find the retention time of each cell with the reference estimate
%   - Select cells to characterise in detail
%   - Perform detailed characterisation of the selected cells
%   - Compute the optimal reference for the selected cells to maximise the
%     retention time
%   - (Optional: repeat to see if the selected cells are indeed the ones
%     that limit the retention time)
%   - Measure the power consumption to calculate the power consumption in
%     all applications
%   - Measure the total latency (capture delay) to limit the frequency at
%     which it can be used

%% Connect to measurement setup (if necessary)
if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end

%% Memory settings
memoryConfig(1) = struct( ...
    "memoryNumber", 0, ...
    "fullMemoryReadTimeRT1", 60, ...
    "worstCellReadTimeRT1Range", 30:5:120, ...
    "fullMemoryReadTimeRT2", 10, ...
    "fullMemoryReadTimeWT1", 10, ...
    "characterisationLimits", [0, 0.6], ...
    "retentionTimeRT1Range", 50:10:120, ...
    "holdTimeSteps", 11, ...
    "holdTimeComparisonCount", 1023);

memoryIdx = 1;

%% Create memory characterisation class instance
mem = DynamicMemoryCharacterisation(memoryConfig(memoryIdx));

%% Characterise all cells for read T1 = fullMemoryReadTime
mem.characteriseReadTimeFullMemory(MSI);
mem.processReadTimeFullMemoryData();

%% Characterise worst cell for read T1 setting range
mem.characteriseReadTimeWorstCells(MSI);
mem.processReadTimeWorstCellsData();

%% Retention time test for range of read T1 settings to find worst cells
mem.findWorstRetentionCells(MSI);
mem.processRetentionFullMemoryData();

%% Characterise worst retention cells for hold time range
mem.characteriseHoldTimeWorstCells(MSI);
mem.processRetentionWorstCellsData();