%% Find single cell retention vs temperature
% Find the retention time of a single cell as a function of the
% temperature. Dip slowly during running the main loop to get a rough
% temperature sweep.

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\" + ...
    "memoryCharacterisation\";
workspaceName = sprintf("%ssingleCellTemperatureRetention_%s", ...
    testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Settings
% Selected cell: [memory (0-5), row (0-31), column (0-31)]
cellNo = [4, 4, 3];
dataNo = [0,1];
% Memory settings
% Reference voltage
memoryReferenceVoltages = [0.127, 0.187, 0.204, 0.324, 0.970, 0.931];
% Settings register values
settingsRegister = [ ...
    bin2dec("000"+"110"+"001010"+"1010"+"01111000"+"11111111"), ...
    bin2dec("000"+"110"+"001010"+"1010"+"01100100"+"11111111"), ...
    bin2dec("000"+"110"+"001111"+"0100"+"00101101"+"11111111"), ...
    bin2dec("000"+"110"+"001111"+"0100"+"00101101"+"11111111"), ...
    bin2dec("000"+"110"+"011110"+"1111"+"10110100"+"11111111"), ...
    bin2dec("000"+"110"+"011110"+"1111"+"10110100"+"11111111")];
settingsRegisterReset = [...
    bin2dec("000"+"111"+"011110"+"1111"+"10000000"+"11111111"), ...
    bin2dec("000"+"011"+"011110"+"1111"+"10000000"+"11111111")];
clear testReferenceVoltage
% Test ref voltages for reset to 1 (for data = 0)
testReferenceVoltage(:,1) = max(memoryReferenceVoltages - 0.15, 0);
% Test ref voltages for reset to 0 (for data = 1)
testReferenceVoltage(:,2) = min(memoryReferenceVoltages + 0.15, 1.1);

% Allocate output variables
maxNumberOfRuns = 500;
temperature = NaN(1, numel(dataNo) * maxNumberOfRuns + 1);
retention = NaN(numel(dataNo), maxNumberOfRuns);

%% Main loop

% Load program
inputProgram = Program("cellRetentionFastRst");
inputProgram.fromAsmFile("cellRetentionFastRst.asm");
inputProgram.Register(3) = 31 - cellNo(3);
inputProgram.Register(4) = cellNo(1) * 64 + 32;
inputProgram.Register(5) = settingsRegister(cellNo(1) + 1);
inputProgram.Register(6) = cellNo(1) * 64 + cellNo(2);

MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, ...
    memoryReferenceVoltages(cellNo(1) + 1))

temperature(1) = MSI.getTemperature();

for runIdx = 1:maxNumberOfRuns
    fprintf("Run %d/%d @ %.3f K\n", runIdx, maxNumberOfRuns, ...
        temperature(numel(dataNo) * (runIdx - 1) + 1))
    % Run retention time characterisation
    for dataIdx = 1:numel(dataNo)
        data = dataNo(dataIdx);
        MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_TREF, ...
            testReferenceVoltage(cellNo(1) + 1, dataIdx))

        inputProgram.Register(5) = bitset(inputProgram.Register(5), 29, ...
            1 - data);
        inputProgram.Register(7) = 0xFFFFFFFF * data;
        inputProgram.Register(9) = settingsRegisterReset(dataIdx);

        outputProgram = MSI.runProgram(inputProgram, 1);
        retention(dataIdx, runIdx) = outputProgram.Register(1);
        % Capture temperature again
        temperature(numel(dataNo) * runIdx + dataIdx) = ...
            MSI.getTemperature();
    end
end

%% Plot data
figure(130)
% subplot(2,1,1)
legendText = string.empty(numel(dataNo), 0);
for dataIdx = 1:numel(dataNo)
    x1 = temperature(dataIdx:numel(dataNo):(end-numel(dataNo)));
    x2 = temperature((dataIdx+1):numel(dataNo):(end-numel(dataNo)+1));
    xm = mean([x1; x2]);
    y1 = max((retention(dataIdx,:) - 1) * 360e-9 + 300e-9, 0);
    y2 = (retention(dataIdx,:)) * 360e-9 + 300e-9;
    ym = y2;
%     errorbar(xm, ym, ym - min(y1, y2), max(y1, y2) - ym, ...
%         xm - min(x1, x2), max(x1, x2) - xm, 'k', "LineStyle", "none")
    scatter(xm, ym, "k.")
    hold on
    legendText(dataIdx) = sprintf("Data = %d", dataNo(dataIdx));
end
hold off
set(gca, "YScale", "log")
xlabel("Temperature [K]")
ylabel("Retention time [s]")
title(sprintf("Cell (%d, %d, %d)", ...
    cellNo))
legend(legendText, "Location", "best", "AutoUpdate", "off")

% Arrhenius plot
% subplot(2,1,2)
% for dataIdx = 1:numel(dataNo)
%     x1 = 1./temperature(dataIdx:numel(dataNo):(end-numel(dataNo)));
%     x2 = 1./temperature((dataIdx+1):numel(dataNo):(end-numel(dataNo)+1));
%     xm = 1./mean([1./x1; 1./x2]);
%     y1 = log(1./max((retention(dataIdx,:) - 1) * 360e-9 + 300e-9, 0));
%     y2 = log(1./(retention(dataIdx,:) * 360e-9 + 300e-9));
%     ym = y2;
%     errorbar(xm, ym, ym - min(y1, y2), max(y1, y2) - ym, ...
%         xm - min(x1, x2), max(x1, x2) - xm, 'k', "LineStyle", "none")
%     hold on
% end
% hold off
% xlabel("1/Temperature [1/K]")
% ylabel("ln(1/Retention time [s])")
% title(sprintf("Arrhenius plot: Cell (%d, %d, %d)", cellNo))
% legend(legendText, "Location", "best", "AutoUpdate", "off")

% Calculate activation energies
% Temperature [K] ranges which separate the very low/low/high activation 
% energy regions
% veryLowEaTemp = [5, 20];
% veryLowEaFit = NaN(numel(dataNo), 2);
lowEaTemp = [80, 150];
lowEaFit = NaN(numel(dataNo), 2);
highEaTemp = [190, 270];
highEaFit = NaN(numel(dataNo), 2);

kb = 8.617333262e-5; % Boltzmann constant in [eV/K]

% subplot(2,1,2)
% hold on
for dataIdx = 1:numel(dataNo)
    x1 = 1./temperature(dataIdx:numel(dataNo):(end-numel(dataNo)));
    x2 = 1./temperature((dataIdx+1):numel(dataNo):(end-numel(dataNo)+1));
    xm = 1./mean([1./x1; 1./x2]);
    y1 = log(1./max((retention(dataIdx,:) - 1) * 360e-9 + 300e-9, 0));
    y2 = log(1./(retention(dataIdx,:) * 360e-9 + 300e-9));
    ym = y2;

    x = xm(1./xm > highEaTemp(1) & 1./xm < highEaTemp(2));
    y = ym(1./xm > highEaTemp(1) & 1./xm < highEaTemp(2));
    highEaFit(dataIdx, :) = polyfit(x, y, 1);
%     plot(1./highEaTemp, highEaFit(dataIdx, :) * ...
%         [1./highEaTemp; ones(size(highEaTemp))], 'r--', 'LineWidth', 1.5)

    xm = xm(500:end);
    ym = ym(500:end);
    x = xm(1./xm > lowEaTemp(1) & 1./xm < lowEaTemp(2));
    y = ym(1./xm > lowEaTemp(1) & 1./xm < lowEaTemp(2));
    lowEaFit(dataIdx, :) = polyfit(x, y, 1);
%     plot(1./lowEaTemp, lowEaFit(dataIdx, :) * ...
%         [1./lowEaTemp; ones(size(lowEaTemp))], 'r--', 'LineWidth', 1.5)

%     x = xm(1./xm > veryLowEaTemp(1) & 1./xm < veryLowEaTemp(2));
%     y = ym(1./xm > veryLowEaTemp(1) & 1./xm < veryLowEaTemp(2));
%     veryLowEaFit(dataIdx, :) = polyfit(x, y, 1);
%     plot(1./veryLowEaTemp, veryLowEaFit(dataIdx, :) * ...
%         [1./veryLowEaTemp; ones(size(veryLowEaTemp))], 'r')

    
%     [~, idx] = min(abs(t - mean(highEaTemp)));
%     text(1/t(idx), log(1/retentionFit(idx))+1, sprintf("E_a = %.3f eV", ...
%         -highEaFit(dataIdx, 1) * kb))
% 
%     [~, idx] = min(abs(t - mean(lowEaTemp)));
%     text(1/t(idx), log(1/retentionFit(idx))+0.7, sprintf("E_a = %.4f eV", ...
%         -lowEaFit(dataIdx, 1) * kb))
end
% hold off
% xlim([0.003, 0.02])
% grid on

% subplot(2,1,1)
hold on
for dataIdx = 1:numel(dataNo)
    t = linspace(min(temperature), max(temperature), 100);
%     retentionFit = 1./(...
%         exp(veryLowEaFit(dataIdx,:) * [1./t; ones(size(t))]) + ...
%         exp(lowEaFit(dataIdx,:) * [1./t; ones(size(t))]) + ...
%         exp(highEaFit(dataIdx,:) * [1./t; ones(size(t))]));
    retentionFit = 1./(...
        exp(lowEaFit(dataIdx,:) * [1./t; ones(size(t))]) + ...
        exp(highEaFit(dataIdx,:) * [1./t; ones(size(t))]));
    plot(t, retentionFit, 'r--', 'LineWidth', 1.5)
    [~, idx] = min(abs(t - mean(highEaTemp)));
    text(t(idx), retentionFit(idx) * 2, sprintf("E_a = %.3f eV", ...
        -highEaFit(dataIdx, 1) * kb))

    [~, idx] = min(abs(t - mean(lowEaTemp)));
    text(t(idx), retentionFit(idx) * 2.5, sprintf("E_a = %.4f eV", ...
        -lowEaFit(dataIdx, 1) * kb))
end
hold off
ylim([1e-7, 1e0])
grid on

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "cellNo", ...
    "memoryReferenceVoltages", "settingsRegister", "dataNo", ...
    "settingsRegisterReset", "maxNumberOfRuns", "temperature", "retention")