%% Test 11: Characterise SNM Source Follower
% This script is used to characterise the source follower of the SNM
% measurement structure.

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\tests\" + ...
    "11_snm_source_follower\";
workspaceName = sprintf("%stest11_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Set SNM array settings and sweep SNM input
% Set the SNM array to directly connect the SNM input the input of the
% source follower and set up the SMU for readout.

sfBiasCurrent = 20e-6;

snmSettingsRegisterAddress = 64 * 9 + 32;
snmSettingsRegisterData = bin2dec("00000000000000000000" + "1" + "00" + ...
    "0000" + "00000");

% Set SNM settings register
MSI.memoryWrite(snmSettingsRegisterData, snmSettingsRegisterAddress)
% (Optional: turn off clock)

% Set up output SMU
outputSmuChannel = 1;
MSI.setSmuCurrentSafe(outputSmuChannel, sfBiasCurrent, 2.5)

% Sweep input voltage
sfInputVoltages = linspace(0, 1.1, 56);
sfOutputVoltages = zeros(size(sfInputVoltages));

for inputIdx = 1:numel(sfInputVoltages)
    MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_SNM, ...
        sfInputVoltages(inputIdx))
    sfOutputVoltages(inputIdx) = MSI.getSmuVoltage(outputSmuChannel);
end

%% Plot source follower transfer
% Plot the transfer of the source follower.

% Linear fit
linearFit = polyfit(sfInputVoltages, sfOutputVoltages, 1);

% Plot transfer function together with linear fit
figure(110)
plot(sfInputVoltages, sfOutputVoltages)
hold on
plot(sfInputVoltages, polyval(linearFit, sfInputVoltages))
hold off
xlim([0, 1.1])
ylim([0, 2.5])
xlabel("Input [V]")
ylabel("Output [V]")
title("SNM source follower transfer")
legend(["Data", sprintf("Linear Fit: %f + %f * V_i", linearFit(2), ...
    linearFit(1))], "Location", "best")

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "sfBiasCurrent", ...
    "sfInputVoltages", "sfOutputVoltages")