%% Test 12: Measure Cell SNM
% This script is used to measure the transfer curves of various cells from
% which the SNM can be derived.

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\tests\" + ...
    "12_snm_transfer\";
workspaceName = sprintf("%stest12_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Set SNM array settings and sweep SNM input
% Set the SNM array to directly connect the SNM input the input of the
% source follower and set up the SMU for readout.

% List of cells to measure (0-255 = SVT, 256-511 = LVT)
cellNo = [0:7, 504:511];

% Set up output SMU (20 uA, max 2.7 V)
sfBiasCurrent = 20e-6;
outputSmuChannel = 1;
MSI.setSmuCurrentSafe(outputSmuChannel, sfBiasCurrent, 2.7)

% Sweep input voltage
inputVoltages = linspace(0, 1.1, 56);

% Allocate result matrices
wl0Output = zeros(numel(cellNo), numel(inputVoltages));
wl1bl0Output = zeros(numel(cellNo), numel(inputVoltages));
wl1bl1Output = zeros(numel(cellNo), numel(inputVoltages));

% Loop over cells
for cellIdx = 1:numel(cellNo)
    fprintf("Cell %d/%d (= %d)\n", cellIdx, numel(cellNo), cellNo(cellIdx))
    % ----- WL = 0 -----
    % (Optional: turn on clock)
    snmSettingsRegisterAddress = 64 * 9 + 32;
    snmSettingsRegisterData = cellNo(cellIdx);
    
    % Set SNM settings register
    MSI.memoryWrite(snmSettingsRegisterData, snmSettingsRegisterAddress)
    % (Optional: turn off clock)
    
    for inputIdx = 1:numel(inputVoltages)
        MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_SNM, ...
            inputVoltages(inputIdx))
        wl0Output(cellIdx, inputIdx) = MSI.getSmuVoltage(outputSmuChannel);
    end
    
    % ----- WL = 1, BL = 0 -----
    % (Optional: turn on clock)
    snmSettingsRegisterData = snmSettingsRegisterData + 512;
    
    % Set SNM settings register
    MSI.memoryWrite(snmSettingsRegisterData, snmSettingsRegisterAddress)
    % (Optional: turn off clock)
    
    for inputIdx = 1:numel(inputVoltages)
        MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_SNM, ...
            inputVoltages(inputIdx))
        wl1bl0Output(cellIdx, inputIdx) = ...
            MSI.getSmuVoltage(outputSmuChannel);
    end
    
    % ----- WL = 1, BL = 1 -----
    % (Optional: turn on clock)
    snmSettingsRegisterData = snmSettingsRegisterData + 1024;
    
    % Set SNM settings register
    MSI.memoryWrite(snmSettingsRegisterData, snmSettingsRegisterAddress)
    % (Optional: turn off clock)
    
    for inputIdx = 1:numel(inputVoltages)
        MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_SNM, ...
            inputVoltages(inputIdx))
        wl1bl1Output(cellIdx, inputIdx) = ...
            MSI.getSmuVoltage(outputSmuChannel);
    end
end

%% Process measurement data
% Reverse the source follower transfer to find the cell output voltage and
% plot the transfer curves.

% Find cell transfer by mapping source follower output to input
wl0Transfer = interp1(sfOutputVoltages, sfInputVoltages, wl0Output, "linear", "extrap");
wl1bl0Transfer = interp1(sfOutputVoltages, sfInputVoltages, wl1bl0Output, "linear", "extrap");
wl1bl1Transfer = interp1(sfOutputVoltages, sfInputVoltages, wl1bl1Output, "linear", "extrap");

% Compute distance to diagonal
xDiag = linspace(0,1.1,111);
wl0YDiag = zeros(numel(cellNo), numel(xDiag));
wl1bl0YDiag = zeros(numel(cellNo), numel(xDiag));
wl1bl1YDiag = zeros(numel(cellNo), numel(xDiag));
for cellIdx = 1:numel(cellNo)
    for xInd = 1:numel(xDiag)
        x = xDiag(xInd);
        wl0YDiag(cellIdx, xInd) = (interp1(wl0Transfer(cellIdx,:) ...
            - (1.1 - 2 * x + inputVoltages), inputVoltages, 0) - x) ...
            * sqrt(2);
        wl1bl0YDiag(cellIdx, xInd) = (interp1(wl1bl0Transfer(cellIdx,:) ...
            - (1.1 - 2 * x + inputVoltages), inputVoltages, 0) - x) ...
            * sqrt(2);
        wl1bl1YDiag(cellIdx, xInd) = (interp1(wl1bl1Transfer(cellIdx,:) ...
            - (1.1 - 2 * x + inputVoltages), inputVoltages, 0) - x) ...
            * sqrt(2);
    end
end

% Plot SNM figures (SVT)
figure(121)

% Hold SNM
subplot(1,3,1)
plot(inputVoltages, wl0Transfer(cellNo < 256,:))
hold on
plot(wl0Transfer(cellNo < 256,:), inputVoltages)
hold off
daspect([1,1,1])
xlim([0, 1.1])
ylim([0, 1.1])
holdSnmSVT = min(max(max(wl0YDiag(cellNo < 256,:)) - fliplr(min(...
    wl0YDiag(cellNo < 256,:)))), max(-(max(wl0YDiag(cellNo < 256,:)) - ...
    fliplr(min(wl0YDiag(cellNo < 256,:))))));
title(sprintf("Hold SNM = %.3f V", holdSnmSVT))

% Read SNM
subplot(1,3,2)
plot(inputVoltages, wl1bl1Transfer(cellNo < 256,:))
hold on
plot(wl1bl1Transfer(cellNo < 256,:), inputVoltages)
hold off
daspect([1,1,1])
xlim([0, 1.1])
ylim([0, 1.1])
readSnmSVT = min(max(max(wl1bl1YDiag(cellNo < 256,:)) - fliplr(min(...
    wl1bl1YDiag(cellNo < 256,:)))), max(-(max(wl1bl1YDiag(cellNo < 256,...
    :)) - fliplr(min(wl1bl1YDiag(cellNo < 256,:))))));
title(sprintf("Read SNM = %.3f V", readSnmSVT))

% Write SNM
subplot(1,3,3)
plot(inputVoltages, wl1bl0Transfer(cellNo < 256,:))
hold on
plot(wl1bl1Transfer(cellNo < 256,:), inputVoltages)
hold off
daspect([1,1,1])
xlim([0, 1.1])
ylim([0, 1.1])
writeSnmSVT = -min(max(wl1bl0YDiag(cellNo < 256,:)) - ...
    fliplr(min(wl1bl1YDiag(cellNo < 256,:))));
title(sprintf("Write SNM = %.3f V", writeSnmSVT))

% Plot SNM figures (LVT)
figure(122)

% Hold SNM
subplot(1,3,1)
plot(inputVoltages, wl0Transfer(cellNo > 255,:))
hold on
plot(wl0Transfer(cellNo > 255,:), inputVoltages)
hold off
daspect([1,1,1])
xlim([0, 1.1])
ylim([0, 1.1])
holdSnmLVT = min(max(max(wl0YDiag(cellNo > 255,:)) - fliplr(min(...
    wl0YDiag(cellNo > 255,:)))), max(-(max(wl0YDiag(cellNo > 255,:)) - ...
    fliplr(min(wl0YDiag(cellNo > 255,:))))));
title(sprintf("Hold SNM = %.3f V", holdSnmLVT))

% Read SNM
subplot(1,3,2)
plot(inputVoltages, wl1bl1Transfer(cellNo > 255,:))
hold on
plot(wl1bl1Transfer(cellNo > 255,:), inputVoltages)
hold off
daspect([1,1,1])
xlim([0, 1.1])
ylim([0, 1.1])
readSnmLVT = min(max(max(wl1bl1YDiag(cellNo > 255,:)) - fliplr(min(...
    wl1bl1YDiag(cellNo > 255,:)))), max(-(max(wl1bl1YDiag(cellNo > 255,...
    :)) - fliplr(min(wl1bl1YDiag(cellNo > 255,:))))));
title(sprintf("Read SNM = %.3f V", readSnmLVT))

% Write SNM
subplot(1,3,3)
plot(inputVoltages, wl1bl0Transfer(cellNo > 255,:))
hold on
plot(wl1bl1Transfer(cellNo > 255,:), inputVoltages)
hold off
daspect([1,1,1])
xlim([0, 1.1])
ylim([0, 1.1])
writeSnmLVT = -min(max(wl1bl0YDiag(cellNo > 255,:)) - ...
    fliplr(min(wl1bl1YDiag(cellNo > 255,:))));
title(sprintf("Write SNM = %.3f V", writeSnmLVT))

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "sfBiasCurrent", ...
    "cellNo", "inputVoltages", "wl0Output", "wl1bl0Output", ...
    "wl1bl1Output", "wl0Transfer", "wl1bl0Transfer", "wl1bl1Transfer")