%% Test 1: Test Shift Registers
% This script is used to test the shift registers at various speeds and see
% at what speeds they fail.

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\tests\" + ...
    "1_test_shift_registers\";
workspaceName = sprintf("%stest1_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Loop over various speeds and shift data in/out
% Shift standard shift register test pattern '001100110011...' through the
% shift registers at increasing speeds by changing the shift speed register
% on the FPGA.

testPattern = [string(char(kron(ones(1,240), '0011'))) + ...
    string(char(kron(ones(1,480), '01'))), ...
    string(char(kron(ones(1,6), '0011'))) + ...
    string(char(kron(ones(1,10), '01'))) + "0"];

% Perform a binary search to find the speed at which it starts to fail
NUM_TESTS = 18;
testSpeeds = -ones(NUM_TESTS, 2);
testDistances = -ones(NUM_TESTS, 2);

testSpeeds(1,:) = uint16(Inf);
testSpeeds(2,:) = uint16(2);
for iTest = 1:2
    for iReg = 1:2
        testDistances(iTest, iReg) = getShiftSpeedErrors(MSI.FpgaHandle,...
            testSpeeds(iTest, iReg), testPattern(iReg), iReg - 1);
    end
end

slowestError = max(testSpeeds.*(testDistances > 0), [], 1);
fastestCorrect = 1./(min(1./abs(testSpeeds.*(testDistances == 0)), [], 1));

for iTest = 3:NUM_TESTS
    testSpeeds(iTest,:) = uint16((fastestCorrect + slowestError)/2);
    for iReg = 1:2
        testDistances(iTest, iReg) = getShiftSpeedErrors(MSI.FpgaHandle,...
            testSpeeds(iTest, iReg), testPattern(iReg), iReg - 1);
        if testDistances(iTest, iReg) == 0
            fastestCorrect(iReg) = testSpeeds(iTest, iReg);
        else
            slowestError(iReg) = testSpeeds(iTest, iReg);
        end
    end    
end

% Return shift speed to default value
MSI.FpgaHandle.setShiftSpeed(MSI.FpgaHandle.DEFAULT_SHIFT_SPEED)

%% Report maximum speed
% Report maximum speed which is the highest frequency at which the pattern 
% that is shifted out is still correct.
[speedsSorted, idxSorted] = sort(testSpeeds,1);
distancesSorted(:,1) = testDistances(idxSorted(:,1),1) / 1920;
distancesSorted(:,2) = testDistances(idxSorted(:,2),2) / 45;

% Show normalised errors per shift cycle
figure(11)
semilogx(speedsSorted * 10, distancesSorted, 'o-', 'LineWidth', 1)
xlabel("Shift clock period [ns]")
ylabel("Normalised shift data Hamming distance")
title("Normalised shift error rate")
grid on

fprintf("Fastest correct reg 0 shift period setting: %d = %f kHz\n" + ...
    "Fastest correct reg 1 shift period setting: %d = %f kHz\n", ...
    fastestCorrect(1), 1e8 / fastestCorrect(1) / 1e3, fastestCorrect(2),...
    1e8 / fastestCorrect(2) / 1e3);

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "testPattern", ...
    "testSpeeds", "testDistances")
