%% Test 3: Test Ring Oscillator/Delay Chains
% This script is used to test the speed of the delay chains by measuring
% the output frequency of the ring oscillator.

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\tests\" + ...
    "3_test_ro\";
workspaceName = sprintf("%stest3_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Enable the RO for various delays
% Turn on the ring oscillator for maximum delay (~ 100 MHz output) and half
% of that (~200 MHz output) and measure the frequency using the
% oscilloscope (manually).

% Disable reset to prevent settings register automatically resetting
MSI.memoryReset(0)

% RO delay settings to loop over (256 possible settings), offset required 
% for enabling the RO, and the address of the RO settings register
roDelay = round(linspace(255,50,10));
roEnable = 256;
% Address: local controller 8 (weight 64) and settings register (weight 32)
roAddress = 8 * 64 + 32;

roFrequency = zeros(size(roDelay));

% Enable RO at specified delays
for iDelaySetting = 1:numel(roDelay)
    userInput = -1;
    while userInput < 0
        MSI.memoryWrite(roDelay(iDelaySetting) + roEnable, roAddress)
        % Manually enter the RO output frequency for every point
        userInput = input("What is the RO output frequency (in MHz)?" + ...
            " [-1 to resend] ");
    end
    roFrequency(iDelaySetting) = 1e6 * userInput;
end

% Enable reset, also resets the 
MSI.memoryReset(1)

%% Calculate delay chain setting delay
ringDelay = 1./(2*roFrequency);
delayFit = polyfit(roDelay, ringDelay, 1);
fprintf("Delay [ps] = %.3f * setting + %.3f\n", delayFit*1e12)

% Plot frequency and fit vs setting
figure(31)
scatter(roDelay, roFrequency * 1e-6, 'o')
hold on
plot(0:255, 1e-6./(2*(delayFit(1) * (0:255) + delayFit(2))) )
hold off
xlim([0,255])
ylim([0,1e3])
xlabel("Delay chain setting")
ylabel("RO frequency [MHz]")
title("RO frequency")
grid on

% Plot data and fit
figure(32)
scatter(roDelay, ringDelay * 1e12, 'o')
hold on
plot([0,255], (delayFit(1) * [0,255] + delayFit(2)) * 1e12)
hold off
xlim([0,255])
xlabel("Delay chain setting")
ylabel("Chain delay [ps]")
title("RO chain delay")
legend(["Data", sprintf("%.3f * setting + %.3f\n", delayFit*1e12)], ...
    'Location', 'best')
grid on

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "roDelay", ...
    "roFrequency")
