%% Test 4: Test Memory Functionality
% This script is used to test the functionality of all memories by writing
% and reading in fast succession (faster than retention time).

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\tests\" + ...
    "4_test_memories\";
workspaceName = sprintf("%stest4_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Upload test programs to the global controller
% Upload test programs to the global controller to perform write-read
% operations at high speeds with different data patterns.

% Memory local controller indices and reference voltage settings
memoryNo = 0;
% Reference voltages mostly same for SVT/LVT, SRAM value does not matter
referenceVoltage = [0.1, 0.1, 0.1, 0.1, 0.9, 0.9, 1.1, 1.1];
% Settings register values the same for SVT/LVT
settingsRegister = kron([ ...
    bin2dec("000"+"000"+"100000"+"1010"+"00111100"+"11111111"),...
    bin2dec("000"+"000"+"100000"+"0100"+"00011001"+"11111111"), ...
    bin2dec("000"+"000"+"100000"+"1010"+"00111100"+"11111111"), ...
    bin2dec("0"+"0000"+"0"+"000111"+"0000"+"00011010"+"11111111")], [1,1]);

% Keep track of error counters for each run
errorCounters = zeros(numel(memoryNo), 32);

% Load starting program
inputProgram = Program("test_4");
inputProgram.fromAsmFile(testFolder + "test_4.asm")

% Loop over the different memories
for memoryIdx = 1:numel(memoryNo)
    % Set the reference voltage
    MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, ...
        referenceVoltage(memoryNo(memoryIdx) + 1))
    
    % Update program to the right address range and settings
    % Start address
    inputProgram.Register(1) = 64 * memoryNo(memoryIdx);
    % Stop address
    inputProgram.Register(2) = 64 * memoryNo(memoryIdx) + 31;
    % Setting register address
    inputProgram.Register(4) = 64 * memoryNo(memoryIdx) + 32;
    % Setting register data
    inputProgram.Register(5) = settingsRegister(memoryNo(memoryIdx) + 1);

    % Upload, run, and download the program
    outputProgram = MSI.runProgram(inputProgram);
    
    % Save the results of the program
    errorCounters(memoryIdx, :) = outputProgram.ErrorCounter(:);
end


%% Verify results
% Check if the results from all memories are correct
assert(sum(errorCounters, 'all') == 0)
fprintf("All memory cells are functional!\n")

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "memoryNo", ...
    "errorCounters", "referenceVoltage", "settingsRegister")
