%% Test 5: Characterise Sense Amplifiers
% This script is used to characterise the sense amplifiers of all the
% memories.

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\tests\" + ...
    "5_characterise_sense_amplifiers\";
workspaceName = sprintf("%stest5_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Set reference voltages and perform measurements
% Sweep the reference voltages and perform many reads on the SAs to
% characterise them. The number of 0s and 1s can be used to determine the
% input referred noise and offset of the sense amplifiers.

% Reference (relative to 'ideal reference' below) voltage range
refRange = [-0.05, 0, 0.05];%linspace(-0.2, 0.2, 9);
% Max number of test references per memory/reference
maxNumOfTestRef = 100;
% Test reference range per memory (min/max values)
testRefRange = kron([ ...
    [0;0.6],...
    [0;0.6], ...
    [0.5;1.2], ...
    [0.5;1.2]], [1,1]);

% Reference voltages
memoryNo = 0:5;
% memoryReferenceVoltages = [0.085, 0.175, 0.404, 0.1, 0.977, 0.940];
% Settings register values
% settingsRegister = [ ...
%     bin2dec("000"+"001"+"001010"+"1010"+"00110010"+"11111111"), ...
%     bin2dec("000"+"001"+"001010"+"1010"+"01100100"+"11111111"), ...
%     bin2dec("000"+"001"+"001111"+"0100"+"00101000"+"11111111"), ...
%     bin2dec("000"+"001"+"100000"+"0100"+"00011110"+"11111111"), ...
%     bin2dec("000"+"001"+"011110"+"1111"+"00111100"+"11111111"), ...
%     bin2dec("000"+"001"+"011110"+"1111"+"00110010"+"11111111")];

% % 4K
% memoryReferenceVoltages = [0.127, 0.319, 0.204, 0.324, 0.977, 0.931];
memoryReferenceVoltages = [0.085, 0.175, 0.404, 0.1, 0.977, 0.940];
settingsRegister = [ ...
    bin2dec("000"+"111"+"001010"+"1010"+"01111000"+"11111111"), ...
    bin2dec("000"+"111"+"001010"+"1010"+"01100100"+"11111111"), ...
    bin2dec("000"+"111"+"001111"+"0100"+"00101101"+"11111111"), ...
    bin2dec("000"+"111"+"001111"+"0100"+"00101101"+"11111111"), ...
    bin2dec("000"+"111"+"011110"+"1111"+"00111100"+"11111111"), ...
    bin2dec("000"+"111"+"011110"+"1111"+"10110100"+"11111111")];

settingsRegisterReset = kron([ ...
    bin2dec("000"+"000"+"100000"+"1111"+"11000000"+"11111111"),...
    bin2dec("000"+"000"+"100000"+"1111"+"11000000"+"11111111"), ...
    bin2dec("000"+"000"+"100000"+"1111"+"11000000"+"11111111"), ...
    bin2dec("0"+"1000"+"0"+"100000"+"1111"+"11000000"+"11111111")], [1,1]);

resetData = uint32(kron(double([0x00000000, 0x00000000, 0xFFFFFFFF, ...
    0xFFFFFFFF]), [1,1]));

% Load input program
inputProgram = Program("test_5");
%inputProgram.fromAsmFile(testFolder + "test_5.asm")
inputProgram.fromAsmFile(testFolder + "test_5_singleSApolarity.asm")

% Allocate result matrices
errorCounters = NaN(32, maxNumOfTestRef, numel(refRange), ...
    numel(memoryNo));
reference = NaN(32, maxNumOfTestRef, numel(refRange), ...
    numel(memoryNo));
testReference = NaN(32, maxNumOfTestRef, numel(refRange), ...
    numel(memoryNo));

% Loop over memories
for memoryIdx = 1:numel(memoryNo)
    fprintf("Memory %d/%d\n", memoryIdx, numel(memoryNo))
    inputProgram.Register(3) = 64 * memoryNo(memoryIdx);
    inputProgram.Register(4) = 64 * memoryNo(memoryIdx) + 32;
    inputProgram.Register(5) = settingsRegisterReset(...
        memoryNo(memoryIdx) + 1);
    inputProgram.Register(6) = settingsRegister(memoryNo(memoryIdx) + 1);
    inputProgram.Register(7) = resetData(memoryNo(memoryIdx) + 1);

    % References
    referenceVoltages = memoryReferenceVoltages(memoryNo(memoryIdx) ...
        + 1) + refRange;

    % Loop over reference voltages
    for refIdx = 1:numel(referenceVoltages)
        referenceVoltage = referenceVoltages(refIdx);

        MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, ...
            referenceVoltage)

        % Outer most test ref values: should result in all-1 and all-0
        % Lowest test reference
        testRefIdx = 1;
        testRefMin = testRefRange(1, memoryNo(memoryIdx) + 1);

        [errorCounters, reference, testReference] = test_5_measurement( ...
            MSI, errorCounters, reference, testReference, inputProgram, ...
            testRefMin, testRefIdx, refIdx, memoryIdx);

        % Largest test reference
        testRefIdx = 2;
        testRefMax = testRefRange(2, memoryNo(memoryIdx) + 1);

        [errorCounters, reference, testReference] = test_5_measurement( ...
            MSI, errorCounters, reference, testReference, inputProgram, ...
            testRefMax, testRefIdx, refIdx, memoryIdx);

        % For all SAs, perform a binary search to their offset voltage with
        % 1mV step size (supply resolution)
        target = 1023 / 2;
        testRefIdx = 3;
        for SAIdx = 1:32
            % Find limits between which the offset lies and round to mV
            % - Note: will be large for the first couple SAs, but pretty
            % small for later SAs
            upperLimit = testReference(SAIdx, errorCounters(SAIdx, :, ...
                refIdx, memoryIdx) > target, refIdx, memoryIdx);
            lowerLimit = testReference(SAIdx, errorCounters(SAIdx, :, ...
                refIdx, memoryIdx) < target, refIdx, memoryIdx);
            if min(upperLimit) > max(lowerLimit)
                upperLimit = ceil(min(upperLimit) * 1000) / 1000;
                lowerLimit = floor(max(lowerLimit) * 1000) / 1000;
            else
                upperLimit = floor(max(upperLimit) * 1000) / 1000;
                lowerLimit = ceil(min(lowerLimit) * 1000) / 1000;
            end

            % As long as the steps are larger than 1mV:
            while abs(upperLimit - lowerLimit) > 1e-3
                % Compute the next test reference voltage
                testRefVoltage = round((upperLimit + lowerLimit) * 500) ...
                    / 1000;

                % Perform measurement at selected voltage
                [errorCounters, reference, testReference] = ...
                    test_5_measurement(MSI, errorCounters, reference, ...
                    testReference, inputProgram, testRefVoltage, ...
                    testRefIdx, refIdx, memoryIdx);
                testRefActual = testReference(SAIdx, testRefIdx, ...
                        refIdx, memoryIdx);
            
                % Update the limits
                result = errorCounters(SAIdx, testRefIdx, refIdx, ...
                    memoryIdx);
                if result > target
                    if abs(testRefActual - lowerLimit) < abs(upperLimit ...
                            - lowerLimit)
                        upperLimit = testRefActual;
                    else
                        break
                    end
                else
                    if abs(upperLimit - testRefActual) < abs(upperLimit ...
                            - lowerLimit)
                        lowerLimit = testRefActual;
                    else
                        break
                    end
                end
            
                % Increase the test reference index, break loop if it 
                % becomes too large
                testRefIdx = testRefIdx + 1;
                if testRefIdx > maxNumOfTestRef
                    break
                end
            end
            fprintf("%c", 128)
            if testRefIdx > maxNumOfTestRef
                break
            end
        end
        fprintf("\n")
    end
end

% Reset references to reasonable values
MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, 0.25)
MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_TREF, 0.25)

%% Compute sense amplifier statistics
% Use the results from the measurements to compute the input referred noise
% and offset of all the sense amplifiers in the design.

% Matrices for storing the calculated offset and noise
offset = zeros(32, numel(refRange), numel(memoryNo));
noise = zeros(32, numel(refRange), numel(memoryNo));

% Clear loop variables to remove size information
clear diffInputVoltage avgOutputData

for iMemory = 1:numel(memoryNo)
    for iRef = 1:numel(refRange)
        for iSA = 1:32
            avgOutputData = errorCounters(iSA, :, iRef, iMemory) / 1023;
            avgOutputData = avgOutputData(~isnan(avgOutputData));
            diffInputVoltage = reference(iSA, :, iRef, iMemory) - ...
                testReference(iSA, :, iRef, iMemory);
            diffInputVoltage = diffInputVoltage(~isnan(diffInputVoltage));
            % Normal CDF fit (from https://nl.mathworks.com/matlabcentral/
            % answers/345595-fitting-cumulative-normal-distribution-
            % function-to-data)
            [~,minIdx] = min(diffInputVoltage);
            if avgOutputData(minIdx) < 0.5
                fcn = @(b,x) normcdf(x, b(1), b(2));
            else
                fcn = @(b,x) 1 - normcdf(x, b(1), b(2));
            end
            NRCF = @(b) norm(avgOutputData - fcn(b, diffInputVoltage));
            [~,offsetEstimateIdx] = min(abs(avgOutputData - 0.5));
            offsetEstimate = diffInputVoltage(offsetEstimateIdx);
            B = fminsearch(NRCF, [offsetEstimate; 4e-3]);

            offset(iSA, iRef, iMemory) = B(1);
            noise(iSA, iRef, iMemory) = B(2);
        end
    end

    figure(50 + iMemory)
    subplot(2,1,1)
    hold on
    for iSA = 1:32
        plot(reshape(mean(reference(iSA, :, :, iMemory), 2, 'omitnan'), ...
            1, []), offset(iSA, :, iMemory)*1e3, '-o')
    end
    hold off
    title("Input-referred offset")
    xlabel("Reference [V]")
    ylabel("Offset [mV]")
    subplot(2,1,2)
    hold on
    for iSA = 1:32
        plot(reshape(mean(reference(iSA, :, :, iMemory), 2, 'omitnan'), ...
            1, []), noise(iSA, :, iMemory)*1e3, '-o')
    end
    hold off
    title("Input-referred noise")
    xlabel("Reference [V]")
    ylabel("Noise RMS [mV]")
end

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

referenceSAoffset = reshape(mean(reference, 2, 'omitnan'), ...
    [32, numel(refRange), numel(memoryNo)]);

save(workspaceName, "startTemperature", "logFileName", "errorCounters", ...
    "reference", "testReference", "offset", "noise", "memoryNo", ...
    "refRange", "settingsRegister", "settingsRegisterReset", ...
    "resetData", "referenceSAoffset")
