%% Test 5: Characterise Sense Amplifiers
% This script is used to characterise the sense amplifiers of all the
% memories.

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\tests\" + ...
    "5_characterise_sense_amplifiers\";
workspaceName = sprintf("%stest5_around0_satest_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Set reference voltages and perform measurements
% Sweep the reference voltages and perform many reads on the SAs to
% characterise them. The number of 0s and 1s can be used to determine the
% input referred noise and offset of the sense amplifiers.

% Reference (relative to 'ideal reference' below) voltage range
refRange = kron(ones(1,3), [-0.1:0.01:0.1, 0.1:-0.01:-0.1]);

% Reference voltages
memoryNo = 4;
memoryReferenceVoltages = kron([0.03, 0.03, 1.05], [1,1]);
% Settings register values the same for SVT/LVT
settingsRegister = kron([ ...
    bin2dec("000"+"001"+"100000"+"1111"+"01000000"+"11111111"),...
    bin2dec("000"+"001"+"100000"+"1111"+"00011000"+"11111111"), ...
    bin2dec("000"+"001"+"011110"+"1010"+"00111100"+"11111111")], [1,1]);

settingsRegisterReset = kron([ ...
    bin2dec("000"+"000"+"100000"+"1010"+"01101110"+"11111111"),...
    bin2dec("000"+"000"+"100000"+"0100"+"00100000"+"11111111"), ...
    bin2dec("000"+"000"+"100000"+"1000"+"10000000"+"11111111")], [1,1]);

% Load input program
inputProgram = Program("test_5");
inputProgram.fromAsmFile(testFolder + "test_5.asm")
%inputProgram.fromAsmFile(testFolder + "test_5_singleSApolarity.asm")
% Reset registers to 0
inputProgram.Register(7) = 0x00000000;
% Reset registers to 1
%inputProgram.Register(7) = 0xFFFFFFFF;

% Allocate result matrices
errorCounters = NaN(32, numel(refRange), numel(memoryNo));
reference = NaN(32, numel(refRange), numel(memoryNo));
testReference = NaN(32, numel(refRange), numel(memoryNo));

MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_TREF, 1.1)

% Loop over memories
for memoryIdx = 1:numel(memoryNo)
    fprintf("Memory %d/%d\n", memoryIdx, numel(memoryNo))
    inputProgram.Register(3) = 64 * memoryNo(memoryIdx);
    inputProgram.Register(4) = 64 * memoryNo(memoryIdx) + 32;
    %inputProgram.Register(5) = settingsRegisterReset( ...
    %    memoryNo(memoryIdx) + 1);
    inputProgram.Register(5) = settingsRegister(memoryNo(memoryIdx) + 1);

    % References
    referenceVoltages = memoryReferenceVoltages(memoryNo(memoryIdx) ...
        + 1) + refRange;

    % Loop over reference voltages
    for refIdx = 1:numel(referenceVoltages)
        referenceVoltage = referenceVoltages(refIdx);

        MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, ...
            referenceVoltage)

        outputProgram = MSI.runProgram(inputProgram);

        reference(:, refIdx, memoryIdx) = ...
            MSI.measureSupplyChannelVoltage(MSI.SUPPLY_CHANNEL_REF);
        testReference(:, refIdx, memoryIdx) = ...
            MSI.measureSupplyChannelVoltage(MSI.SUPPLY_CHANNEL_TREF);
        errorCounters(:, refIdx, memoryIdx) = ...
            outputProgram.ErrorCounter(:);
        
    end
end

%% Plot data
figure(5001)
plot3(reference(1,:,1)-testReference(1,:,1), errorCounters(1,:,1), ...
    ones(1,numel(refRange)))
hold on
for i=2:32
    plot3(reference(i,:,1)-testReference(i,:,1), errorCounters(i,:,1), ...
        i * ones(1,numel(refRange)))
end
hold off

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "errorCounters", ...
    "reference", "testReference", "memoryNo", "refRange", ...
    "settingsRegister", "writeData")
