%% Test 5: Characterise Sense Amplifiers
% This script is used to characterise the sense amplifiers of all the
% memories.

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\tests\" + ...
    "5_characterise_sense_amplifiers\";
workspaceName = sprintf("%stest5_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Set reference voltages and perform measurements
% Sweep the reference voltages and perform many reads on the SAs to
% characterise them. The number of 0s and 1s can be used to determine the
% input referred noise and offset of the sense amplifiers.

% Reference (relative to 'ideal reference' below) and test reference 
% (relative to reference) voltage ranges
refRange = linspace(-0.15, 0.15, 7);
testRefRange = linspace(-0.1, 0.1, 21);

% Reference voltages
memoryNo = 0:7;
memoryReferenceVoltages = [0.35, 0.35, 0.25, 0.25, 0.75, 0.75, 0.85, 0.85];
% Settings register values the same for SVT/LVT
settingsRegister = kron([ ...
    bin2dec("000"+"001"+"100000"+"1000"+"10000000"+"11111111"),...
    bin2dec("000"+"001"+"100000"+"0100"+"00100000"+"11111111"), ...
    bin2dec("000"+"001"+"100000"+"1000"+"10000000"+"11111111"), ...
    bin2dec("0"+"1000"+"1"+"100000"+"1000"+"10000000"+"11111111")], [1,1]);

% Load input program
inputProgram = Program("test_5");
inputProgram.fromAsmFile(testFolder + "test_5.asm")

% Allocate result matrices
errorCounters = zeros(numel(memoryNo), numel(refRange), ...
    numel(testRefRange), 32);
reference = zeros(numel(memoryNo), numel(refRange), ...
    numel(testRefRange), 32);
testReference = zeros(numel(memoryNo), numel(refRange), ...
    numel(testRefRange), 32);

% Loop over memories
for memoryIdx = 1:numel(memoryNo)
    fprintf("Memory %d/%d\n", memoryIdx, numel(memoryNo))
    inputProgram.Register(4) = 64 * memoryNo(memoryIdx) + 32;
    inputProgram.Register(5) = settingsRegister(memoryIdx);
    inputProgram.Register(3) = 64 * memoryNo(memoryIdx);

    % References
    referenceVoltages = memoryReferenceVoltages(memoryIdx) + refRange;

    % Loop over reference voltages
    for refIdx = 1:numel(referenceVoltages)
        referenceVoltage = referenceVoltages(refIdx);

        % Test references (reference +- 20 mV)
        testReferenceVoltages = referenceVoltage + testRefRange;
    
        MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, ...
            referenceVoltage)

        for testRefIdx = 1:numel(testReferenceVoltages)
            testReferenceVoltage = testReferenceVoltages(testRefIdx);
            MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_TREF, ...
                testReferenceVoltage)
    
            outputProgram = MSI.runProgram(inputProgram);
    
            errorCounters(memoryIdx, refIdx, testRefIdx, :) = ...
                outputProgram.ErrorCounter(:);
            reference(memoryIdx, refIdx, testRefIdx, :) = ...
                MSI.measureSupplyChannelVoltage(MSI.SUPPLY_CHANNEL_REF);
            testReference(memoryIdx, refIdx, testRefIdx, :) = ...
                MSI.measureSupplyChannelVoltage(MSI.SUPPLY_CHANNEL_TREF);
        end
    end
end

%% Compute sense amplifier statistics
% Use the results from the measurements to compute the input referred noise
% and offset of all the sense amplifiers in the design.

% Matrices for storing the calculated offset and noise
offset = zeros(numel(memoryNo), numel(refRange), 32);
noise = zeros(numel(memoryNo), numel(refRange), 32);

% Clear loop variables to remove size information
clear diffInputVoltage avgOutputData

for iMemory = 1:numel(memoryNo)
    for iRef = 1:numel(refRange)
        for iSA = 1:32
            avgOutputData(:) = errorCounters(iMemory, iRef, :, iSA) / 1023;
            diffInputVoltage(:) = reference(iMemory, iRef, :, iSA) - ...
                testReference(iMemory, iRef, :, iSA);
            % Normal CDF fit (from https://nl.mathworks.com/matlabcentral/
            % answers/345595-fitting-cumulative-normal-distribution-
            % function-to-data)
            if avgOutputData(1) < 0.5
                fcn = @(b,x) 1 - normcdf(x, b(1), b(2));
            else
                fcn = @(b,x) normcdf(x, b(1), b(2));
            end
            NRCF = @(b) norm(avgOutputData - fcn(b, diffInputVoltage)); 
            B = fminsearch(NRCF, [0; 1e-2]);

            offset(iMemory, iRef, iSA) = B(1);
            noise(iMemory, iRef, iSA) = B(2);
        end
    end

    figure(50 + iMemory)
    subplot(2,1,1)
    hold on
    for iSA = 1:32
        plot(mean(reference(iMemory, :, :, iSA), 3), offset(iMemory, :, ...
            iSA), '-o')
    end
    hold off
    title("Input-referred offset")
    xlabel("Reference [V]")
    ylabel("[V]")
    subplot(2,1,2)
    hold on
    for iSA = 1:32
        plot(mean(reference(iMemory, :, :, iSA), 3), noise(iMemory, :, ...
            iSA), '-o')
    end
    hold off
    title("Input-referred noise")
    xlabel("Reference [V]")
    ylabel("[V]")
end

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "errorCounters",...
    "reference", "testReference", "offset", "noise")
