%% Test 6: Cell Retention Time (simple)
% This script is used to find the simple retention time (fixed reference
% voltage) for all dynamic cells.

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\tests\" + ...
    "6_cell_retention_time_simple\";
workspaceName = sprintf("%stest6_mult_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Set reference voltages and perform measurements
% Set the reference voltage, loop over the memories and rows, and sweep the
% hold time between write and read instruction to find the time to failure
% for each cell.

% memoryNo = inputData(1:(end-2), 1)';
% memoryReferenceVoltages = inputData(1:(end-2), 5)';
% readT1 = inputData(1:(end-2), 2)';
% readT2 = inputData(1:(end-2), 3)';
% writeT1 = inputData(1:(end-2), 4)';

memorySel = 1:30; %[25, 30];%[6:10, 12:15, 17:20];
memoryNo = inputData(memorySel, 1)';
memoryReferenceVoltages = inputData(memorySel, 5)';
readT1 = inputData(memorySel, 2)';
readT2 = inputData(memorySel, 3)';
writeT1 = inputData(memorySel, 4)';
settingsRegister = zeros(size(memoryNo));

for i = 1:numel(memoryNo)
	settingsRegister(i) = DynamicMemoryCharacterisation.createSetting(...
        255, readT1(i), readT2(i), writeT1(i), 0, 1, 1);
end

settingsRegisterReset = bin2dec("000"+"111"+"011110"+"1111"+"10000000"+ ...
    "11111111");
	
clear testReferenceVoltage
% Test ref voltages for reset to 1 (for data = 0)
testReferenceVoltage(:,1) = max(memoryReferenceVoltages - 0.15, 0);
% Test ref voltages for reset to 0 (for data = 1)
testReferenceVoltage(:,2) = min(memoryReferenceVoltages + 0.15, 1.1);

% Rows of interest (select single row for anomalies or script testing)
rowNo = 0:31;
% Data of interest (can select single value)
writeData = 0:1;

% Delay settings for first program iteration
firstRunDelayInitial = 0;
firstRunDelayStep = 500;

% Load starting program
inputProgram = Program("test_6_rst");
inputProgram.fromAsmFile(testFolder + "test_6_rst.asm")

% Allocate result matrices (every memory, every row, every data polarity, 
% 2 runs, 32 cells per run)
errorCounters = zeros(numel(memoryNo), numel(rowNo), numel(writeData), ...
    3, 32);
secondRunDelay = zeros(numel(memoryNo), numel(rowNo), numel(writeData), 2);
thirdRunDelay = zeros(numel(memoryNo), numel(rowNo), numel(writeData), 2);

% Loop over memories
for memoryIdx = 1:numel(memoryNo)
    fprintf("Memory %d/%d (= %d)\n", memoryIdx, numel(memoryNo), ...
        memoryNo(memoryIdx))
    % Load memory settings
    inputProgram.Register(4) = 64 * memoryNo(memoryIdx) + 32;
    inputProgram.Register(5) = settingsRegister(memoryIdx);
    inputProgram.Register(9) = settingsRegisterReset;

    % Set memory reference voltage
    MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, ...
        memoryReferenceVoltages(memoryIdx))

    % Loop over rows of interest
    for rowIdx = 1:numel(rowNo)
%         fprintf("Row %d/%d (= %d)\n", rowIdx, numel(rowNo), rowNo(rowIdx))
        inputProgram.Register(6) = 64 * memoryNo(memoryIdx) + ...
            rowNo(rowIdx);
        % Loop over write data of interest
        for writeDataIdx = 1:numel(writeData)
%             fprintf("Data %d/%d (= %d)\n", writeDataIdx, ...
%                 numel(writeData), writeData(writeDataIdx))
            % Set test reference for reset
            MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_TREF, ...
                testReferenceVoltage(memoryIdx, ...
                writeData(writeDataIdx) + 1))

            inputProgram.Register(5) = bitset(inputProgram.Register(5), ...
                29, 1 - writeData(writeDataIdx));
            inputProgram.Register(7) = 0xFFFFFFFF * ...
                writeData(writeDataIdx);
            % Set the loop step sizes and initial value
            % Run program 3x:
            % 1. Large steps to find the point where cells start failing
            % 2. Small steps to find the point where cells start failing
            % 3. Sweep over critical range (32 x 32)

            % ----- 1 -----
            inputProgram.Register(1) = firstRunDelayStep;
            inputProgram.Register(2) = firstRunDelayStep;
            inputProgram.Register(3) = firstRunDelayInitial;

            outputProgram = MSI.runProgram(inputProgram);
            
            % Save results
            errorCounters(memoryIdx, rowIdx, writeDataIdx, 1, :) = ...
                outputProgram.ErrorCounter(:);

            % Calculate step size for the second program iteration
            delay = firstRunDelayInitial + (0:1022) * firstRunDelayStep;

            [worstErrorCount, worstCell] = max(outputProgram...
                .ErrorCounter(:));
            if worstErrorCount == 1023
%                 warning("Cell (%d,%d) always fails, decrease the " + ...
%                     "initial delay", rowNo(rowIdx), worstCell - 1)
                worstErrorCount = max(outputProgram.ErrorCounter( ...
                    outputProgram.ErrorCounter < 1023));
                if isempty(worstErrorCount)
%                     warning("All cells in row %d fail", rowNo(rowIdx))
                    continue
                end
            end
            secondRunDelay(memoryIdx, rowIdx, writeDataIdx, 1) = ...
                delay(1023 - worstErrorCount);

            [bestErrorCount, bestCell] = min(outputProgram...
                .ErrorCounter(:));
            if bestErrorCount == 0
%                 warning("Cell (%d,%d) always works, increase the " + ...
%                     "initial delay or step size", rowNo(rowIdx), ...
%                     bestCell - 1)
                bestErrorCount = min(outputProgram.ErrorCounter( ...
                    outputProgram.ErrorCounter > 0));
                if isempty(bestErrorCount)
%                     warning("All cells in row %d work", rowNo(rowIdx))
                    continue
                end
            end
            secondRunDelay(memoryIdx, rowIdx, writeDataIdx, 2) = ...
                ceil((delay(1024 - bestErrorCount) - ...
                secondRunDelay(memoryIdx, rowIdx, writeDataIdx, 1)) ...
                / 1023);

            % ----- 2 -----
            inputProgram.Register(1) = secondRunDelay(memoryIdx, ...
                rowIdx, writeDataIdx, 2);
            inputProgram.Register(2) = secondRunDelay(memoryIdx, ...
                rowIdx, writeDataIdx, 2);
            inputProgram.Register(3) = secondRunDelay(memoryIdx, ...
                rowIdx, writeDataIdx, 1);

            outputProgram = MSI.runProgram(inputProgram);
            
            % Save results
            errorCounters(memoryIdx, rowIdx, writeDataIdx, 2, :) = ...
                outputProgram.ErrorCounter(:);

            % Calculate step size for the second program iteration
            worstErrorCount = max(outputProgram.ErrorCounter(:));
            bestErrorCount = min(outputProgram.ErrorCounter( ...
                outputProgram.ErrorCounter > 0));

            if ~any(outputProgram.ErrorCounter > 0)
%                 warning("No second run errors in row %d", rowNo(rowIdx))
                thirdRunDelay(memoryIdx, rowIdx, writeDataIdx, :) = ...
                    secondRunDelay(memoryIdx, rowIdx, writeDataIdx, :);
            else
                delay = secondRunDelay(memoryIdx, rowIdx, writeDataIdx, ...
                    1) + (-1:1023) * secondRunDelay(memoryIdx, rowIdx, ...
                    writeDataIdx, 2);
    
                thirdRunDelay(memoryIdx, rowIdx, writeDataIdx, 1) = ...
                    delay(1024 - worstErrorCount);
                thirdRunDelay(memoryIdx, rowIdx, writeDataIdx, 2) = ...
                    ceil((delay(1025 - bestErrorCount) - ...
                    thirdRunDelay(memoryIdx, rowIdx, writeDataIdx, 1))/32);
            end

            % ----- 3 -----
            inputProgram.Register(1) = 0;
            inputProgram.Register(2) = thirdRunDelay(memoryIdx, ...
                rowIdx, writeDataIdx, 2);
            inputProgram.Register(3) = thirdRunDelay(memoryIdx, ...
                rowIdx, writeDataIdx, 1);

            outputProgram = MSI.runProgram(inputProgram);
            
            % Save results
            errorCounters(memoryIdx, rowIdx, writeDataIdx, 3, :) = ...
                outputProgram.ErrorCounter(:);
            
        end
    end
end


%% Construct cell retention time heatmap
% Use the results from the measurements to construct a heatmap showing the
% best and worst cells in terms of worst case retention time.

% Assumption: cells suddenly start to fail
% - First N groups of 32: no failure (error counter + 0)
% - N+1 th group: failing sometimes (error counter + 1 to 32)
% - N+2 to 32th group: full failure (error counter + 32)
% For every element in errorCounters(:,:,:,3,:) -> find delay of N+1th
% group

delayGroup = 32 - ceil(errorCounters(:,:,:,3,:) / 32);
criticalDelay = thirdRunDelay(:,:,:,1) + thirdRunDelay(:,:,:,2) .* ...
    delayGroup;
criticalDelay(errorCounters(:,:,:,1,:) == 0) = Inf;
criticalDelay(errorCounters(:,:,:,1,:) == 1023) = 0;
criticalDelay(criticalDelay < 0) = 0;

retTime = NaN(32, 32);
worstCellRetTime = zeros(size(memoryNo));
fcn = @(b,x) normcdf(x, b(1), b(2));

for memoryIdx = 1:numel(memoryNo)
	retTime(rowNo+1,:) = 8e-8 + 2e-8 * reshape(min(criticalDelay( ...
    memoryIdx,:,:,1,:), [], 3), [numel(rowNo), 32]);
	NRCF = @(b) norm(sum(retTime(retTime ~= Inf)' <= retTime( ...
		retTime ~= Inf), 2)/numel(retTime) - fcn(b, log(retTime( ...
		retTime ~= Inf))));
	B = fminsearch(NRCF, [-5; 1]);
    worstCellRetTime(memoryIdx) = min(retTime, [], 'all');
    fprintf("%2d: mu, sigma, min = %7.3f, %7.3f, %e (%dx)\n", ...
		memoryIdx, B(1), B(2), worstCellRetTime(memoryIdx), ...
		sum(retTime == min(retTime, [], 'all'), 'all'))
end

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "errorCounters",...
    "secondRunDelay", "thirdRunDelay", "delayGroup", "criticalDelay", ...
    "memoryNo", "rowNo", "writeData", "memoryReferenceVoltages", ...
    "testReferenceVoltage", "settingsRegisterReset")
