%% Test 7: Characterise Dynamic Cells
% This script is used to characterise dynamic cells. It sweeps the
% reference voltage and hold time to determine the RBL voltage after a read
% operation as a function of the hold time.

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\tests\" + ...
    "7_characterise_dynamic_cells\";
workspaceName = sprintf("%stest7_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Set reference voltages and perform measurements
% Perform multiple measurements with various references to find the RBL
% voltage of a specified cell for a given delay.

% Number of comparisons at each voltage
numComparisons = 1023;
% List of number of delay loops between write and read operation to test
%delayList = 2.^(6:12);
%delayList = 2.^(0:7);
%delayList = ([100,500,1000,1500,2000,3000,4000]-80)/20; % ns to setting
% delayList = ([logspace(2, 7, 11)]-80)/20; % ns to setting
% delayList = ([logspace(2, log10(2e4), 11)]-80)/20; % ns to setting

delayList = zeros(1,30);
temp = NaN(1, numel(delayList) + 1);

% List of reference limits for each of the memories to sweep between
memoryReferenceLimits = kron([0, 0.6; 0, 0.6; 0.7, 1.2], [1;1]);
% Total number of reference voltages to try
numberOfReferences = 11;
% Settings register values (room temp)
% settingsRegister = [ ...
%     bin2dec("000"+"110"+"001010"+"1010"+"00111100"+"11111111"), ...
%     bin2dec("000"+"110"+"001010"+"1010"+"01010000"+"11111111"), ...
%     bin2dec("000"+"110"+"001010"+"0100"+"00101000"+"11111111"), ...
%     bin2dec("000"+"110"+"100000"+"0100"+"00011110"+"11111111"), ...
%     bin2dec("000"+"110"+"011110"+"1111"+"00111100"+"11111111"), ...
%     bin2dec("000"+"110"+"011110"+"1111"+"00111100"+"11111111")];
% Settings register values (cryo)
settingsRegister = [ ...
    bin2dec("000"+"110"+"001010"+"1010"+"01100100"+"11111111"), ...
    bin2dec("000"+"110"+"001010"+"1010"+"01100100"+"11111111"), ...
    bin2dec("000"+"110"+"001111"+"0100"+"00011110"+"11111111"), ...
    bin2dec("000"+"110"+"001111"+"0100"+"00011110"+"11111111"), ...
    bin2dec("000"+"110"+"011110"+"1111"+"01010000"+"11111111"), ...
    bin2dec("000"+"110"+"011110"+"1111"+"00110010"+"11111111")];
settingsRegisterReset = bin2dec("000"+"111"+"011110"+"1111"+"10000000"+ ...
    "11111111");
% Test ref voltages for reset to 1 (when data = 0) and 0 (when data = 1)
clear testReferenceVoltage
testReferenceVoltage(:,1) = min(memoryReferenceLimits, [], 2);
testReferenceVoltage(:,2) = max(memoryReferenceLimits, [], 2);

% List of cells to test: row = memory number (0-5), row number (0-31),
% column number (0-31)
% cellList = zeros(16, 3);
% cellList(:,1) = 4;
% cellList(:,2) = kron(0:8:30, ones(1,4));
% cellList(:,3) = kron(ones(1,4), 0:8:30);

% cellList = [4, 0, 0;
%             4, 0, 1;
%             5, 0, 0;
%             5, 0, 1];

% cellList = [0, 0, 26;
%             0, 10, 13;
%             0, 0, 30;
%             1, 22, 5;
%             1, 6, 15;
%             2, 15, 1;
%             2, 30, 16;
%             2, 20, 29;
%             3, 11, 10;
%             3, 30, 11;
%             3, 3, 21;
%             4, 4, 23;
%             4, 0, 0;
%             5, 8, 14;
%             5, 4, 0];

cellList = [1, 11, 24];

% List of data polarities to test (0, 1, or both)
% dataList = [0, 1];
dataList = 0;

% Allocate result matrices
errorCounters = zeros(32, numberOfReferences, numel(dataList), ...
    numel(delayList), size(cellList,1));
references = zeros(numberOfReferences, numel(dataList), ...
    numel(delayList), size(cellList, 1));

% Load starting program
% inputProgram = Program("test_7");
% inputProgram.fromAsmFile(testFolder + "test_7.asm")
inputProgram = Program("test_7_rst");
inputProgram.fromAsmFile(testFolder + "test_7_rst.asm")

% Loop over the cells of interest
for cellIdx = 1:size(cellList,1)
    fprintf("Cell %d/%d\n", cellIdx, size(cellList, 1))
    fprintf("Temp = %f K\n", MSI.getTemperature());
    memoryNo = cellList(cellIdx, 1);
    rowNo = cellList(cellIdx, 2);
    colIdx = cellList(cellIdx, 3) + 1;

    inputProgram.Register(2) = numComparisons;
    inputProgram.Register(4) = 64 * memoryNo + 32;
    inputProgram.Register(5) = settingsRegister(memoryNo + 1);
    inputProgram.Register(6) = 64 * memoryNo + rowNo;
    inputProgram.Register(9) = settingsRegisterReset;
    % Loop over the data polarities
    for dataIdx = 1:numel(dataList)
        inputProgram.Register(5) = bitset(inputProgram.Register(5), ...
            29, 1 - dataList(dataIdx));
        inputProgram.Register(7) = 0xFFFFFFFF * dataList(dataIdx);

        MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_TREF, ...
            testReferenceVoltage(memoryNo + 1, dataList(dataIdx) + 1))

        % Loop over the delays
        for delayIdx = 1:numel(delayList)
            temp(delayIdx) = MSI.getTemperature;
    
            inputProgram.Register(8) = delayList(delayIdx);
            % Measure the outer references first, followed by binary search
            % to find the reference voltage for which the number of errors
            % is roughly 50%.
            % Throw an warning if the 50% point lies outside the reference
            % limits and continue with the next data point
            references(1, dataIdx, delayIdx, cellIdx) = ...
                memoryReferenceLimits(memoryNo + 1, 1);
            references(2, dataIdx, delayIdx, cellIdx) = ...
                memoryReferenceLimits(memoryNo + 1, 2);
            for refIdx = 1:2
                MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, ...
                    references(refIdx, dataIdx, delayIdx, cellIdx))
                outputProgram = MSI.runProgram(inputProgram);
                errorCounters(:, refIdx, dataIdx, delayIdx, cellIdx) = ...
                    outputProgram.ErrorCounter(:);
            end
            errors = errorCounters(colIdx, 1:2, dataIdx, delayIdx, ...
                cellIdx);
            refs = references(1:2, dataIdx, delayIdx, cellIdx);
            % Check if one reference gives all 0 and one gives all 1
            if min(errors) > (numComparisons - 1)/2 || max(errors) < ...
                    (numComparisons + 1)/2
                warning("Reference voltage limits for (%d, %d, %d) " + ...
                    "too tight: [%f, %f] V gives [%d, %d]", cellIdx, ...
                    delayIdx, dataIdx, refs(1), refs(2), errors(1), ...
                    errors(2))
                continue
            end
            minRef = refs(errors == min(errors));
            maxRef = refs(errors == max(errors));

            for refIdx = 3:numberOfReferences
                newRef = (minRef + maxRef) / 2;
                references(refIdx, dataIdx, delayIdx, cellIdx) = newRef;
                MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, ...
                    newRef);  
                outputProgram = MSI.runProgram(inputProgram);
                errorCounters(:, refIdx, dataIdx, delayIdx, cellIdx) = ...
                    outputProgram.ErrorCounter(:);
                cellErrorCount = errorCounters(colIdx, refIdx, dataIdx, ...
                    delayIdx, cellIdx);
                if cellErrorCount < (numComparisons + 1)/2
                    minRef = newRef;
                else
                    maxRef = newRef;
                end
            end
            fprintf("%c", 128)
        end
    end
    fprintf("\n")
end
temp(end) = MSI.getTemperature;

%% Calculate RBL voltages of cells of interest
% Calculate the likely RBL voltage (and spread?) for the specified cells,
% delays, and data polarities by fitting the references and normalised
% error counts to a normal CDF.

% Allocate results matrix
RBLVoltage = zeros(numel(dataList), numel(delayList), size(cellList, 1));
RBLVoltageStd = zeros(numel(dataList), numel(delayList), ...
    size(cellList, 1));

% Clear loop variables to remove size information
clear refVoltages avgOutputData

% Loop over cells, delays, and data lists
for cellIdx = 1:size(cellList,1)
    memoryNo = cellList(cellIdx, 1);
    colIdx = cellList(cellIdx, 3) + 1;
    for delayIdx = 1:numel(delayList)
        for dataIdx = 1:numel(dataList)
            refVoltages(:) = references(:, dataIdx, delayIdx, cellIdx);
            avgOutputData(:) = errorCounters(colIdx, :, dataIdx, ...
                delayIdx, cellIdx) / numComparisons;

            % Normal CDF fit (from https://nl.mathworks.com/matlabcentral/
            % answers/345595-fitting-cumulative-normal-distribution-
            % function-to-data)
            % Depending on if the average output goes from 0 to 1 or from 1
            % to 0, the CDF must be flipped.
            if avgOutputData(1) < 0.5
                fcn = @(b,x) normcdf(x, b(1), b(2));
            else
                fcn = @(b,x) 1 - normcdf(x, b(1), b(2));
            end
            NRCF = @(b) norm(avgOutputData - fcn(b, refVoltages));
            offsetEstimate = refVoltages(end);
            B = fminsearch(NRCF, [offsetEstimate; 2e-3]);

            RBLVoltage(dataIdx, delayIdx, cellIdx) = B(1);% - interp1( ...
                %referenceSAoffset(colIdx, :, memoryNo + 1), ...
                %offset(colIdx, :, memoryNo + 1), B(1), 'spline');
            RBLVoltageStd(dataIdx, delayIdx, cellIdx) = B(2);
            
        end
    end
%     figure(70 + cellIdx)
%     legendText = string.empty(numel(dataList),0);
%     hold on
%     for dataIdx = 1:numel(dataList)
%         errorbar((8+2*delayList)*1e-8, RBLVoltage(dataIdx, :, cellIdx), ...
%             RBLVoltageStd(dataIdx, :, cellIdx))
%         legendText(dataIdx) = sprintf("Data = %d", dataList(dataIdx));
%     end
%     hold off
%     xlabel("Hold time [s]")
%     ylabel("RBL voltage + SA offset/noise [V]")
%     legend(legendText)
end

% for memoryNo = unique(cellList(:,1))'
%     figure(70 + memoryNo)
%     hold on
%     for cellIdx = find(cellList(:,1) == memoryNo)'
%         errorbar((8+2*delayList)*1e-8, RBLVoltage(1, :, cellIdx), ...
%             RBLVoltageStd(1, :, cellIdx), 'r')
%         errorbar((8+2*delayList)*1e-8, RBLVoltage(2, :, cellIdx), ...
%             RBLVoltageStd(2, :, cellIdx), 'b')
%     end
%     hold off
%     xlabel("Hold time [s]")
%     ylabel("RBL voltage + SA offset/noise [V]")
% end

% figure(72)
% histogram(RBLVoltage(1,:,:), 15)
% xlabel("RBL voltage + SA offset/noise [V]")

figure(73)
errorbar((temp(1:(end-1)) + temp(2:end))/2, RBLVoltage(1, :, cellIdx), ...
    RBLVoltageStd(1, :, cellIdx), RBLVoltageStd(1, :, cellIdx), ...
    diff(temp)/2, diff(temp)/2, 'LineStyle', 'none')
xlabel("Temperature [K]")
ylabel("RBL voltage [V]")
title("RBL voltage at hold time = 0")

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "errorCounters", ... 
    "references", "cellList", "delayList", "dataList", "RBLVoltage", ...
    "RBLVoltageStd", "settingsRegister", "numComparisons", "temp")
