%% Test 8: Characterise Dynamic Cells
% This script is used to characterise dynamic cells. It sweeps the
% reference voltage and read precharge time to determine the RBL voltage 
% after a read operation as a function of the read reset/precharge.

runTime = string(datetime('now', 'Format', 'yyyyMMddHHmmssSSS'));
testFolder = "C:\Users\radamsteegt\Documents\MATLAB\tests\" + ...
    "8_characterise_dynamic_cells\";
workspaceName = sprintf("%stest8c_%s", testFolder, runTime);

%% Run startup script
% Connect with the FPGA and instruments and set up the initial instrument 
% output values and configurations.

if ~exist("MSI", "var")
    MSI = MeasurementSetupInterface();
end
startTemperature = MSI.getTemperature();
logFileName = MSI.LogHandle.FileName;

%% Set reference voltages and perform measurements
% Perform multiple measurements with various references to find the RBL
% voltage of a specified cell for a given delay.

% List of number of read times to test
readPrechargeDurationList = 0:2:15;
% List of reference limits for each of the memories to sweep between
memoryReferenceLimits = kron([0, 0.6; 0, 0.5; 0.6, 1.1], [1;1]);
% Total number of reference voltages to try
numberOfReferences = 11;
% Settings register values the same for SVT/LVT
settingsRegister = [ ...
    bin2dec("000"+"110"+"001010"+"0000"+"00111100"+"11111111"), ...
    bin2dec("000"+"110"+"001010"+"0000"+"01100100"+"11111111"), ...
    bin2dec("000"+"110"+"100000"+"0000"+"00011000"+"11111111"), ...
    bin2dec("000"+"110"+"100000"+"0000"+"00011000"+"11111111"), ...
    bin2dec("000"+"110"+"100000"+"0000"+"00110010"+"11111111"), ...
    bin2dec("000"+"110"+"100000"+"0000"+"00110010"+"11111111")];
settingsRegisterReset = bin2dec("000"+"111"+"011110"+"1111"+"10000000"+ ...
    "11111111");
% Test ref voltages for reset to 0
%testReferenceVoltage = max(memoryReferenceLimits, [], 2);
% Test ref voltages for reset to 1
testReferenceVoltage = min(memoryReferenceLimits, [], 2);

% List of cells to test: row = memory number (0-5), row number (0-31),
% column number (0-31)
cellList = zeros(16, 3);
cellList(:,1) = 1;
cellList(:,2) = kron(0:8:30, ones(1,4));
cellList(:,3) = kron(ones(1,4), 0:8:30);
% List of data polarities to test (0, 1, or both)
dataList = [0, 1];

% Allocate result matrices
errorCounters = zeros(32, numberOfReferences, numel(dataList), ...
    numel(readPrechargeDurationList), size(cellList,1));
references = zeros(numberOfReferences, numel(dataList), ...
    numel(readPrechargeDurationList), size(cellList, 1));

% Load starting program
% inputProgram = Program("test_8c");
% inputProgram.fromAsmFile(testFolder + "test_8c.asm")
inputProgram = Program("test_8c_rst");
inputProgram.fromAsmFile(testFolder + "test_8c_rst.asm")

% Loop over the cells of interest
for cellIdx = 1:size(cellList,1)
    fprintf("Cell %d/%d\n", cellIdx, size(cellList, 1))
    memoryNo = cellList(cellIdx, 1);
    rowNo = cellList(cellIdx, 2);
    colIdx = cellList(cellIdx, 3) + 1;

    inputProgram.Register(4) = 64 * memoryNo + 32;
    inputProgram.Register(5) = settingsRegister(memoryNo + 1);
    inputProgram.Register(6) = 64 * memoryNo + rowNo;
    inputProgram.Register(12) = settingsRegisterReset;

    % Loop over the delays
    for delayIdx = 1:numel(readPrechargeDurationList)
        inputProgram.Register(9) = readPrechargeDurationList(delayIdx);
        % Loop over the data polarities
        for dataIdx = 1:numel(dataList)
            inputProgram.Register(5) = bitset(inputProgram.Register(5), ...
                29, 1 - dataList(dataIdx));
            inputProgram.Register(7) = 0xFFFFFFFF * dataList(dataIdx);
            % Measure the outer references first, followed by binary search
            % to find the reference voltage for which the number of errors
            % is roughly 50%.
            % Throw an error if the 50% point lies outside the reference
            % limits.
            references(1, dataIdx, delayIdx, cellIdx) = ...
                memoryReferenceLimits(memoryNo + 1, 1);
            references(2, dataIdx, delayIdx, cellIdx) = ...
                memoryReferenceLimits(memoryNo + 1, 2);
            for refIdx = 1:2
                MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, ...
                    references(refIdx, dataIdx, delayIdx, cellIdx))
                outputProgram = MSI.runProgram(inputProgram);
                errorCounters(:, refIdx, dataIdx, delayIdx, cellIdx) = ...
                    outputProgram.ErrorCounter(:);
            end
            errors = errorCounters(colIdx, 1:2, dataIdx, delayIdx, ...
                cellIdx);
            refs = references(1:2, dataIdx, delayIdx, cellIdx);
            % Check if one reference gives all 0 and one gives all 1
            if min(errors) > 511 || max(errors) < 512
                error("Reference voltage limits to tight: [%f, %f] V" + ...
                    " gives [%d, %d]", refs(1), refs(2), errors(1), ...
                    errors(2))
            end
            minRef = refs(errors == min(errors));
            maxRef = refs(errors == max(errors));

            for refIdx = 3:numberOfReferences
                newRef = (minRef + maxRef) / 2;
                references(refIdx, dataIdx, delayIdx, cellIdx) = newRef;
                MSI.setSupplyChannelVoltageSafe(MSI.SUPPLY_CHANNEL_REF, ...
                    newRef);
                outputProgram = MSI.runProgram(inputProgram);
                errorCounters(:, refIdx, dataIdx, delayIdx, cellIdx) = ...
                    outputProgram.ErrorCounter(:);
                cellErrorCount = errorCounters(colIdx, refIdx, dataIdx, ...
                    delayIdx, cellIdx);
                if cellErrorCount < 512
                    minRef = newRef;
                else
                    maxRef = newRef;
                end
            end
        end
        fprintf("%c", 128)
    end
    fprintf("\n")
end

%% Calculate RBL voltages of cells of interest
% Calculate the likely RBL voltage (and spread?) for the specified cells,
% delays, and data polarities by fitting the references and normalised
% error counts to a normal CDF.

% Allocate results matrix
RBLVoltage = zeros(numel(dataList), numel(readPrechargeDurationList), ...
    size(cellList, 1));
RBLVoltageStd = zeros(numel(dataList), numel(readPrechargeDurationList), ...
    size(cellList, 1));

% Clear loop variables to remove size information
clear refVoltages avgOutputData

% Loop over cells, delays, and data lists
for cellIdx = 1:size(cellList,1)
    colIdx = cellList(cellIdx, 3) + 1;
    for delayIdx = 1:numel(readPrechargeDurationList)
        for dataIdx = 1:numel(dataList)
            refVoltages(:) = references(:, dataIdx, delayIdx, cellIdx);
            avgOutputData(:) = errorCounters(colIdx, :, dataIdx, ...
                delayIdx, cellIdx) / 1023;

            % Normal CDF fit (from https://nl.mathworks.com/matlabcentral/
            % answers/345595-fitting-cumulative-normal-distribution-
            % function-to-data)
            % Depending on if the average output goes from 0 to 1 or from 1
            % to 0, the CDF must be flipped.
            if avgOutputData(1) < 0.5
                fcn = @(b,x) normcdf(x, b(1), b(2));
            else
                fcn = @(b,x) 1 - normcdf(x, b(1), b(2));
            end
            NRCF = @(b) norm(avgOutputData - fcn(b, refVoltages)); 
            B = fminsearch(NRCF, [0.55; 5e-2]);

            RBLVoltage(dataIdx, delayIdx, cellIdx) = B(1);% - interp1( ...
                %referenceSAoffset(colIdx, :, memoryNo + 1), ...
                %offset(colIdx, :, memoryNo + 1), B(1), 'spline');
            RBLVoltageStd(dataIdx, delayIdx, cellIdx) = B(2);
            
        end
    end
%     figure(80 + cellIdx)
%     legendText = string.empty(numel(dataList),0);
%     hold on
%     for dataIdx = 1:numel(dataList)
%         errorbar(readPrechargeDurationList, RBLVoltage(dataIdx, :, cellIdx), ...
%             RBLVoltageStd(dataIdx, :, cellIdx))
%         legendText(dataIdx) = sprintf("Data = %d", dataList(dataIdx));
%     end
%     hold off
%     xlabel("Read precharge duration setting (16-step delay chain)")
%     ylabel("RBL voltage + SA offset/noise [V]")
%     legend(legendText)
end

figure(82)
hold on
for cellIdx = 1:size(cellList,1)
    errorbar(readPrechargeDurationList, RBLVoltage(1, :, cellIdx), ...
        RBLVoltageStd(1, :, cellIdx), 'r')
    errorbar(readPrechargeDurationList, RBLVoltage(2, :, cellIdx), ...
        RBLVoltageStd(2, :, cellIdx), 'b')
end
hold off
xlabel("Read precharge duration setting (16-step delay chain)")
ylabel("RBL voltage + SA offset/noise [V]")

%% Save data
% Save the measurement data from the active workspace into a .mat file in
% the test directory.

save(workspaceName, "startTemperature", "logFileName", "errorCounters", ... 
    "references", "cellList", "readPrechargeDurationList", "dataList", ...
    "settingsRegister", "RBLVoltage", "RBLVoltageStd")
